	static char sccsid[]="%Z%	backref.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	backref.c [%P%]

Description:

	This module contains those functions necessary to handle backreferencing.
	Backreferencing, available in "grep" and in "cgrep" in "grep" mode,
	is the ability to match the exact same text that previously matched a
	regular expression.

	For example, "cgrep -G '\([A-Za-z])\1' <files>" matches any letter that
	occurs twice in succession.

Functions:

	re_brexec()
	brfn()
	Memccp()
	re_brinit()
	spew()
	seq()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

static int Memcmp();
static void seq();
static void spew();
#ifdef DEBUG
static void eg_brpr();
extern void eg_epr();
#endif

typedef enum BRFN_RET {
	CONTINUE,
	NEXT,
	DONE
} BRFN_RET; 
static BRFN_RET brfn();

static CHAR *Save_b, *Save_e;
static CHAR *Nb, *Ne;
static BOOL ChgGroup;

/******************************************************************************

Routine:
	re_brexec()

Syntax:

	BOOL re_brexec(pat, fcn, b, e)

	PATTERN *pat;	// Pointer to pattern to be matched
	BOOL (*fcn)();	// Pointer to subsequent pattern matching function
			//	(0 if this is the final algorithm used)
	CHAR *b;	// Pointer to first character of text to be searched
	CHAR *e;	// Pointer to last character of text to be searched

Description:

	re_brexec() is the main routine for text matching of regular expressions
	containing backreferencing.

	Regular expressions containing backreferencing are broken up during
	compilation into subexpressions, which are matched sequentially.
	brfn() is called recursively to match each successive subexpression.

	When all subexpressions are matched, brfn() calls grepmatch() for each
	match that is found, to store any match information that needs to be stored.
	After a match is found, re_brexec() will continue to search for matches,
	unless grepmatch() indicates by its return value that the search should stop.

Global Variables:

	None

Files:
	None

Return Status:

	YES 	// At least one match was found, and the search was then
		//   terminated because grepmatch() indicated that
		//   it should be terminated.

	NO	// The entire text was searched for matches.   Any number
		//   of matches may have been found.

*******************************************************************************/

BOOL
re_brexec(pat, fcn, b, e)
PATTERN *pat;
BOOL (*fcn)();
CHAR *b, *e;
{
#ifdef	DEBUG
	if (TRACE(3)) {
		PR("dfabr(%d->%d, %d, %d)\n", pat->re_ptr, pat->re_ptr->br, b, e);
		eg_brpr(pat->re_ptr->br);
	}
#endif
	/* Save the last two arguments as file scoped statics, for convenience. */
	Save_b = b;
	Save_e = e;

	/* Zero pointers to the beginning and end of matched text. */
	Nb = Ne = 0;

	/* Search the text for matches. */
	while (b <= e) {
		/* Search for the next match, and store it, if found.  */
		if (brfn(pat, fcn, pat->re_ptr->br, b, e) == DONE) {
			/* We found and stored a match, and brfn() has told
			 * us we need look for no further matches.  Return YES,
			 * indicating we need not search for any more matches.
			 */
			return(YES);
		}
		if (!Ne) {
			/* We did not find a match this time.
			 * Break out of the loop.
			 */
			break;
		}
		/* We found a match to the pattern.  Set <b> to the end of the
		 * match, so we start the new search there.
		 */
		b = Ne;
		/* Zero pointers to the beginning and end of matched text.
		 * They will be filled in if a match is found.
		 */
		Nb = Ne = 0;
	}
	/* We searched the entire text for matches. */
	return(NO);
}

/******************************************************************************

Routine:
	brfn()

Syntax:

	static BRFN_RET brfn(pat, fcn, br, b, e)

	PATTERN *pat;	// Pointer to pattern to be matched
	BOOL (*fcn)();	// Pointer to subsequent pattern matching function
			//	(0 if this is the final algorithm used)
	BR *br;		// Pointer to BR struct for next subexpression
	CHAR *b;	// Pointer to first character of text to be searched
	CHAR *e;	// Pointer to last character of text to be searched

Description:

	brfn() is called recursively to match each subexpression.

	Regular expressions containing backreferencing are broken up during
	compilation into subexpressions, which are matched sequentially.
	brfn() is called recursively to match each successive subexpression.

	When all subexpressions are matched, brfn() calls grepmatch() for each
	match that is found, to store any match information that needs to be stored.
	After a match is found, brfn() will continue to search for matches,
	unless grepmatch() indicates by its return value that the search should stop.

Global Variables:

	None

Files:
	None

Return Status:

	DONE 	// A match was found to the entire pattern, and grepmatch()
		//   has indicated that we need not search for more matches.

	NEXT 	// A match was found to the entire pattern, and grepmatch()
		//   has indicated that we should search for more matches, but
		//   we do not need to search for extensions of the same match.

	CONTINUE// A match may or may not have been found.  If a match was
		//   found, grepmatch() has indicated that we not only 
		//   should search for more matches, but that we also have
		//   to search for extensions to the same match in order
		//   to make sure that we find the longest possible match.

*******************************************************************************/

static BRFN_RET
brfn(pat, fcn, br, b, e)
PATTERN *pat;
BOOL (*fcn)();
BR *br;
CHAR *b, *e;
{
	BRFN_RET ret;
	register STATE *s, *t;
	CHAR *start;
	int length;
	CHAR *save_b = b;

	/* Are there more subexpressions to match? */
	if (br == 0) {
		/* We have matched all subexpressions.  Call grepmatch()
		 * to store the match, if appropriate.
		 */
		if (grepmatch(pat, fcn, Save_b, Save_e,
		   pat->re_ptr->br->start, b-1, &Nb, &Ne)) {

			/* grepmatch() indicates we not look for more matches. */
			return(DONE);
		}
		/* If grepmatch() extended the match beyond the end of the
		 * pattern, then the full extent of the match need not be
		 * known.  Return NEXT to indicate that we can begin looking
		 * for the next match
		 */
		if (Ne >= b) return(NEXT);

		/* Return CONTINUE to indicate that we need the fullest extent of
		 * this match, so we must keep looking for longer versions of
		 * this match.
		 */
		return(CONTINUE);
	}

	/* <br> points to the next subexpression to match.  Initialize to zero
	 * the start and end points of the matched text to this subexpression.
	 */
	br->start = br->end = 0;

	/* Switch on the type of subexpression. */
	switch(br->type) {
	case br_brplus:
	case br_brstar:
	case br_brquest:
	case br_br:
		/* The next subexpression is a Backref, or a Backref Plus,
		 * Star, or Quest.  Get the start of the text and the length
		 * of the text matched by the corresponding group.
		 */
		start = br->group->first->start;
		length = br->group->last->end + 1 - start;

		/* If the type of subexpression is br_brstar or br_brquest,
		 * or if the length of text matching the corresponding group
		 * is zero, this subexpression matches the null string.
		 * Therefore, check if the rest of the pattern matches the
		 * text starting at <b>.
		 */
		if (br->type == br_brstar || br->type == br_brquest || !length) {
			/* Set the match to this subexpression to be the null
			 * string.
			 */
			br->start = b;
			br->end = b - 1;
			/* Now check for matches to the rest of the pattern. */
			ret = brfn(pat, fcn, br->br_next, b, e);

			/* If brfn() indicated we should not continue looking
			 * for longer versions of this match, we return.
			 */
			if (ret != CONTINUE) return(ret);

			/* If the length of text that matched the group is
			 * zero, we return, since we have already handled the
			 * only possible match to this subexpression.
			 */
			if (!length) return(CONTINUE);
		}
		/* Check if the next <length> characters beginning at start
		 * are a match.
		 *
		 * If there are at least <length> characters remaining in the
		 * text we are searching, call Memcmp() to compare the
		 * next <length> characters in the text with the <length>
		 * characters that matched the corresponding group.
		 */
		while (b + length <= e+1 && !Memcmp(b, start, length, pat->cmap)) {
			/* We have matched the subexpression.  Set br->start and
			 * br->end to point to the first and last characters in
			 * the matched text, and advance <b> to the character
			 * just after the last matched character.
			 */
			br->start = save_b;
			b += length;
			br->end = b - 1;
			/* Call brfn() to check if the text beginning at <b>
			 * matches the next subexpression.
			 */
			ret = brfn(pat, fcn, br->br_next, b, e);

			/* If brfn() indicated we should not continue looking
			 * for longer versions of this match, we return.
			 */
			if (ret != CONTINUE) return(ret);

			/*
			 * If <br->type> is "br_br" or "br_brquest", we return, since
			 * the subexpression can match the text at most once.
			 */
			if (br->type == br_br || br->type == br_brquest) {
				return(CONTINUE);
			}
			/* <br->type> is "br_brstar" or "br_brplus".  We can
			 * match repeated instances, so we keep looking for
			 * more matches.
			 */
		} /* end of while loop */
		break;
	case br_re:
		/* The subexpression is a regular expression.
		 * Get the initial state of the regular expression.
		 */
		s = &br->r->states[0];

		/* Check if the initial state is a match (i.e., if the regular
		 * expression matches the null string.
		 */
		if (s->is_match) {
			/* This subexpression matches the null string.
			 * Set the match to this subexpression to be the null
			 * string.
			 */
			br->start = b;
			br->end = b - 1;
			/* Now check for matches to the rest of the pattern. */
			ret = brfn(pat, fcn, br->br_next, b, e);

			/* If brfn() indicated we should not continue looking
			 * for longer versions of this match, we return.
			 */
			if (ret != CONTINUE) return(ret);
		}
		/* Now, we read text characters, beginning at <b>, one at a time,
		 * but not going beyond <e>.
		 */
		for (; b <= e; b++) {
			/* Get the next mapped character */
			register CHAR c = br->r->cmap[*b];

			/* Get the next state of the finite state machine. */
			if ((t = s->tab[c]) != 0) s = t;
			else s = nextstate(br->r, s, (int)c);

			/* If this is the Empty state, we are done.  Stop
			 * reading text.  The Empty State occurs when the
			 * expression is anchored, and we have a mismatch.
			 */
			if (s == &Empty_state) break;

			/* If we don't have a match yet, read the next character. */
			if (!s->is_match) continue;

			/* We have a match to the subexpression.  <b> points to the
			 * last character of the match.
			 */
			br->end = b;

			if (br->r->anchored) {
				/* The match is anchored. Set the start
				 * of the matched text to <save_b>.
				 */
				br->start = save_b;
			} else if (br->r->length) {
				/* The match to the subexpression always
				 * has the same length. Set the start of the
				 * match.
				 */
				br->start = br->end + 1 - br->r->length;
			} else if (!pat->count_ptr) {
				/* We don't care where the match to this
				 * subexpression begins.
				 */
				br->start = br->end;
			} else if (!pat->allbounds) {
				/* This subexpression has its reversal stored
				 * in pat->count_ptr, and we just need to know
				 * the start of the longest possible match.
				 */
				br->start = get_beg(pat->count_ptr, save_b, b);
			} else {
				/* This subexpression has its reversal stored
				 * in pat->count_ptr, and we need to know
				 * the start of every possible match.
				 */
				register RE_RE *_r = pat->count_ptr;
				register CHAR *_e = b;
				register STATE *_s, *_t;
				/* Check if the initial state is a match
				 * (i.e., if the regular expression matches
				 * the null string).
				 */
				_s = &_r->states[0];
				if (_s->is_match) {
					/* This subexpression matches the null
					 * string.  Set the match to this
					 * subexpression to be the null
					 * string.
					 */
					br->start = b+1;
					/* Now check for matches to the rest
					 * of the pattern.
					 */
					ret = brfn(pat, fcn, br->br_next, b+1, e);

					/* If brfn() indicated we should not
					 * continue looking for longer versions
					 * of this match, we return.
					 */
					if (ret != CONTINUE) return(ret);
				}
				/* Now, we read text characters backwards,
				 * beginning at <e>, one at a time,
				 * but not going beyond <save_b>.
				 */
				for (; _e >= save_b; _e--) {
					register CHAR _c = _r->cmap[*_e];

					/* Get the next state of the finite
					 * state machine.
					 */
					if ((_t = _s->tab[_c]) != 0) _s = _t;
					else _s = nextstate(_r, _s, (int)_c);

					/* If this is the Empty state, we are done.
					 * Stop reading text.
					 */
					if (_s == &Empty_state) break;
					if (_s->is_match) {
						/* We have a match to the
						 * subexpression.  <_e> points to
						 * the first character of the match.
						 */
						br->start = _e;
						/* Now check for matches to the rest
						 * of the pattern.
						 */
						ret = brfn(pat, fcn, br->br_next, b+1, e);

						/* If brfn() indicated we should
						 * not continue looking for longer
						 * versions of this match, we return.
						 */
						if (ret != CONTINUE) return(ret);
					}
				}
				/* In this case, we have already checked for
				 * matches to the rest of the pattern.
				 * Continue reading more text.
				 */
				continue;
			}
			/* Now check for matches to the rest of the pattern. */
			ret = brfn(pat, fcn, br->br_next, b+1, e);

			/* If brfn() indicated we should not continue looking
			 * for longer versions of this match, we return.
			 */
			if (ret != CONTINUE) return(ret);
		} /* End of for loop reading text characters */
		break;
	}
	return(CONTINUE);
}

/******************************************************************************

Routine:
	Memcmp()

Syntax:

	static int Memcmp (s1, s2, n, cmap)

	CHAR *s1;	// Pointer to first string
	CHAR *s2;	// Pointer to second string
	int n;		// Number of characters in each string
	CHAR *cmap;	// Character map

Description:

	Compares two character arrays of the same length to see
	if they are identical (after being mapped).  Mapping
	handles the case where the match is case insensitive.

Global Variables:

	None

Files:
	None

Return Status:

	0	// Character arrays are identical.
	>0	// Character arrays are not identical.  Positive integer
		//   returned is <n> minus the length of the longest
		//   common initial substring of the two strings.

*******************************************************************************/

static int
Memcmp (s1, s2, n, cmap)
register CHAR *s1, *s2;
register int n;
register CHAR *cmap;
{
	while (--n >= 0 && cmap[*s1++] == cmap[*s2++]);
	return(++n);
}

/******************************************************************************

Routine:
	re_brinit()

Syntax:

	void re_brinit(pat, r)

	PATTERN *pat;	// Pointer to pattern to be matched
	RE_RE	*r;	// Pointer to compiled regular expression structure.

Description:

	If backreferencing is used in a pattern, "re_brinit()" must be called
	after the regular expression is compiled (by "re_recomp()"), but before
	any text matching is attempted (by "re_brexec()").  "re_brinit()" calls
	"seq()" to break the expression down in to subexpressions and to allocate
	the backreferencing structures needed for each subexpression.  "re_brinit()"
	calls "re_reinit()" to perform necessary PRE_MATCHING initialization
	of the regular expression (RE_RE) structure for each subexpression that
	is a regular expression before it can be used for text matching.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
re_brinit(pat, r)
PATTERN *pat;
RE_RE *r;
{
	register GR *gr;
	register BR *br;

	/* Allocate space for 10 "GR" structures for backreference groups.
	 * Since possible groups are numbered 1-9, array element 0 will be unused.
	 */
	pat->groups = (GR *)cgcalloc(10, sizeof(GR), "BR groups");

	/* Initialize the pointer to the chain of "BR" structures to zero. */
	r->br = 0;

	/* Call seq() to traverse the regular expression, breaking it into
	 * subexpressions, and setting up the chain of "BR" structures, one for
	 * each subexpression.
	 */
	seq(r, r->root, pat->groups);

	/* Now, step through the subexpressions, from left to right.
	 * We need to initialize all "br_re" subexpressions for matching.
	 * "br_br" and "br_star" subexpressions need no additional initialization.
	 */
	for (br = r->br; br; br = br->br_next) {
		/* Check if we have a regular expression ("br_re") subexpression. */
		if (br->type == br_re) {
			register RE_RE *re = br->r;

			/* Create a new "EOP" expression to be the new root of
			 * the subexpression, making the old root its child.
			 */
			re->root = newexpr(re, EOP, '#', re->root, (EXPR *)0);

			/* newexpr() augmented r->maxid, and gave the root of
			 * the subexpression this new ID number.  We reuse the
			 * ID number of the root of the entire expression,
			 * and we therefore decrement <re->maxid>.
			 */
			re->root->id = r->root->id;
			re->maxid--;

			/* Fill in the minimum and maximum lengths of text
			 * that can match the subexpression.  There is no
			 * approximate matching for backreferences.
			 */
			re->min_length = re->root->min_length;
			re->max_length = re->root->max_length_ex;

			/* If the minimum and maximum lengths are the same,
			 * the length is fixed.  Set it if it is fixed, or
			 * set it to 0 otherwise.
			 */
			re->length = (re->min_length == re->max_length ? re->min_length : 0);

			/* Check if this new regular expression has a nested
			 * Group or Backref.  This should never happen,
			 * because seq() and spew() separate such expressions
			 * into their individual components.
			 */
			re->backref = re->root->backref || re->root->group;

			if (re->backref == YES) {
				exitmsg0("%s: INTERNAL ERROR: Composite Group\n");
			}
			re_reinit(re);
			/* All subexpressions but the very first are "anchored".
			 * This means that the matched text must begin at the
			 * very first text character scanned.
			 * The first subexpression is anchored only if the
			 * entire expression is anchored.
			 */
			br->r->anchored = (br == r->br ? r->anchored : YES);
		} /* end of if (br->type == br_re) */
	} /* end of loop over subexpressions */

	/* We may need to know the exact beginning of the text matched by
	 * the first subexpression.
	 * If the subexpression has a fixed length, or if it is anchored,
	 * we can easily find its beginning, so we do nothing special.
	 */
	if (r->br->r->length || r->br->r->anchored) return;

	/* If the first subexpression is a Group or part of a Group, we need
	 * to know every possible exact beginning of the match (not just the
	 * longest match).  Set <pat->allbounds> to indicate this.
	 */
	for (gr = pat->groups+1; gr < pat->groups+10; gr++) {
		if (gr->first == r->br) {
			pat->allbounds = YES;
			break;
		}
	}
	/* If we need to know the exact extent of the match, call set_count_ptr
	 * to initialize an expression that is the reversal of the first
	 * subexpression.  There are three cases of this:
	 *
	 * 1) The first subexpression is a Group or part of a Group for
	 * backreferencing.  (pat->allbounds == YES)
	 *
	 * 2) We need to know the exact extent of the match to the entire
	 * pattern.  (pat->pbounds == YES)
	 *
	 * 3) The first subexpression can match text that causes the match
	 * to span a new-line, thereby possibly extending the match to additional
	 * lines.
	 */
	if (pat->allbounds || pat->pbounds || interior_nl(r->br->r->root, LEFT)) {
		pat->count_ptr = set_count_ptr(r->br->r);
	}
}

#ifdef	DEBUG
static void
eg_brpr(br)
BR *br;
{
	BR *b;
	char buf[EPRINTSIZE];

	for (b = br; b; b = b->br_next) {
		PR("%d@", b);
		switch(b->type) {
		case br_brplus:
			PR("BRPLUS(%d) ", b->group);
			break;
		case br_brstar:
			PR("BRSTAR(%d) ", b->group);
			break;
		case br_brquest:
			PR("BRQUEST(%d) ", b->group);
			break;
		case br_br:
			PR("BR(%d) ", b->group);
			break;
		case br_re:
			eg_epr((EXPR *)b->r->root, buf);
			if (((EXPR *)b->r->root)->backref) PR("X");
			if (((EXPR *)b->r->root)->group) PR("G");
			if (b->group) PR("GROUP%d(%s) ", b->group, buf);
			else PR("RE(%s) ", buf);
			break;
		default:
			PR("BADTYPE(%d) ", b->type);
			break;
		}
	}
	PR("\n");
}
#endif

/******************************************************************************

Routine:
	spew()

Syntax:

	static void spew(r, type, e, groups, group)

	RE_RE	*r;	// Pointer to compiled regular expression structure.
	BR_TYPE type;	// Type of subexpression to create or augment.
			//	br_re:		regular expression
			//	br_br:		Backref
			//	br_brstar:	Backref Star
			//	br_brplus:	Backref Plus
			//	br_brquest:	Backref Quest
	EXPR	*e;	// Pointer to current expression being scanned.
	GR *groups;	// Pointer to array of GR structures for Groups
	GR *group;	// Pointer to GR structure for Group referenced by Backref
			//	(0 if BR_TYPE is "re_re")


Description:

	"spew()" either creates a new subexpression or augments the one currently
	being scanned.  If <type> is "br_re", "spew()" must determine if <e>
	can be added to the current subexpression "r->br", or if <e> begins
	a new subexpression.  For other values of <type>, spew() always
	creates a new subexpression.  If it is determined that the current
	subexpression is to be augmented, "spew()" calls "newexpr()" to do so,
	and then returns.

	Otherwise, "spew()" creates a BR structure for the new subexpression,
	adding it to the head of the subexpression chain.  Then, "spew()"
	initializes the BR structure.  For subexpressions of type "br_re",
	this requires the allocation of an "RE_RE" structure to hold the
	regular expression.

	Finally, "spew()" loops over each backreferencing Group.  If this
	particular subexpression is inside one or more backreferencing groups,
	"spew()" updates as appropriate the "first" and "last" pointers in
	the GR structures for the corresponding backreferencing Groups.

Global Variables:

	ChgGroup

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
spew(r, type, e, groups, group)
RE_RE *r;
BR_TYPE type;
EXPR *e;
GR *groups;
GR *group;
{
	BR *br;
	GR *gr;

	/* We must check whether <e> is the first expression in a new
	 * subexpression, or if it is next expression in the current
	 * subexpression.  For it to be the next expression in the
	 * current subexpression, all of the following must be true:
	 *
	 * 1) It must have type "br_re", (i.e., it does not contain a Backref).
	 *
	 * 2) There must be a current subexpression "r->br", and
	 * it must also have type "br_re".
	 *
	 * 3) The ChgGroup flag must be off, meaning that the scan has
	 * not just entered or exited a backreferencing Group.
	 */
	if (type == br_re && r->br && r->br->type == br_re && !ChgGroup) {
		/* <e> is the next character in the subexpression.
		 * Concatenate <e> with the previous root of the
		 * subexpression.  This Cat becomes the new root of the
		 * subexpression.  The concatenation goes from right to
		 * left, since we are scanning from right to left.
		 */
		r->br->r->root = newexpr(r, Cat, '\0', e, r->br->r->root);
		return;
	}
	/* We are starting a new subexpression */

	/* Turn off the ChgGroup flag. */
	ChgGroup = NO;

	/* Allocate space for a new "BR" structure for the new subexpression. */
	br = (BR *)cgmalloc(sizeof(BR), "back ref");

	/* Set the type of "BR". */
	br->type = type;

	/* Add this "BR" to the head of the chain of "BR"s. */

	br->br_next = r->br;
	r->br = br;

	/* Type dependent initialization of the BR structure. */
	switch(type) {
	case br_br:
	case br_brplus:
	case br_brstar:
	case br_brquest:
		/* For a br_br*, set the number of the corresp. Group. */
		br->group = group;
		break;
	case br_re:
		/* Set the number of the Group to zero. */
		br->group = 0;

		/* Allocate a separate regular expression ("RE_RE")
		 * structure for matching this subexpression.
		 * The <r> member of the "BR" struct is set to point to it.
		 */
		br->r = (RE_RE *)cgmalloc(sizeof(RE_RE), "re_br");

		/* Copy the "RE_RE" structure for the whole expression
		 * into it.  This serves to initialize most members
		 * to the correct values.
		 */
		(void)memcpy(br->r, r, (int)sizeof(*r));

		/* The regular expression needs a pointer to the
		 * corresponding "BR" structure.
		 */
		br->r->br = br;

		/* Set the root expression of this "RE_RE" to the current
		 * expression.
		 */
		br->r->root = e;
		break;
	}

	/* For each backreferencing Group, we need to know the first and
	 * last subexpressions that are contained in the group.  Therefore,
	 * we loop over the nine possible backreferencing groups.
	 */
	for (gr = groups+1; gr < groups+10; gr++) {
		/* See if we are inside the Group */
		if (gr->inside) {
			/* We are inside the group.  Since we are
			 * scanning from right to left, we scan the
			 * last subexpression first.  So, if the
			 * last subexpression in the group is not yet set,
			 * then this is it.
			 */
			if (!gr->last) gr->last = br;
			/* This latest subexpression is now the first one
			 * in the group.
			 */
			gr->first = br;
		}
	}

}

/******************************************************************************

Routine:
	seq()

Syntax:

	static void seq(r)

	RE_RE	*r;	// Pointer to compiled regular expression structure.
	EXPR	*e;	// Pointer to current expression being scanned.

Description:

	"seq()" scans the regular expression containing backreferencing
	from right to left to break the expression down in to subexpressions.
	Each subexpression will be either a regular expression containing
	neither a backreferencing Group nor a Backref entity, or it will
	be a Backref entity.

	For example, if the regular expression is "abc\(d\(ef\)\)gh*i\1jkl\2*mno",
	there will be BR structures set up as follows:

	Subexpression:	Type of BR structure

	"abc"		br_re
	"d"		br_re
	"ef"		br_re
	"gh*i"		br_re
	"\1"		br_br
	"jkl"		br_re
	"\2*"		br_brstar
	"mno"		br_re

	When "seq()" encounters expression <e>, it checks to see if <e>
	can be a single subexpression or can be contained in a single 
	subexpression.  If so, "seq()" calls "spew()" to either add the
	expression to the current subexpression or to begin a new
	subexpression.  If not, "seq()" calls itself recursively, stepping
	down the expression tree.
	

Global Variables:

	ChgGroup

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
seq(r, e, groups)
RE_RE *r;
EXPR *e;
GR *groups;
{
	GR *group;

	/* Switch on the type of expression we are scanning. */
	switch(e->type) {
	case Alternate:
		re_error("Only one expression allowed when backreferencing.");
		break;
	case Cat:
		/* If the concatenation contains a Backref or a Group,
		 * it may be part of multiple subexpressions.  We sequence
		 * through the concatenation from left to right.
		 */
		if (e->backref || e->group) {
			/* Sequence through the pattern from right to left */
			seq(r, e->r, groups);
			seq(r, e->l, groups);
		} else {
			/* The concatenation does not contain a Backref
			 * or a Group.  It must fall entirely within one
			 * subexpression.  Call spew() to create
			 * a new subexpression or to add the concatenation
			 * to the current subexpression.
			 */
			spew(r, br_re, e, groups, (GR *)0);
		}
		break;
	case Group:
#ifdef	DEBUG
		printf("SPEW GROUP NO: %c\n", e->lit); 
#endif
		/* Get a pointer to the BR structure for the Group. */
		group = &groups[e->lit - '0'];

		/* A Group change always begins a new subexpression. */
		ChgGroup = YES;

		/*  Indicate we are entering the group. */
		group->inside = YES;

		/* If the Group contains a nested Backref or Group,
		 * it may be part of multiple subexpressions.  We sequence
		 * through the concatenation from left to right.
		 */
		if (e->l->backref || e->l->group) {
			/* Sequence inside the group */
			seq(r, e->l, groups);
		} else {
			/* The Group does not contain a nested Backref
			 * or Group.  It must comprise one subexpression.
			 * Call spew() to create the new subexpression.
			 */
			spew(r, br_re, e->l, groups, (GR *)0);
		}
		/* Indicate that we have left the group. */
		group->inside = NO;

		/* A group always consists of one or more complete
		 * subexpressions.  The next entity will begin a new
		 * subexpression.
		 */
		ChgGroup = YES;
		break;
	case EOP:
		/* Sequence through the pattern */
		seq(r, e->l, groups);
		break;
	case Backref:
#ifdef	DEBUG
		printf("SPEW BR NO: %c\n", e->lit); 
#endif
		/* Get a pointer to the BR structure for the corresp. Group. */
		group = &groups[e->lit - '0'];

		/* Create a "br_br" BR structure for the Backref (e.g. "\1"). */
		spew(r, br_br, (EXPR *)0, groups, group);
		break;
	case Plus: case Star: case Quest:
		/* See if this is a Backref Plus, Star, or Quest (e.g. "\1*").
		 * NOTE: Due to optimization, we can wind up with a Backref Plus
		 * or a Backref Quest.
		 */ 
		if (e->l->type == Backref) {
			BR_TYPE br_type;
		/* Get a pointer to the BR structure for the corresp. Group. */
			group = &groups[e->l->lit - '0'];

			/* Get the type of Backref */
			switch (e->type) {
			case Plus:
#ifdef	DEBUG
				printf("SPEW BRPLUS NO: %c\n", e->l->lit); 
#endif
				/* Create a "br_brplus" BR structure for the Backref Plus. */
				br_type = br_brplus;
				break;
			case Star:
#ifdef	DEBUG
				printf("SPEW BRSTAR NO: %c\n", e->l->lit); 
#endif
				/* Create a "br_brstar" BR structure for the Backref Star. */
				br_type = br_brstar;
				break;
			case Quest:
#ifdef	DEBUG
				printf("SPEW BRQUEST NO: %c\n", e->l->lit); 
#endif
				/* Create a "br_brquest" BR structure for the Backref Quest. */
				br_type = br_brquest;
				break;
			}
			spew(r, br_type, (EXPR *)0, groups, group);
		} else {
			/* It's not a Backref subexpression.  Start a new
			 * Subexpression or continue an existing one.
			 */
			spew(r, br_re, e, groups, (GR *)0);
		}
		break;
	default:
		/* Start a new Subexpression or continue an existing one.  */
		spew(r, br_re, e, groups, (GR *)0);
		break;
	}
}
