	static char sccsid[]="%Z%	compile.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	compile.c [%P%]

Description:

	compiles patterns into a compiled form that is usable for the matching
	algorithms

Functions:
	compile()
	fgrep_compile()
	egrep_compile()
	fixed_compile()
	set_count_ptr()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

static void fgrep_compile(), egrep_compile(), fixed_compile();

/******************************************************************************

Routine:
	compile()

Syntax:

	void compile(pat)

	PATTERN *pat;	//Pointer to pattern to be compiled.

Description:

	compile() is the main routine for compiling the pattern to be matched,
	i.e., putting the pattern into a structure that can be easily and
	speedily used by the text matching routines.

	compile() first calls either fgrep_compile() or egrep_compile()
	to do the compilation for fixed strings or full regular expressions
	respectively.

	Then, some small additional adjustments are done in two special cases
	that are not fully handled in the compilation routines.

	a) If we are matching the pattern backwards, different matching routines
	must be used.

	b) If -v is on for this pattern, additional adjustments may be needed.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
compile(pat)
PATTERN *pat;
{
	/* Do the compilation.   The compilation routine will set
	 * pat->procfn to point to the primary (i.e first stage)
	 * matching function.  pat->succfn will be set to point
	 * to the secondary (i.e. second stage) matching function,
	 * or to zero if there is no second stage.
	 * Multistage matching is used for many regular expressions.
	 * See egrep_compile() for more details on multistage matching.
	 */
	switch(pat->parser) {
	case fgrepparse:
	case nlparse:
	case nullparse:
		/* The pattern is now zero or more fixed strings. */
		fgrep_compile(pat);
		break;
	case egrepparse:
		/* The pattern is now a full regular expression. */
		egrep_compile(pat);
		break;
	case grepparse:
	default:
		/* cgrep -G mode should have been converted to one of the above. */
		re_error("INTERNAL ERROR: grep mode conversion error");
	}

	/* If the beginning delimiter is matched backwards, i.e., if the text
	 * is searched backwards from the match, we change each possible
	 * matching equivalent to its backwards equivalent.
	 */
	if (pat->backward) {
		/* We switch pat->procfn to its reverse equivalent.
		 * Since re_nullexec() matches nothing, we do nothing
		 * in that case.
		 */

		if (pat->procfn == re_cwexec) pat->procfn = re_cbexec;
		else if (pat->procfn == re_reexec) pat->procfn = re_rbexec;
		else if (pat->procfn == re_lnexec) pat->procfn = re_lbexec;
		else if (pat->procfn == re_brexec) {
			/* Since backreferencing is not reversible, we can't
			 * use re_brexec as the primary if we are matching
			 * backwards.  We therefore add a new primary stage
			 * in which we match every line in the backwards
			 * direction, and do the full matching in the second
			 * stage.
			 */
			pat->procfn = re_lbexec;
			pat->succfn = re_brexec;
			pat->thirdfn = 0;
		}
		/* The only possible reversible value of pat->succfn
		 * at this point is re_reexec (regular expression matching).
		 */
		if (pat->succfn == re_reexec) pat->succfn = re_rbexec;

	}
	/* Performance enhancement:
	 * If -v is on for this pattern, and the pattern matches
	 * every line or no line, we can turn -v off, by matching
	 * every line instead of no line, or vice versa.
	 */
	if (pat->vflag || pat->vfast) {
		if ((pat->procfn == re_lnexec || pat->procfn == re_lbexec) &&
		    pat->succfn == 0) {
			/* Pattern matches every line, but -v is on.
			 * Turn off -v, and match no line.
			 */
			pat->procfn = re_nullexec;
			pat->vflag = pat->vfast = NO;
		} else if (pat->procfn == re_nullexec) {
			/* Pattern matches no line, but -v on.
			 * Turn off -v, and match every line.
			 * NOTE: Since re_nullexec() matches no lines,
			 * don't check above that pat->succfn is 0.
			 */
			pat->procfn = (pat->backward ? re_lbexec : re_lnexec);
			pat->succfn = 0;
			pat->vflag = pat->vfast = NO;
		} 
	}

	/* If -v is on for this pattern, and we will be looking for matches
	 * on a line-by-line basis, (i.e., pat->vflag is set rather than
	 * pat->vfast), we must add an extra layer to implement the line
	 * by line search, unless the primary matching function is already
	 * set to match every line.
	 */
	if (pat->vflag && pat->procfn != re_lnexec && pat->procfn != re_lbexec) {
		pat->thirdfn = pat->succfn;
		pat->succfn = pat->procfn;
		pat->procfn = (pat->backward ? re_lbexec : re_lnexec);
	}
}

/******************************************************************************

Routine:
	fgrep_compile()

Syntax:

	void fgrep_compile(pat)

	PATTERN *pat;	//Pointer to pattern to be compiled.

Description:

	fgrep_compile() is the main routine for compiling a pattern that
	consists only of zero or more fixed strings, or that has been
	previously converted to such.   Except in the trivial cases
	where the pattern is known to match every line or no lines,
	the Commentz-Walter algorithm will be used to match the text.
	(The well known Boyer-Moore algorithm is actually a special case
	of the Commentz-Walter algorithm specifically for one string).

	The algorithm has four functions, listed in the order in which they
	are called.
		re_cwinit()	Initializes structures needed for compilation
		re_cwadd()	Adds a string to the compilation input
				Called once for each string.
		re_cwcomp()	Does the compilation
				(called by fixed_compile())
		re_cwexec()	Does the pattern matching with text.
				(Not called during compilation).

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
fgrep_compile(pat)
PATTERN *pat;
{
	register CHAR *s, *e;

	/* A single match to the pattern cannot span multiple lines of text.
	 * This is because any pattern that can span multiple lines must contain
	 * either a character class ([...]), or a literal new-line (either
	 * \<new-line> or, if -a was used, \n.  In any of these cases,
	 * the '[' or '\' metacharacter would have caused "full regular
	 * expression" format to be used, rather than fixed string format.
	 * Another way that a match can span multiple lines of text is if
	 * approximate matching is being used, in which case, "full regular
	 * expression" format is also used.
	 */
	pat->mult_lines = 0;

	/* Two special cases can be "compiled" trivially. */
	switch(pat->parser) {
	case nlparse:
		/* It is already known the pattern will match every line. */
		pat->procfn = re_lnexec;
		pat->succfn = 0;
		return;
	case nullparse:
		/* It is already known the pattern will match no line. */
		pat->procfn = re_nullexec;
		pat->succfn = 0;
		return;
	}

	/* Do the initialization for the Commentz-Walter algorithm. */
	pat->cw_ptr = re_cwinit(pat);

	/* Loop through the pattern to be matched. */
	s = pat->expression;
	while (*s) {
		/* We step through the pattern until we encounter
		 * a new line (the separator between multiple strings)
		 * or a null (end of pattern).
		 * Each character is mapped through pat->cmap[],
		 * which, if -i (case insensitive match) is on for
		 * this pattern, maps lower case letters into upper case.
		 */
		for (e = s; *e && (*e != '\n'); e++) *e = pat->cmap[*e];

		/* We found a new-line or the end of the pattern. */
		/* Add the string to the algorithm if it is not null. */
		if (e - s) {
			re_cwadd(pat->cw_ptr, s, e);
		}
		/* If we found a new-line, advance to the following character.
		 * Otherwise, we found a null, and we advance to the null
		 * which will terminate the loop.
		 */
		s = *e ? (e + 1) : e;
	}
	/* Since we have only fixed strings, there is no second stage
	 * of the matching algorithm.
	 */
	pat->succfn = 0;
	/* Now, choose an algorithm, and compile the pattern using the chosen
	 * algorithm.  Set pat->procfn to point to the text matching function
	 * for that algorithm.  This is all done by fixed_compile().
	 */
	fixed_compile(pat);
}

/******************************************************************************

Routine:
	egrep_compile()

Syntax:

	void egrep_compile(pat)

	PATTERN *pat;	//Pointer to pattern to be compiled.

Description:

	egrep_compile() is the main routine for compiling a pattern that
	is a regular expression.

	The relevant routines are not entirely analogous to those for
	the fixed string algorithm (Commentz-Walter) described in the
	prologue to fgrep_compile().

	The relevant functions, listed in the order in which they
	are called, are:

	re_recomp()	Does the compilation.

	re_reinit()	Does some post-compilation initialization.

	re_brinit()	Does some extra post compilation initialization
			  needed only if backreferencing.

	re_recw()	Attempts to find fixed strings embedded in the regular
			  expression.  If successful, an initial match can be
			  done to the fixed strings using a faster algorithm,
			  so that the slower regular expression algorithm need
			  not be used on the entire file.  In some cases,
			  re_recw() may find that the entire regular expression
			  is equivalent to fixed strings, and a fixed string
			  algorithm can be used for the entire match.

			For example, if our pattern is 'ab.*xyz', we can do a match
			  to find "xyz" (the longest literal substring) first,
			  and then check to see if the full regular expression
			  is matched.  If our pattern contains multiple strings
			  (e.g. 'ab.*one' and 'two..ui') we can use a fixed string
			  algorithm to find 'one' or 'two' first, and then check
			  to see if the full regular expression is matched.

	re_reexec()	Does the pattern matching with text.
			   (Not called during compilation).

	re_rbexec()	Does the pattern matching with text in the case where
			  the text is being scanned backwards.
			   (Not called during compilation).

	re_brexec()	Does the pattern matching with text in the special case
			  where backreferencing is in use.
			  (Not called during compilation).

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
egrep_compile(pat)
PATTERN *pat;
{
	int bestlen;

#ifdef	DEBUG
	printf("PATTERN %s\n", pat->expression);
#endif
	/* Compile the pattern as an egrep regular expression */
	pat->re_ptr = re_recomp(pat);

	/*
	 * re_recw() reads the compiled regular expression.
	 * It tries to set up a match to a set of one or more
	 * fixed strings, which may be a single stage match or the
	 * first stage of a multistage match.  If appropriate,
	 * it calls re_cwinit() and re_cwadd() to set up input
	 * for compilation for the Commentz-Walter algorithm.
	 * It returns the following information:
	 *
	 * 	"bestlen", "pat->cw_ptr->npat", "pat->succfn"
	 *
	 *	If "bestlen" > 0, "pat->cw_ptr->npat" gives the number
	 *	of fixed strings in the potential Commentz-Walter match,
	 *	and "bestlen" gives the length of the shortest such string,
	 *	and "pat->succfn" indicates if the Commentz-Walter match
	 *	is a first stage match prior to a full regular expression
	 *	match (pat->succfn != 0), or if the fixed string match
	 *	is the only stage needed (pat->succfn == 0).
	 *
	 *	If "bestlen" == 0, no Commentz-Walter match is possible,
	 *	even a partial one, because one or more of the fixed strings
	 *	had zero length.  In this case, the return values of
	 *	"pat->cw_ptr->npat" and "pat->succfn" are not defined.
	 */
	pat->procfn = 0;
	bestlen = re_recw(pat, (CHAR **)0);
	/*
	 * Do Commentz-Walter match if possible, except that one character
	 * multistage matches are not cost-effective, so we don't use them.
	 */
	if (bestlen > 1 || (bestlen == 1 && !pat->succfn)) {
		/* We have a Commentz-Walter match, and it's either longer
		 * than a single character or it is a single stage match.
		 */
		/* Now, choose an algorithm, and compile the pattern using
		 * the chosen algorithm.  Set pat->procfn to point to the
		 * text matching function for that algorithm.
		 */
		fixed_compile(pat);
		/*
		 * If the match is multistage, a subsequent match must use
		 * the full regular expression matching algorithm.
		 * In the special case of backreferencing, we must use the
	         * algorithm specific to backreferencing.
		 */
		if (pat->succfn) {
			pat->succfn = (pat->backref ? re_brexec : re_reexec);
		} else {
			/* We have a single stage match to one or more
			 * fixed strings.  We can simply return, because
			 * we will not be doing any regular expression
			 * matching to this pattern.
			 */
			return;
		}
	}

	if (pat->procfn == 0 && pat->cost_limit > 0 && pat->cost_ins > 0) {
	/*
	 * We could not or would not do a Commentz-Walter match above.
	 * Furthermore, we are doing approximate matching, and the cost
	 * of an insertion of an extra character is non-zero.  In this
	 * case, we MAY still be able to do a first stage Commentz-Walter match.
	 *
	 * If we allow at most N errors, we can divide each string into
	 * N+1 substrings.  The text, then, must exactly match at least
	 * one of the substrings.  For example, cgrep -A1 '...abcdefgh...'
	 * can only match text containing either "abcd" or "efgh".
	 */
		register CHAR **pats;
		register int i, j, pieces, tmp;
		register short ctmp;
		unsigned int nwords;

		/* First, let's free up some memory allocated by re_recw()
		 * that we can reuse, namely the stored Commentz-Walter information.
		 */
		FREE(pat->cw_ptr->bmpat);
		FREE(pat->cw_ptr->shift1);
		FREE(pat->cw_ptr->tab);
		re_cwfreenodes(pat->cw_ptr->root);
		FREE(pat->cw_ptr);

		/* Save the maximum number of errors allowed.  */
		ctmp = pat->cost_limit;

		/* Turn off approximate matching, and call re_recw().
		 * re_recw() will store each fixed string in a Commentz-Walter
		 * structure as if we were exact matching.  We do this just
		 * to get a count on the number of patterns.
		 *
		 * Note also that re_recw() does not modify
		 * the compiled regular expression in pat->re_ptr, it only
		 * reads it.  So, the compiled regular expression itself
		 * will still be correct for approximate matching, and need
		 * not be recompiled.
		 */
		pat->cost_limit = 0;
		pat->approx = NO;
		(void)re_recw(pat, (CHAR **)0);

		/* Get the pattern count. */
		nwords = pat->cw_ptr->npat;

		/* Free the Commentz-Walter structures. */
		FREE(pat->cw_ptr->bmpat);
		FREE(pat->cw_ptr->shift1);
		FREE(pat->cw_ptr->tab);
		re_cwfreenodes(pat->cw_ptr->root);
		FREE(pat->cw_ptr);

		/* Call re_recw() yet again.  This time store each fixed string
		 * in the pats[] array.
		 *
		 * Note also that re_recw() does not modify
		 * the compiled regular expression in pat->re_ptr, it only
		 * reads it.  So, the compiled regular expression itself
		 * will still be correct for approximate matching, and need
		 * not be recompiled.
		 */
		pats = (CHAR **)cgmalloc2(nwords, sizeof(CHAR *), "approx pats");
		bestlen = re_recw(pat, pats);
		/* We have stored each fixed string.
		 * Restore the approximate matching parameters.
		 */
		pat->cost_limit = ctmp;
		pat->approx = YES;

		/* Calculate the minimum cost of a matching error. */
		ctmp = MIN(pat->cost_ins, pat->cost_del);
		ctmp = MIN(ctmp, pat->cost_sub);

		/* Calculate the number of pieces into which the pattern
		 * must be split for a first stage match.  See explanation
		 * above.
		 */
		pieces = (pat->cost_limit/ctmp) + 1; 

		/*  If the shortest fixed string in the pattern
		 *  is long enough to be split into the requisite number of
		 *  pieces, we split each string into the requisite number of
		 *  pieces, and set up for a Commentz-Walter match on the
		 *  whole set of pieces.
		 */
		if (bestlen/pieces > 1) {
			/* Initialize a Commentz-Walter structure. */
			pat->cw_ptr = re_cwinit(pat);

			/* Loop over each fixed string. */
			for (i = 0; i < nwords; i++) {
				/* Break the string into pieces,
				 * and store each piece in a
				 * Commentz-Walter structure.
				 */
				register CHAR *b, *e;
				b = pats[i];
				e = b + strlen((char *)b);
				for (j = pieces; j > 0 ; --j) {
					tmp = (e - b)/j;
					re_cwadd(pat->cw_ptr, b, b+tmp);
					b += tmp;
				}
			}
			/* Now compile the Commentz-Walter strings. */
			re_cwcomp(pat->cw_ptr);
			pat->procfn = re_cwexec;
			/* The second stage will be a full regular expression
			 * match.  (Backreferencing is illegal with
			 * approximate matching, so we don't have to check for
			 * it.)
			 */
			pat->succfn = re_reexec;
		}
		/* Now we can free the pats arrays. */
		for (i= 0; i < nwords; i++) {
			FREE(pats[i]);
		}
		FREE(pats);
	}
	if (pat->procfn == 0) {
		/*
		 * We cannot or choose not to do Commentz-Walter matching,
		 * even just as a first stage.
		 *
		 * We the full regular expression matching algorithm.
		 * In the special case of backreferencing, we must
		 * use the algorithm specific to backreferencing.
		 */
		pat->procfn = (pat->backref ? re_brexec : re_reexec);
		pat->succfn = 0;
	}
	/* At most, we have two stage matching at this point. */
	pat->thirdfn = 0;

	/* Regular expression matching requires some post compilation
	 * initialization.  Initialize the compiled regular expression.
	 */
	if (pat->re_ptr->backref) {
		/* Initialization for backreferencing */
		re_brinit(pat, pat->re_ptr);
	} else {
		/* Initialization for general regular expression. */
		re_reinit(pat->re_ptr);

		/* If we are using a general regular expression match,
		 * and we need to know the full extent of the match,
		 * store, if necessary, the reversed version of the expression
		 * in pat->count_ptr.  We do this so that beginning at
		 * the end of the match, we can search backwards for
		 * the beginning of the match.  set_count_ptr()
		 * creates a reversed copy of the regular expression,
		 */
		if (pat->pbounds || pat->mult_lines) {
			pat->count_ptr = set_count_ptr(pat->re_ptr);
		}
	}

	return;
}

/******************************************************************************

Routine:
	fixed_compile()

Syntax:

	void fixed_compile(pat)

	PATTERN *pat;	//Pointer to pattern to be compiled.

Description:

	fixed_compile() compiles a pattern that consists of zero or
	more fixed strings, or that has been previously converted to such.
	Except in the trivial cases where the pattern is matches every line
	or no lines, the Commentz-Walter algorithm will be used to match the text.
	The well known Boyer-Moore algorithm is a special case of the
	Commentz-Walter algorithm for one string.

	If pat->succfn is 0 upon input, the match is single stage, i.e., the
	entire pattern is, or has been converted to, zero or more fixed strings.

	If pat->succfn is non-zero upon input, the match is multistage, i.e.,
	the full pattern is a regular expression, and we are compiling a match
	to fixed strings that are a proper subset of the regular expression.
	For example, if the regular expression is 'ab.*xyz', we can do
	a Commentz-Walter match to find "xyz" (the longest literal substring)
	first, and then check to see if the full regular expression is matched.
	If our pattern contains multiple strings (e.g. 'ab.*one' and 'two..ui'),
	we can use Commentz-Walter to find 'one' or 'two' first, and then check
	to see if the full regular expression is matched.  This can be much
	faster than directly attempting the regular expression matching routine
	on the entire file.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
fixed_compile(pat)
PATTERN *pat;
{
	int nwords = pat->cw_ptr->npat;

	if (nwords == 0) {
		/* There is no pattern, so it matches nothing.  */
		pat->procfn = re_nullexec;
	} else {
		/* 
		 * Compile the pattern for matching using the Commentz-Walter
		 * algorithm, except in one special case.  If the pattern is
		 * a new-line, skip compilation, and just match every line.
		 * This is much faster, and is an important special case,
		 * as it occurs every time a line count is used to delimit
		 * windows (e.g. cgrep +-1 <pattern> <file>).
		 */
		re_cwcomp(pat->cw_ptr);
		pat->procfn = re_cwexec;
		nwords = pat->cw_ptr->npat;
		if (nwords == 1 && pat->cw_ptr->mindepth == 1 && pat->cw_ptr->bmpat[0] == NL) {
			/* The single string is just a new-line.
			 * Match every line.
			 */
			pat->procfn = re_lnexec;
		}
	}
}

/******************************************************************************

Routine:
	set_count_ptr()

Syntax:

	RE_RE *set_count_ptr(r)

	RE_RE *r;	// Pointer to the regular expression for which
			//   the full extent of each text match needs to be
			//   determined.

Description:

	set_count_ptr() is called after compilation for a regular expression
	when it is necessary to know the full length of the text matching that
	expression.

	Unless text matching the regular expression has a known fixed length,
	set_count_ptr() sets up a regular expression that is the reversal of the
	one pointed to by "r", so we can match backwards to find the beginning
	of the match.


Global Variables:

	None

Files:

	None

Return Status:

	0	// The  text matching the regular expression is constant length,
		//	in which case this length can be used directly.

	Otherwise, returns a pointer to a new regular expression suitable for
		matching in the reverse direction.

*******************************************************************************/

RE_RE *
set_count_ptr(r)
register RE_RE *r;
{
	register RE_RE *cp;

	/* If the expression has a fixed length, don't bother. We can just
	 * use the constant length.
	 */
	if (r->length) return((RE_RE *)0);

	/* Allocate space for the reversed compiled regular expression. */
	cp = (RE_RE *)cgmalloc(sizeof(RE_RE), "re_count");

	/* Copy the compiled regular expression into the allocated space. */
	(void)memcpy(cp, r, (int)sizeof(*r));

	/* Indicate the reversed direction.   Note that the regular expression
	 * does not have to be recompiled, since compiled regular expressions
	 * are inherently symmetric.  Furthermore, we need not even copy
	 * those portions of the compiled regular expression that were
	 * not copied by the memcpy() above (i.e, those that are accessed
	 * through pointers).  This is because they are not modified after
	 * compilation, and can therefore be shared.
	 *
	 * NOTE: Backreferencing is non-symmetric.  However, backreferencing
	 * is done piecemeal using multiple regular expressions, each of
	 * which is individually symmetric.
	 */
	cp->backward = (BOOL)(!r->backward);

	/* Anchor the reversed expression.  This means that the text matched
	 * must begin at the first character of the search.  This is needed to
	 * guarantee that the reversed match finds only the same instance
	 * of matched text as the original match.
	 */
	cp->anchored = YES;

	/* Do the post-compilation initialization.
	 * The reversed expression needs a private copy of these,
	 * separate from that of the original expression.
	 */
	re_reinit(cp);

	return(cp);
}
