	static char sccsid[]="%Z%	cw.c	%E% %U%";
/******************************************************************************

		###############################################
		# Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
		###############################################

Module:	cw.c [%P%]

Description:

	Contains the code for implementing the Commentz-Walter algorithm,
	one of the algorithms used to search for text matches to a pattern.
	The code for initialization, compilation, and matching is all included.

	This algorithm is used to simultaneously search for text matches to
	one or more fixed strings.

	The Commentz-Walter algorithm is well known in the literature as an
	algorithm for matching multiple strings.  When used to match a single
	string, it is essentially equivalent to the even better known Boyer-Moore
	algorithm.

Functions:
	re_cwinit()
	re_cwadd()
	re_cwcomp()
	newnode();
	lips();
	shift();
	shiftprop();
	shifttab();
	spurious();
	re_cwexec()
	re_cbexec()
	re_cwfreenodes()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"
#define		INCR	1000

static NODE *newnode();
static void lips(), shift(), shiftprop(), shifttab();
static void finish_trim();
static BOOL spurious();

static NODE *Next;	/* Pointer to next NODE to be used */
static NODE *Lim;	/* Pointer to next NODE to be memory allocated */

#define	ADD(N)	if (qtail) qtail = qtail->link = (N);\
			else qtail = qhead = (N)
#define	DEL()	{ NODE *tmp = qhead; if ((qhead = qhead->link) == 0)\
			qtail = 0; tmp->link = 0;}

/******************************************************************************

Routine:
	re_cwinit()

Syntax:
	RE_CW *re_cwinit(pat)

	PATTERN *pat;	// Pointer to the pattern to be matched.

Description:

	Initializes an instance of the structure used to implement
	the Commentz-Walter algorithm.  This must be called as the first step
	before attempting a match to a set of fixed strings using the
	Commentz-Walter algorithm.

Global Variables:

	None

Files:
	None

Return Status:

	Returns a pointer to an initialized Commentz-Walter structure.

*******************************************************************************/

RE_CW *
re_cwinit(pat)
PATTERN *pat;
{
	register RE_CW *c;

	/* Initialize the pointers to the nodes. */

	Next = Lim = 0;

	/* Allocate memory for a Commentz-Walter structure,
	 * and initialize its contents to zero.
	 */
	c = (RE_CW *)cgmalloc((unsigned int)(sizeof *c), "cwinit");
	(void)memset((char *)c, 0, (int)sizeof(RE_CW));

	/* Initialize minimum length of strings to match. */
	c->mindepth = LARGE;

	/* Initialize character mapping array. */
	c->cmap = pat->cmap;

	/* Initialize whether or not we need to know the full extent of a match. */
	c->fullextent = ((pat->pbounds || pat->mult_lines) ? YES : NO);

	/* Create a "root" node for the Commentz-Walter algorithm. */
	c->root = newnode(c, (CHAR)'\0', 0);

	/* return a pointer to the initialized structure. */
	return(c);
}

/******************************************************************************

Routine:
	re_cwadd()

Syntax:
	void re_cwadd(c, s, e)

	RE_CW *c;	// Pointer to Commentz-Walter structure previously
			//	returned by re_cwinit().
	CHAR *s;	// Pointer to first character of string.
	CHAR *e;	// Pointer to one past last character of string.

Description:

	Adds one string to Commentz-Walter structures for future compilation.

	NOTE:  If this structure is to be used for a match in the
	backward direction, each string must be reversed before being
	passed into this routine.  "s" must point to the first character
	of the reversed string, and "e" must point one character past
	the end of the reversed string.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
re_cwadd(c, s, e)
register RE_CW *c;
register CHAR *s, *e;
{
	register NODE *p, *state;
	int depth;

	/* Start with the last character in the string to be added,
	 * and with the initial state as the current state.
	 */
	for (state = c->root; s <= --e;) {
		/* Check if a NODE already exists for the characters
		 * we have scanned so far.  
		 */
		for (p = state->child; p; p = p->sibling) {
			if (p->llit == c->cmap[*e]) {
				break;
			}
		}
		if (p == 0) {
			/* There is no NODE for this character.  We must
			 * create one.  Break out of the loop.
			 */
			break;
		} else {
			/* A NODE already exists for the characters scanned
			 * so far.  Set the current state to that node.
			 */
			state = p;

			/* If this state is a match to the pattern, and we do
			 * not need to know the full extent of the match,
			 * we can ignore this string, since we are also
			 * searching for a proper substring.
			 */
			if (state->matched && !c->fullextent) return;
		}
		/* Get the previous character of the string to be added. */
	}
	/* Check that we have not run off the beginning of the string.
	 * If we ran off, this string is a terminal substring (not necessarily
	 * a proper substring) of a previously encountered string, and no new
	 * NODES will be created.
	 */
	if (s <= e) {
		/* We have not run off the beginning of the string.
		 * The depth of the next node will be one more than
		 * the depth of the last state encountered.
		 */
		depth = state->depth + 1;
		/* Create a new NODE for the new state. */
		p = newnode(c, *e--, depth++);
		/* Add this new state to the head of the <child> chain of
		 * children of the old state.
		 */
		p->sibling = state->child;
		state->child = p;
		/* Set <state> to the new state. */
		state = state->child;
		/* As long as there are characters remaining at the
		 * beginning of the string, add a new NODE for each one.
		 */
		while (s <= e) {
			state->child = newnode(c, *e--, depth++);
			state = state->child;
		}
	} else {
		/* We ran off the beginning of the string.  This string is
		 * a terminal substring of a previously encountered string.
		 * If we do not need to know the full extent of the match,
		 * remove the remainder of the previously encountered string,
		 * if any.
		 */
		if (!c->fullextent) state->child = 0;
	}
	/* <state->depth> is the length of the newly added string.
	 * If it is shorter than the minimum string length seen so far,
	 * increment the minimum string length.
	 */
	if (c->mindepth > state->depth) {
		c->mindepth = state->depth;
	}
	/* Mark that this last state is a match to the pattern. */
	state->matched = 1;

	/* Increment the number of patterns. */
	c->npat++;
}

/******************************************************************************

Routine:
	newnode()

Syntax:

	static NODE *newnode(c, llit, depth)

	RE_CW *c;	// Pointer to Commentz-Walter structure previously
			//	returned by re_cwinit().
	CHAR llit;	// Literal character
	int depth;	// Depth of node

Description:

	newnode() allocates one NODE structure for use by the Commentz-Walter
	matching algorithm.  The NODEs for a set of patterns to be matched form
	a tree.  newnode() sets the <llit>, <depth>, and <id> attributes,
	and initializes some other attribute values to 0.

	For example, suppose the following patterns are to be matched:

	one
	sane
	sent

	The nodes that will be generated (by re_cwadd()) are as follows:

	depth	llit	comment				parent node

	0	NA	root node			none

	1	'e'	'e' in "one" and "sane" 	root node
	1	't'	't' in "sent"			root node
	
	2	'n'	'n' in "one" and "sane"		'e' in "one" and "sane"
	2	'n'	'n' in "sent"			't' in "sent"

	3	'o'	'o' in "one"			'n' in "one" and "sane"
	3	'a'	'a' in "sane"			'n' in "one" and "sane"
	3	'e'	'e' in "sent"			'n' in "sent"

	4	's'	's' in "sane"			'a' in "sane"
	4	's'	's' in "sent"			'e' in "sent"

	There is no "parent" member.  Rather, the <child> attribute of a
	NODE will point to one child, and the children will form a singly
	linked list linked by the <sibling> attribute.

Global Variables:

	Next	// Pointer to next available NODE slot
	Lim	// Upper limit of allocated memory for NODEs

Files:
	None

Return Status:

	Returns a pointer to the new NODE.

*******************************************************************************/

static NODE *
newnode(c, llit, depth)
RE_CW *c;
CHAR llit;
int depth;
{
	/* If we need more memory for NODEs, allocate enough for INCR NODEs. */
	if (Next == Lim) {
		Next = (NODE *)cgmalloc2(INCR, sizeof(NODE), "node");
		Lim = Next + INCR;
	}
	/* Store the mapped value of the character to be matched, and the
	 * depth of the NODE.
	 */
	Next->llit = c->cmap[llit];
	Next->depth = depth;

	/*  The following members must be initialized to 0. */
	Next->child = 0;
	Next->matched = 0;
	Next->sibling = 0;
	Next->link = 0;

	/* The <lips>, <shift1>, and <shift2> members will be initialized
	 * in lips() and shift() respectively.
	 */

	/* Give the NODE a unique identifier. */
	Next->id = c->nodeid++;

	/* Return a pointer to the new NODE, and increment <Next> to
	 * point to the next NODE to be allocated.
	 */
	return(Next++);
}

/******************************************************************************

Routine:
	re_cwcomp()

Syntax:
	void re_cwcomp(b)

	RE_CW *c;	// Pointer to Commentz-Walter structure previously
			//	returned by re_cwinit().

Description:

	Compiles the set of strings for use by the Commentz-Walter algorithm.
	The structure must have been created and initialized by re_cwinit(),
	and had strings added to it by re_cwadd().
	re_cwcomp() must be called before either of the text matching routines
	re_cwexec() or re_cbexec() is invoked to check for matches.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
re_cwcomp(c)
RE_CW *c;
{
	int i;

	/* Call lips() and shift() to calculate the step sizes that indicate
	 * how many characters we can advance each time during text matching.
	 */
	lips(c);
	shift(c);
	/* If there is only one pattern (after possible trimming), we can
	 * get a significant performance improvement by not using the NODEs
	 * during text matching.  Rather, we define two simple arrays,
	 * one, <bmpat>, containing the pattern to be matched, and the other,
	 * <shift1>, containing the <shift1> values for the corresponding
	 * NODEs.
	 */
	if (c->npat == 1) {
		/* Get the root NODE. */
		NODE *state = c->root;

		/* Allocate the two arrays. */
		c->bmpat = cgmalloc2(c->mindepth, sizeof(CHAR *), "bmpat");
		c->shift1 = (int *)cgmalloc2(c->mindepth, sizeof(int), "shift1");

		/* Loop through the NODEs. */
		for (i = c->mindepth - 1; i >= 0; i--) {
			/* Note that <shift1[i]> is taken from the parent NODE
			 * of the NODE from which <bmpat[i]> is taken.  That
			 * is because the value of <shift1> in a NODE is the
			 * shift length for a mismatch of the FOLLOWING
			 * character, not for a mismatch of the character
			 * represented by the NODE.
			 */
			c->shift1[i] = state->shift1;
			state = state->child;
			c->bmpat[i] = state->llit;
		}
	}
#ifdef DEBUG
	cwpr(c->root, 1);
	printf("\n");
#endif
}


/******************************************************************************

Routine:
	lips()

Syntax:

	static void lips(c)

	RE_CW *c;	// Pointer to Commentz-Walter structure previously
			//	returned by re_cwinit().

Description:

	lips() sets the <lips> attribute of each NODE to point to the
	"Longest Initial Proper Substring NODE" of that NODE.  It also may remove
	a string from the strings to be matched if a proper substring is also
	to be matched.

	SETTING <lips>:

	For the purposes of this explanation, identify each NODE uniquely by
	the character string specifying the characters that must be matched to
	reach that node.  For example, if we are trying to match the string
	"tables", the node with <llit> 'b' is identified by "bles".  Note that
	the identifier of the root node is the null string.

	An "initial proper substring NODE" of a given NODE is a NODE that is
	identified by an initial proper substring of the string representing
	the NODE.  For example, a NODE is a "initial proper substring NODE"
	of NODE "bles" if and only if the NODE is identified by "ble", "bl",
	"b", or "".  The "longest initial proper substring NODE" of a NODE is
	the NODE represented by the logest initial proper substring. 

	Note that the root is a initial proper substring NODE of every other NODE,
	since the root is identified by the null string.  A NODE other than the
	root is an initial proper substring NODE of a given NODE if and only if:
	
	1) The NODE and the given node have the same <llit> attribute.
	(In the above example, the identifier of all initial proper substring
	must begin with 'b').

	AND

	2) The NODE is a child of an initial proper substring NODE of the parent
	of the given NODE

	lips() uses a queue to process nodes so that all nodes
	of a given depth (i.e. a generation) are processed before any nodes
	of greater depth (i.e., a later generation) are processed.  Simple
	recursive calls cannot do this, because that would simply guarantee
	that the ancestors of a NODE would be processed before the NODE itself.
	It would not guarantee that aunts, great-aunts, etc. have been
	processed.

	REMOVING STRINGS:

	If the full extent of the match need not be known (i.e. c->fullextent
	is not set), a string can be deleted if a substring is also being
	matched.  For example, if "stand" and "tan" are being matched,
	"stand" can be deleted, because any match to "stand" will also match
	"tan".  This situation is detected as follows:  If <p> is the longest
	initial proper substring NODE of NODE <s>, and <p> is a terminal NODE,
	the string beginning with <p> is a substring of the string containing
	<s>.  If the full extent of the match need not be known, the string
	containing <s> can be deleted.  lips() deletes any descendents
	of <s>, and <s> is itself marked for deletion by having <s->matched>
	set to -1.  After lips() is done processing all NODEs, finish_trim()
	is called to complete the removal of the marked strings.  If any
	string removal has occurred, lips() then calls itself to recalculate
	the <lips> attribute of the remaining NODEs.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
lips(c)
RE_CW *c;
{
	NODE *qhead = 0, *qtail = 0;
	NODE *root = c->root;
	register NODE *state, *p, *r, *s;
	register CHAR a;
	BOOL trimmed = NO;

	/* The root node has no longest initial proper substring,
	 * because the null string has no proper substrings.
	 */
	root->lips = 0;

	/* The only substring of each child of the root is the root
	 * itself, because the only proper substring of a one character
	 * string is the null string.
	 *
	 * Add each child of the root to a queue,
	 * and set its <lips> attribute to point to the root.
	 */
	for (s = root->child; s; s = s->sibling) {
		ADD(s);
		s->lips = root;
	}
	/* Loop over all NODEs in the queue.  Eventually, all nodes will
	 * be processed in this loop, because descendent nodes are added
	 * to the queue as encountered.
	 */
	while (qhead) {
		/* Get the first node in the processing queue, and then remove
		 * it from the queue, so we only process it once.
		 */
		r = qhead;
		DEL();
		/* Loop over each child of this node.  */
		for (s = r->child; s; s = s->sibling) {
			/* Add the child <s> to the queue, so it will be
			 * processed later.
			 */
			ADD(s);
			/* Save the literal character of <s>. */
			a = s->llit;

			/* We now set <s->lips> to the longest initial
			 * proper substring of NODE <s>.
			 *
			 * 1) Loop through the initial proper substrings of <r>
			 * (the parent of <s>) longest first.
			 */
			for (state = r->lips; state; state = state->lips) {
				/* If any children of the longest initial proper
				 * substring of <r> have the same <llit> as <s>,
				 * then that child is the longest initial proper
				 * substring of <s>.
				 */
				for (p = state->child; p; p = p->sibling) {
					if (p->llit == a) {
						/* <p> is the longest initial
						 * proper substring of <s>.
						 */
						s->lips = p;
						break;
					}
				}
				/* If <p> represents a match to the pattern,
				 * or is marked for deletion, then the string
				 * containing <s> can be deleted if we do not
				 * need the full extent of the match.  This
				 * is because the string containing <p> is a
				 * proper substring of the string containing <s>.
				 */
				if (p && p->matched && !c->fullextent) {
					/* We can delete any strings containing <s>.
					 * We delete all descendents of <s> now,
					 * and mark <s> itself for later deletion.
					 */
					s->child = 0;
					s->matched = -1;
					trimmed = YES;
				}
				if (p) break;
			}
			/* If we did not yet find the longest initial proper
			 * substring of <s>, the longest must be the root.
			 */
			if (!state) s->lips = root;
		} /* End of loop over the children of NODE <r>. */
	} /* End of loop over all NODES. */

	/* If we marked any NODEs for deletion, delete them now, and then
	 * call lips() again to recalculate the <lips> attributes.
	 * There should be no trimming done on the second pass.
	 */
	if (trimmed) {
		finish_trim(root, (NODE *)0);
		lips(c);
	}
}

/******************************************************************************

Routine:

Syntax:

	static void finish_trim(state, parent)

	NODE *state;	// Pointer to current NODE
	NODE *parent;	// Pointer to the parent NODE of <state>.
			//	0 if <state> is the root node.

Description:

	finish_trim() is called by lips() in the event that any NODES were
	marked for deletion due to removing strings that are superstrings
	of other strings being matched.

	finish_trim() calls itself recursively to delete NODES from the bottom up.
	Any NODE will be deleted (recursively) if it has no children and does
	not represent a match.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
finish_trim(state, parent)
NODE *state;
NODE *parent;
{
	register NODE *p;

	/* We trim recursively from the bottom up. */
	for (p = state->child; p; p = p->sibling) {
		finish_trim(p, state);
	}
	/* Remove <state> if it has no children and does not represent a
	 * match to the pattern.  All NODEs that represent a match to the
	 * pattern have a <matched> attribute of 1.
	 */
	if (state->child == 0 && state->matched <= 0) {
		/* Remove <state>. */
		if (parent->child == state) {
			parent->child = state->sibling;
		} else for (p = parent->child; p; p = p->sibling) {
			if (p->sibling == state) {
				p->sibling = state->sibling;
				break;
			}
		}
	}
}

/******************************************************************************

Routine:
	shift()

Syntax:

	static void shift(c)

	RE_CW *c;	// Pointer to Commentz-Walter structure previously
			//	returned by re_cwinit().

Description:

	shift(), together with its subroutines spurious(), shiftprop(),
	and shifttab() calculates the following quantities:

	1: <step[]> for each character in the character set.
	2: <delta0[]> for each character in the character set.
	3: <tab[]> for each character in the character set.
	4: The number of patterns to be matched (this is recalculated
		in case some were deleted by lips().
	5: <shift1> for each NODE.


	<step[c]> will be the minimum number of characters we can advance after
	a mismatch occurs with text character <c>, independent of the NODE
	at the mismatch.  <step[c]> will simply be one less than the minimum
	depth of any NODE with a literal (<llit>) of <c>, or if there are
	no such NODES, <step[c]> will be the length of the shortest string
	to be matched.  For example, if we are matching the strings "open"
	and "went", the <step> array elements are all 4, except 'n' and 't'
	are 0, 'e' is 1, 'p' is 2, and 'o' and 'w' are 3.

	<delta0[c]> is identical to <step[c]>, except that if an array
	element of <step> has a value of 0, the corresponding array element
	of <delta0> has a value of LARGE.  <delta0[]> is used by the matching
	algorithm to speed up the detection of a character with a <step> of 0.

	<tab[c]> is a pointer to the NODE that will be entered from the
	root NODE if character <c> is encountered.  If <c> is a mismatch,
	<tab[c]> will be 0 rather than a pointer to the root NODE.

	<s->shift1> will be the minimum number of characters that we can
	advance after a mismatch occurs from NODE <s>, independent of the
	mismatched character.  Three rules apply to limit this number.

	Rule 0)
	We can never advance by more than the minimum length of a matched
	string, because we might miss this string entirely.

	Rule 1)
	If <s> is a "non-spurious" initial proper substring NODE of <r>,
	and the depth of <s> is "k" less than the depth of <r>, we cannot
	automatically advance by more than "k" characters after a mismatch
	from NODE <s>.  See spurious() for a definition of "spurious" initial
	proper substring NODE.

	For example, suppose we are searching for the two strings "went" and
	"open".  An initial proper substring of NODE "ent" is NODE "en".  Therefore,
	if we have matched "en", and then have a mismatch (i.e., the previous
	character is not 'p'), we cannot advance the matching to the right by
	more than one step, because if we do, we might miss "went".

	open
	 went

	Rule 2)
	If <s> is an initial proper substring NODE of terminal NODE <r>, and
	the depth of <s> is "k" less than the depth of <r>, we cannot automatically
	advance by more than "k" characters after a mismatch from NODE <s>
	or any of its descendents.

	For example, suppose we are searching for the string "eye".  Suppose
	we have reached node "ye", and then have a mismatch, because the
	previous character is not an 'e'.  Since "ye" is not an initial proper
	substring of any NODE, Rule 1 does not apply.  Therefore, were it not
	for Rule 2, we could advance three characters, and miss a match to the text
	(e.g. the text contains "xyeye").  However, since NODE "e" is an initial
	proper substring NODE of terminal NODE "eye", the advancement from NODE
	"e" is limited to 2 characters by Rule 1.  By rule 2, this limit of 2
	characters also applies to NODE "ye", a descendent of NODE "e", and we will
	not miss the match.

	shift() uses a queue to process nodes so that all nodes
	of a given depth (i.e. a generation) are processed before any nodes
	of greater depth (i.e., a later generation) are processed.  Simple
	recursive calls cannot do this, because that would simply guarantee
	that the ancestors of a NODE would be processed before the NODE itself.
	It would not guarantee that aunts, great-aunts, etc. have been
	processed.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
shift(c)
RE_CW *c;
{
	NODE *qhead = 0, *qtail = 0;
	register NODE *state, *r, *s;
	int k;

	/* Initialize the step size for each character to the
	 * length of the shortst string to be matched.
	 */
	for (k = 0; k < NCHARS; k++) c->step[k] = c->mindepth;

	c->root->shift1 = 1;
	c->root->shift2 = c->mindepth;

	/* Add all children of the root to a queue. */
	for (s = c->root->child; s; s = s->sibling) {
		ADD(s);
	}
	/* Loop over all NODEs in the queue.  Eventually, all nodes will
	 * be processed in this loop, because descendent nodes are added
	 * to the queue as encountered.
	 */
	while (qhead) {
		/* Get the first node in the processing queue, and then remove
		 * it from the queue, so we only process it once.
		 */
		r = qhead;
		DEL();

		/* Initialize the shift parameters to the length of the
		 * shortest string to be matched.
		 */
		r->shift1 = r->shift2 = c->mindepth;

		/* Now loop through the initial proper substring NODES of <r>. */
		state = r;
		while ((state = state->lips) != c->root) {
			/* If <state> is a "spurious" initial proper
			 * substring NODE of <r>, skip it.
			 */
			if (spurious(r, state)) continue;

			/* Get the difference <k> in depth between <r>
			 * and its initial proper substring NODE, <state>.
			 */
			k = r->depth - state->depth;

			/* By Rule 1, <shift1> for this state can't be more
			 * than <k>.  So if <shift1> is greater than <k>,
			 * set <shift1> to <k>.
			 */
			if (k < state->shift1) state->shift1 = k;

			/* By Rule 2, if <r> is a terminal NODE, <shift1>
			 * for this state and any of its descendents can't
			 * be more than <k>.  We use <shift2> to record this
			 * information.  So if <r> is a terminal NODE and
			 * <shift2> is greater than <k>, set <shift2> to <k>.
			 */
			if (r->matched && k < state->shift2) state->shift2 = k;
		}
		/* Now, add the children of <r> to the queue. */
		for (s = r->child; s; s = s->sibling) {
			ADD(s);
		}
	}
#ifdef DEBUG
	cwpr(c->root, 0);
	printf("\n");
#endif
	/*
	 * Call shiftprop() to set the <step> array recursively,
	 * to finish the calculation of <shift1> for each node,
	 * and to recount the number of patterns (some may have been trimmed).
	 */
	c->npat = 0;
	shiftprop(c, c->root);

	/*
	 * As a performance enhancer, we set up a transition table
	 * from the root NODE.  This avoids the need to traverse the chain
	 * of children of the root during text matching.  It is not cost
	 * effective to do this for other nodes.
	 */
	shifttab(c);

	/* We now modify the <step> and <tab> arrays so that they will be
	 * correct for characters that are mapped into other characters.
	 * We also set up the <delta0> array.
	 */
	for (k = 0; k < NCHARS; k++) {
		register int im = c->cmap[k];
		/* Character <k> is mapped into character <im>.  So, the
		 * <step> and <tab> array elements should have the same values
		 * for <k> and <im>.  Only the <im> element is guaranteed to
		 * be set correctly at this point.
		 */
		if (im != k) {
			c->step[k] = c->step[im];
			c->tab[k] = c->tab[im];
		}
		/* The <delta0> array is identical to the <step> array, except
		 * that if an element of <step> is 0, the value of the same
		 * element of <delta0> is LARGE.  This is a computational
		 * trick to speed up the detection of the last character of
		 * a match.
		 */
		c->delta0[k] = (c->step[k] ? c->step[k] : LARGE);
	}
}

/******************************************************************************

Routine:
	shiftprop()

Syntax:

	static void shiftprop(c, n)

	RE_CW *c;	// Pointer to Commentz-Walter structure previously
			//	returned by re_cwinit().
	NODE *n;	// Pointer to the NODE being processed.

Description:

	shiftprop() is called by shift() to aid in the calculations
	of <step>, <shift1>, and the number of patterns to be matched.
	shiftprop() traverses the NODE tree, calling itself recursively.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
shiftprop(c, n)
register RE_CW *c;
register NODE *n;
{
	register NODE *s;

	/* Loop over the children of <n>. */
	for (s = n->child; s; s = s->sibling) {

		/* If <s> is a terminal NODE, increment the pattern count. */
		if (s->matched) c->npat++;

		/* For each character in the character set, <step> will be set
		 * equal to either <mindepth>, or to the depth of the NODE for
		 * that character with the minimum depth, minus one.  So, set
		 * <step> for the character in the (non-root) node to the depth
		 * of the node, minus one, unless <step> is already equal to or
		 * less than that number.
		 */
		if (c->step[s->llit] >= s->depth) {
			c->step[s->llit] = s->depth - 1;
		}

		/* <shift2> is the maximum shift size for a node or any of its
		 * descendents.  So, if <shift2> for the child is less than
		 * for the parent, reset it to the value for the parent.
		 */
		if (s->shift2 > n->shift2) {
			s->shift2 = n->shift2;
		}

		/* <shift2> is the maximum shift size for a node or any of its
		 * descendents.  So, if <shift1> for the node is less than
		 * <shift2>, reset <shift1> to <shift2>.
		 */
		if (s->shift1 > s->shift2) s->shift1 = s->shift2;

		/* Increment <shift1> by the depth of the node.  This is done
		 * so that the shift offset can be taken from the mismatched
		 * character after the node rather than from the end of the
		 * matched text.
		 */
		s->shift1 += s->depth;

		/* Call shiftprop() recursively for each child. */
		shiftprop(c, s);
	}
}

/******************************************************************************

Routine:
	shifttab()

Syntax:

	static void shifttab(c)

	RE_CW *c;	// Pointer to Commentz-Walter structure previously
			//	returned by re_cwinit().

Description:

	shifttab() is called by shift() to calculate <tab>.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
shifttab(c)
register RE_CW *c;
{
	register NODE *s;
	register NODE **nn;

	/* Calculate the transition table from the root node to its children,
	 * based on the text character read.  The table entry will be 0 for
	 * each character that cannot end a match (i.e., does not correspond
	 * to a child of the root NODE).
	 */
	c->tab = nn = (NODE **)cgcalloc(NCHARS, sizeof(NODE *), "shifttab");
	for (s = c->root->child; s; s = s->sibling) {
		nn[s->llit] = s;
	}
}

/******************************************************************************

Routine:
	spurious()

Syntax:

	static BOOL spurious(r, s)

	NODE *r;	// Pointer to a NODE
	NODE *s;	// Pointer to a NODE that is known to be an
			//   "initial proper substring NODE" of NODE <r>.

Description:

	NODE <s> is an "initial proper substring NODE" of node <r> if and only
	if the string of characters that must be matched to reach NODE <s> is
	an initial proper substring of the characters that must be matched
	to reach NODE <r>.  Remember, characters are matched in reverse order,
	i.e., from right to left.

	For example, if we are matching the strings "open" and "went", NODE "n"
	is an initial proper substring NODE of NODE "nt", and node "en" is an
	initial proper substring NODE of NODE "ent".

	If <s> is an initial proper substring NODE of <r>, this typically
	limits the number of characters we can advance in the search if we
	are at NODE <s> and then have a mismatch.  In the above example,
	we are matching "open" and "went".  Suppose we have so far matched an
	'n' and an 'e'.  The only character that can extend the current match
	is "p".  However, suppose that the next character we encounter scanning
	backwards is not an 'p'.  Then we have a mismatch.  However, if the
	character is a 'w', it could be that this "wen" might be part of a
	match to "went".  For this reason, we can advance the terminal character
	by only one position after a mismatch is encountered.

	However, suppose we are matching "open" and "pent", and so far we have
	matched an 'n' and an 'e'.  The only character that can extend the current
	match is "p".  However, suppose that the next character we encounter
	scanning backwards is not an 'p'.  Then we have a mismatch.  However,
	unlike the first example (if the character is a 'w', it could be part of a
	match to "went"), this mismatched character cannot be part of a match
	to "pent", because we know it is not a 'p'!  In this case, we say that
	the node "en" is a "spurious" initial proper substring of the node "ent",
	and we do not restrict the number of characters advanced because of a
	"spurious" initial proper substring.

	More precisely, if <s> is an initial proper substring of <r>,
	<s> is a "non-spurious" initial proper substring of <r> if and only if:

	1) <r> is a terminal NODE (i.e. represents an entire match)

	OR

	2) There exists at least one character that is a mismatch to the left of
	<s> that can be a match to the left of <r>.

	If neither 1) ir 2) holds, <s> is a "spurious" initial proper
	substring of <r>,

	In the case of "open" and "went", "en" is a non-spurious initial proper
	substring of "ent" because 'w' is a mismatch to the left of "en", but is
	a match to the left of "ent".

	In the case of "open" and "pent", "en" is a spurious initial proper
	substring of "ent" because no character is a mismatch to the left of "en",
	but is a match to the left of "ent".

	NOTE:

Global Variables:

	None

Files:
	None

Return Status:

	YES	// <s> is a spurious initial proper substring of <r>.
	NO	// <s> is a non-spurious initial proper substring of <r>.

*******************************************************************************/

static BOOL
spurious(r, s)
register NODE *r;
register NODE *s;
{
	register NODE *rr;
	register NODE *ss;
	register CHAR a;

	/* If <r> is a terminal node, <s> is a non-spurious initial proper
	 * substring of <r>.
	 */
	if (r->matched) return(NO);

	/* Otherwise, <s> is a non-spurious initial proper substring NODE of <r>
	 * if and only if there exists a character that is a mismatch to the
	 * left of <s> that can be a match to the left of <r>.
	 */
	/* Loop over all characters that can match following <r>. */
	for (rr = r->child; rr; rr = rr->sibling) {
		a = rr->llit;
		/* <a> can follow <r>.  If <a> cannot follow <s> on the left,
		 * <s> is a non-spurious failure node of <r>.
		 */
		for (ss = s->child; ss; ss = ss->sibling) {
			if (ss->llit == a) break;
		}
		if (!ss) {
			/* <s> is a non-spurious initial proper substring
			 * NODE of <r>.
			 */
			return(NO);
		}
	}
	/* <s> is a spurious initial proper substring NODE of <r>. */
	return(YES);
}

/******************************************************************************

Routine:
	re_cwexec()

Syntax:

	BOOL re_cwexec(pat, fcn, s, e)

	PATTERN *pat;	// Pointer to pattern to be matched
	BOOL (*fcn)();	// Pointer to subsequent pattern matching function
			//	(0 if Commentz-Walter is the final algorithm used)
	CHAR *s;	// Pointer to first character of text to be searched
	CHAR *e;	// Pointer to last character of text to be searched

Description:

	re_cwexec() is the text matching routine when the Commentz-Walter
	algorithm is being used, and the text is being searched in the forward
	direction.

	re_cwexec() calls grepmatch() for each match that is found, to store
	any match information that needs to be stored.  After a match is found,
	re_cwexec() will continue to search for matches, unless grepmatch()
	indicates by its return value that the search should stop.

Global Variables:

	None

Files:
	None

Return Status:

	YES 	// At least one match was found, and the search was then
		//   terminated because grepmatch() indicated that
		//   it should be terminated.

	NO	// The entire text was searched for matches.   Any number
		//   of matches may have been found.

*******************************************************************************/

BOOL
re_cwexec(pat, fcn, s, e)
PATTERN *pat;
BOOL (*fcn)();
register CHAR *s;
register CHAR *e;
{
	register NODE *state;
	register NODE *p;
	register CHAR *sp;
	CHAR *save_s = s;
	register int j, k;
	CHAR *nb, *ne;
	RE_CW *c = pat->cw_ptr;

	/* 's' now points to the first character of text.
	 * We can start checking the text at the 'n'th character, where
	 * 'n' is the length of the shortest string for which we are searching.
	 */
	s += c->mindepth - 1;
	/* Now look for matches. */
	while (s <= e) {
		/* For each character checked, advance the number of
		 * characters specified in the "delta0" array for that
		 * character, and, unless we go off the end of the text,
		 * check the new character, and repeat the above.
		 * NOTE:  The Commentz-Walter algorithm can spend the bulk
		 * of its time on this single line!  It's optimization
		 * is critical for performance.
		 */
		while ((int)(e - (s = s + c->delta0[*s])) >= 0 ) {
			;
		}
		/* We have stepped past the end of the text.  This
		 * could be for either of two reasons.
		 *
		 * a) We have just checked a character whose "delta0"
		 * array element has been set to LARGE.  This will
		 * be the case for any character that ends a string
		 * for which we are searching.
		 *
		 * OR
		 *
		 * b) We have reached the end of the text without
		 * finding such a character.  In this case, the
		 * last step cannot be larger than c->mindepth.
		 */
		if ((s - e) <= c->mindepth) {
			/* Case b) above.  We have reached the end of the text.
			 * There are no more matches.  Return, indicating we
			 * have searched the entire text.
			 */
			return(NO);
		}
		/*
		 * Case a) above.  We have hit the last character of a string
		 * we are trying to match.  We have at least a partial match.
		 * Decrement 's' to again point to this character.
		 */
		s -= LARGE;

		/* We have matched one character, the last character of a
		 * pattern.  We now must step backwards to see if we match
		 * an entire pattern.
		 */

		/* If we are matching only one pattern, it is faster to
		 * use the <bmpat> and <shift1> arrays, rather than
		 * traversing the NODEs.
		 */
		if (c->npat == 1) {
			/* Step backwards through the text to see if we
			 * match the entire pattern.
			 */
			for (k = c->mindepth - 2, sp = s - 1; k >= 0; k--, sp--) {
				/* Check if the mapped text character
				 * matches the pattern.
				 */
				if (c->cmap[*sp] != c->bmpat[k]) {
					/* No match. */
					break;
				}
			}
			if (k < 0) {
				/* We went past the beginning of the pattern
				 * without finding a mismatch.  We have a
				 * match.  Increment <sp> to point to the
				 * first character of the match, and
				 * process the match.
				 */
				++sp;
				goto matched;
			}
			/* We did not match the entire pattern.
			 * Determine how many characters we can advance from
			 * the mismatched character to resume the search
			 * for matches.  It will be the maximum of the
			 * shift based on which pattern character was
			 * mismatched and the step based on the value
			 * of the text character.
			 */
			j = c->shift1[k];
			k = c->step[*sp];
			if (k < j) {
				k = j;
			}
			/* Advance, and continue searching for new matches. */
			s = sp + k;
			continue;
		}

		/* We are matching more than one string, so we must use the
		 * NODE structures of the Commentz-Walter Algorithm.
		 */

		/* We have matched one character so far. */
		sp = s;
		state = c->tab[*sp];
		if (state->matched) goto matched;

		/* We now step backwards, character by character, to see
		 * if we have a match to an entire string.
		 */
backscan:
		for (--sp; sp >= save_s; --sp) {
			/* Use the linked <child> list to get the next state. */
			for (p = state->child; p ; p = p->sibling) {
				if (c->cmap[*sp] == p->llit) {
					/* Character matches.  Store
					 * the next state.
					 */
					state = p;
					/* Check if this new state is a match. */
					if (state->matched) goto matched;
					break;
				}
			}
			if (p == 0) {
				/* Character does not match. */
				goto nomatch;
			}
		} /* End of for loop scanning text in reverse. */

		/* We have run off the beginning of the text
		 * we are scanning.  Treat as a non-matching
		 * character, and assume c->step[] is
		 * c->mindepth, the maximum possible value.
		 */
		k = c->mindepth;
		goto fake;
nomatch:
		/* We did not match the entire pattern.
		 * Determine how many characters we can advance from
		 * the mismatched character to resume the search
		 * for matches.  It will be the maximum of the
		 * shift based on which pattern character was
		 * mismatched and the step based on the value
		 * of the text character.
		 */
		k = c->step[*sp];
fake:
		if (k < state->shift1) {
			k = state->shift1;
		}
		/* Advance, and continue searching for new matches. */
		s = sp + k;
		continue;
matched:
		/* We found a match.  grepmatch() will do whatever is necessary
		 * to store the match, if indicated.
		 */
		if (grepmatch(pat, fcn, save_s, e, sp, s, &nb, &ne)) {
			/* A return value of YES from getversion indicates we
			 * should stop looking for matches.  This happens only
			 * if we are counting matches and have reached the
			 * maximum, or if we are doing line-by-line matching
			 * because of -v.
			 */
			return(YES);
		}
		/* If there is a longer pattern containing the one we just
		 * matched, we must continue scanning the text backwards.
		 */ 
		if (c->npat > 1 && state->child) goto backscan;

		/* We don't have to keep scanning backwards. */
		if (ne > s) {
			/* grepmatch() has told us we can skip the rest
			 * of the line.  It will only do this if we do
			 * not need to know the extent of the match,
			 * which implies that the match must be confined
			 * to a single line of text.  Therefore, we can start
			 * checking the text at the 'n'th character, where 'n'
			 * is the length of the shortest string that we
			 * are matching.
			 */
			s = ne + c->mindepth - 1;
		} else {
			/* We can't skip to the next line, so skip to the
			 * next character.
			 */
			s++;
		}
	} /* End of while loop scanning text. */
	return(NO);
}

/******************************************************************************

Routine:
	re_cbexec()

Syntax:

	BOOL re_cbexec(pat, fcn, s, e)

	PATTERN *pat;	// Pointer to pattern to be matched
	BOOL (*fcn)();	// Pointer to subsequent pattern matching function
			//	(0 if Commentz-Walter is the final algorithm used)
	CHAR *s;	// Pointer to first character of text to be searched
	CHAR *e;	// Pointer to last character of text to be searched

Description:

	re_cbexec() is the text matching routine when the Commentz-Walter
	algorithm is being used, and the text is being searched in the BACKWARD
	direction.

	re_cbexec() calls grepmatch() for each match that is found, to store
	any match information that needs to be stored.  After a match is found,
	re_cbexec() will continue to search for matches, unless grepmatch()
	indicates by its return value that the search should stop.

	NOTE: re_cbexec() is identical to re_cwexec(), except that the text
	is checked in the OPPOSITE direction as in re_cwexec().  However, the
	strings to be compared to the text are scanned in the SAME direction
	as in re_bmexec().  This is because if the pattern is to be matched
	backwards, the pattern strings were reversed before being passed in to
	re_cwadd(), so they are already reversed.  This is done so that we
	do not need a separate compilation routine for the backwards case
	(re_cwcomp() works for both cases).

Global Variables:

	None

Files:
	None

Return Status:

	YES 	// At least one match was found, and the search was then
		//   terminated because grepmatch() indicated that
		//   it should be terminated.

	NO	// The entire text was searched for matches.   Any number
		//   of matches may have been found.

*******************************************************************************/

BOOL
re_cbexec(pat, fcn, s, e)
PATTERN *pat;
BOOL (*fcn)();
register CHAR *s;
register CHAR *e;
{
	register NODE *state;
	register NODE *p;
	register CHAR *sp;
	CHAR *save_e = e;
	register int j, k;
	CHAR *nb, *ne;
	RE_CW *c = pat->cw_ptr;

	/* 'e' now points to the last character of text.  We can start
	 * checking the text backwards at the 'n'th character, where
	 * 'n' is the length of the shortest string for which we are searching.
	 */
	e -= (c->mindepth - 1);
	/* Now look for matches. */
	while (e >= s) {
		/* For each character checked, back up the number of
		 * characters specified in the "delta0" array for that
		 * character, and, unless we go off the start of the text,
		 * check the new character, and repeat the above.
		 * NOTE:  The Commentz-Walter algorithm can spend the bulk
		 * of its time on this single line!  It's optimization
		 * is critical for performance.
		 */
		while ((int)((e = e - c->delta0[*e]) - s) >= 0 ) {
			;
		}
		/* We have stepped back past the start of the text.
		 * This could be for either of two reasons.
		 *
		 * a) We have just checked a character whose "delta0"
		 * array element has been set to LARGE.  This will
		 * be the case for any character that begins a string
		 * for which we are searching.
		 *
		 * OR
		 *
		 * b) We have reached the start of the text without
		 * finding such a character.  In this case, the
		 * last step cannot be larger than c->mindepth.
		 */
		if ((s - e) <= c->mindepth) {
			/* Case b) above.  We have reached the start of the
			 * text.  There are no more matches.  Return,
			 * indicating we have searched the entire text.
			 */
			return(NO);
		}
		/*
		 * Case a) above.  We have hit the first character of a string
		 * we are trying to match.  We have at least a partial match.
		 * Increment 'e' to again point to this character.
		 */
		e += LARGE;

		/* We have matched one character, the last character of a
		* pattern.  We now must step backwards to see if we match
		* an entire pattern.
		*/

		/* If we are matching only one pattern, it is faster to
		 * use the <bmpat> and <shift1> arrays, rather than
		 * traversing the NODEs.
		 */

		if (c->npat == 1) {
			/* Step forwards through the text to see if we
			 * match the entire pattern.
			 */
			for (k = c->mindepth - 2, sp = e + 1; k >= 0; k--, sp++) {
				/* Check if the mapped text character
				 * matches the pattern.
				 */
				if (c->cmap[*sp] != c->bmpat[k]) {
					/* No match. */
					break;
				}
			}
			if (k < 0) {
				/* We went past the end of the pattern
				 * without finding a mismatch.  We have a
				 * match.  Decrement <sp> to point to the
				 * last character of the match, and
				 * process the match.
				 */
				--sp;
				goto matched;
			}
			/* We did not match the entire pattern.
			 * Determine how many characters we can advance from
			 * the mismatched character to resume the search
			 * for matches.  It will be the maximum of the
			 * shift based on which pattern character was
			 * mismatched and the step based on the value
			 * of the text character.
			 */
			j = c->shift1[k];
			k = c->step[*sp];
			if (k < j) {
				k = j;
			}
			/* Advance backwards , and continue searching for new matches. */
			e = sp - k;
			continue;
		}

		/* We are matching more than one string, so we must use the
		 * NODE structures of the Commentz-Walter Algorithm.
		 */

		/* We have matched one character so far. */
		sp = e;
		state = c->tab[*sp];
		if (state->matched) goto matched;

		/* We now step backwards, character by character, to see
		 * if we have a match to an entire string.
		 */
backscan:
		for (++sp; sp <= save_e; ++sp) {
			for (p = state->child; p; p = p->sibling) {
				if (c->cmap[*sp] == p->llit) {
					/* Character matches.  Store the next state. */
					state = p;
					/* Check if this new state is a match. */
					if (state->matched) goto matched;
					break;
				}
			}
			if (p == 0) {
				/* Character does not match. */
				goto nomatch;
			}
		} /* End of for loop scanning text in reverse. */

		/* We have run off the end of the text
		 * we are scanning.  Treat as a non-matching
		 * character, and assume c->step[] is
		 * c->mindepth, the maximum possible value.
		 */
		k = c->mindepth;
		goto fake;
nomatch:
		/* We did not match the entire pattern.
		 * Determine how many characters we can advance from
		 * the mismatched character to resume the search
		 * for matches.  It will be the maximum of the
		 * shift based on which pattern character was
		 * mismatched and the step based on the value
		 * of the text character.
		 */
		k = c->step[*sp];
fake:
		if (k < state->shift1) {
			k = state->shift1;
		}
		/* Advance backwards, and continue searching for new matches. */
		e = sp - k;
		continue;
matched:
		/* We found a match.  grepmatch() will do whatever is necessary
		 * to store the match, if indicated.
		 */
		if (grepmatch(pat, fcn, s, save_e, e, sp, &nb, &ne)) {
			/* A return value of YES from getversion indicates we
			 * should stop looking for matches.  This happens only
			 * if we are counting matches and have reached the
			 * maximum, or if we are doing line-by-line matching
			 * because of -v.
			 */
			return(YES);
		}
		/* If there is a longer pattern containing the one we just
		 * matched, we must continue scanning the text backwards.
		 */ 
		if (c->npat > 1 && state->child) goto backscan;

		/* We don't have to keep scanning backwards. */
		if (nb < e) {
			/* grepmatch() has told us we can skip the rest
			 * of the line.  It will only do this if we do
			 * not need to know the extent of the match,
			 * which implies that the match must be confined
			 * to a single line of text.  Therefore, we can start
			 * checking the text at the 'n'th character, where 'n'
			 * is the length of the shortest string that we
			 * are matching.
			 */
			e = nb - (c->mindepth - 1);
		} else {
			e--;
		}
	} /* End of while loop scanning text. */
	return(NO);
}

/******************************************************************************

Routine:
	re_cwfreenodes()

Syntax:

	void re_cwfreenodes(r)

	NODE *r;	// Pointer to root of tree of NODEs to be freed.

Description:

	re_cwfreenodes() frees the memory for the previously allocated NODE
	structures for a pattern.  When called externally, the argument should be
	the root NODE for the pattern.

	Since the memory for the NODEs is allocated in blocks of INCR NODEs
	at a time, the actual address of each allocation is a pointer to a
	NODE whose <id> attribute is an integral multiple of INCR, i.e.,
	0, INCR, 2*INCR, etc.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
re_cwfreenodes(r)
register NODE *r;
{
	register NODE *s;

	for (s = r->child; s; s = s->sibling) {
		re_cwfreenodes(s);
	}
	if ((r->id % INCR) == 0) FREE(r);
}


/****************************************************************************/
#ifdef	DEBUG
static int
cwpr(n, k)
register NODE *n;
int k;
{
	register NODE *p;

	printf("NODE: %d[lit: %c, depth: %d, matched: %d, sh1: %d, sh2: %d substring: %d]:\n", n->id, n->llit, n->depth, n->matched, n->shift1-k*n->depth, n->shift2, (n->lips ? n->lips->id : -1));
	for (p = n->child; p; p = p->sibling) {
		printf("edge from \"%d\" to \"%d\" label {\"%c\"};",
			n->id, p->id, p->llit);
		printf("\n");
		cwpr(p, k);
	}
}
#endif
