	static char sccsid[]="%Z%	recomp.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	recomp.c [%P%]

Description:

	Contains the routines needed to compile a regular expression,
	plus optimization routines.

Functions:

	// COMPILATION ROUTINES

	re_recomp()
	eptr()
	newexpr()
	lex()
	ccl()
	e3()
	e2()
	e1()
	e0()
	eall()
	err()
	fill_expr()

	// OPTIMIZATION ROUTINES

	optimize_exp()
	char_compress()
	trim()
	trim_nl()
	trim_null()
	interior_nl()
	adj_compress()
	adj_lit_compress()
	compress_unlike()
	compress_like()
	lit_contains()
	ccl_compare()
	matches_every_line()
	free_expr()
	delete_expr()
	change_rept()
	delete_rept()
	rept_compress()
	percolate()
	get_next()
	null_alter()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

#define PlusStar(x) ((x) == Plus || (x) == Star)
#define PlusStarQuest(x) (PlusStar(x) || (x) == Quest)
#define StarQuest(x) ((x) == Star || (x) == Quest)

static EXPR_TYPE Toktype;
static int Left_parno;
static unsigned short Left_brackets, Right_brackets;
static CHAR *Toklit, *Ptr, *End;
static EXPR *Groups[10];
static int Group_refs[10];
static CHAR non_alpha[] = "^0-9A-Za-z_\n]\0";
static CHAR *Non_alpha_ccl;
static int Max_deletions;

static EXPR *eall(), *e0(), *e1(), *e2(), *e3();
static EXPR *get_next();
static void eptr();
static void lex(), fill_expr();
static void optimize_exp(), free_expr(), percolate();
static void delete_expr(), delete_rept(), change_rept(), rept_compress();
static void trim_nl(), trim_null(), null_alter();
static void adj_compress(), adj_lit_compress(), char_compress();
static BOOL compress_unlike(), compress_like();
static BOOL matches_every_line(), trim(), lit_contains(), ccl_compare();
static int ccl();
#ifdef DEBUG
static RE_RE *R;
static void print_expr();
#endif

#define MANY 18
void err();

/******************************************************************************

Routine:
	re_recomp()

Syntax:

	RE_RE *re_recomp(pat)

	PATTERN *pat;	// Pointer to pattern to be matched. 

Description:

	The main routine for the compilation of a regular expression for
	use in matching text.

	re_recomp() and re_reinit() must be called (in that order)
	before either of the text matching routines re_reexec() or re_rbexec()
	is invoked to check for matches.

	If backreferencing is being used, re_recomp() and re_brinit() must be
	called (in that order) before the text matching routine re_brexec()
	is invoked to check for matches.

	The compilation done by re_recomp() and its subroutines is done in
	four stages:

	1) INITIALIZATION:  Various quantities in the regular expression
		structure are initialized.

	2) COMPILATION PROPER:  The expression tree for the compiled
		regular expression is created.

	3) OPTIMIZATION:  For performance reasons, the expression tree
		is "optimized".

	4) POST-COMPILATION:  Various quantities in the regular expression
		structure that depend on the contents of the expression tree
		are initialized.

	A fifth stage, PRE-MATCHING, is also required before text matching
	can begin.  PRE-MATCHING is not done by the call to re_recomp(),
	but is accomplished by a call to re_reinit() (or, if backreferencing,
	by a call to re_brinit()).

Global Variables:

	Max_deletions	// Maximum number of deletions allowed.
			//   Set by re_recomp() for use in
			//   compilation/optimization

Files:
	None

Return Status:

	Returns a pointer to the compiled regular expression.

*******************************************************************************/

RE_RE *
re_recomp(pat)
PATTERN *pat;
{
	register RE_RE *r;

	/* STAGE 1: INITIALIZATION */

	/* Allocate space for the compiled regular expression, and initialize
	 * everything to zero.
	 */
	r = (RE_RE *)cgmalloc(sizeof(RE_RE), "allocating RE_RE");
	(void)memset((char *)r, 0, (int)sizeof(RE_RE));

#ifdef DEBUG
	R = r;
#endif
	/* The character map is initialized from the map for the
	 * pattern.  It may be changed during the optimization due to
	 * character compression.
	 */
	r->cmap = pat->cmap;

	/* The approximate matching costs for the pattern are duplicated into
	 * the regular expression for easier use during compilation and matching.
	 */
	r->cost_limit = pat->cost_limit;
	r->cost_sub = pat->cost_sub;
	r->cost_del = pat->cost_del;
	r->cost_ins = pat->cost_ins;

	/* The backward matching indicator and the backreferencing indicator are
	 * initialized from the values for the pattern.  These may be changed
	 * after the compilation.
	 */
	r->backward = pat->backward;
	r->backref = pat->backref;

	/* Initialize the unique identifier for each "lit" expression to 1,
	 * because an identifier of 0 is reserved for non-lit expressions.
	 */
	r->maxid = 1;

	/* Anchoring is normally off. */
	r->anchored = NO;

	/* Store the maximum number of deletions allowed as a global.  This is
	 * used during compilation and optimization.
	 */
	Max_deletions = pat->cost_limit/pat->cost_del;

	/* STAGE 2: COMPILATION PROPER */

	/* Now, do the compilation.  eall() does the compilation, creating
	 * the entire expression tree, and returning the root.
	 */
	r->root = eall(pat, r);

	/* STAGE 3: OPTIMIZATION */

	/* Now, we do various optimizations to speed performance. */
	optimize_exp(pat, r);

	/* STAGE 4: POST-COMPILATION */

	/* Indicate whether we really are backreferencing. */
	pat->backref = r->backref = r->root->backref;

	/* Optimization may have changed the character map. */
	pat->cmap = r->cmap;

	/* Set the minimum length of text that can match the regular expression. */
	r->min_length = r->root->min_length;

	/* Set the maximum length of text that can exactly match
	 * the regular expression.
	 */
	r->max_length = r->root->max_length_ex;

	/* Set the maximum number of lines, minus one, that an exact match
	 * to the expression can span.  This is usually equal
	 * to the maximum new-line count in any one expression, not
	 * including an initial or final new-line. 
	 */
	pat->mult_lines = r->root->nl_count - r->root->nl_first - r->root->nl_last;
	/* There is one case where our count of spanned lines is incorrect.
	 * If the expression with the maximum new-line count has both an
	 * initial and final new-line, and an alternate expression has one
	 * fewer new-line, but neither an initial or a terminal new line,
	 * we can span one more line.  The "nl_alt" flag will have been set
	 * in this case.
	 */
	if (r->root->nl_alt && r->root->nl_first && r->root->nl_last) {
		pat->mult_lines++; 
	}

	/* If we are approximate matching, "max_length" and "mult_lines" must
	 * be adjusted to allow for approximate matches.
	 */
	if (r->cost_limit || !r->cost_ins) {
		/* Get the lesser of the substitution cost or the insertion
		 * cost, but not less than 1, since it always costs at least
		 * 1 unit to insert a new line.
		 */
		int tmp = MIN(r->cost_sub, r->cost_ins);
		if (!tmp) tmp = 1;

		/* The number of spanned lines must be incremented by the
		 * number of allowed insertions or substitutions.
		 */
		pat->mult_lines += (r->cost_limit/tmp); 

		/* The maximum length of a match must be incremented by the
		 * number of allowed insertions, but if the insertion cost is
		 * zero, there is no maximum length.  We represent no limit by -1.
		 */
		if (!r->cost_ins) r->max_length = -1;
		else r->max_length += (r->cost_limit/r->cost_ins);
	}

	/* If all matches to the expression are the same length, set r->length
	 * to this length.  Otherwise, set it to zero.
	 */
	r->length = (r->min_length == r->max_length ? r->min_length : 0);

	/* Backreferencing is not reversible! */
	if (r->backref) r->backward = NO;

	/* Allocate space for the array of "lit" expressions that are
	 * part of the regular expression.
	 */
	r->ptr = (EXPR **)cgmalloc2(r->maxid, sizeof(EXPR *), "ptr");

	/* Fill in the "r->ptr" array with all the "lit" expressions
	 * in the regular expression.
	 */
	eptr(r, r->root);
	return(r);
}

/******************************************************************************

Routine:
	eptr()

Syntax:

	static void eptr(r, e)

	RE_RE *r;	// Pointer to regular expression being compiled

	EXPR *e;	// Pointer to an expression in the expression tree
			//    When called externally, <e> should be the
			//    root expression (i.e. r->root).

Description:

	eptr() calls itself recursively to store a pointer to each "lit" expression
	in the expression tree.   The pointers are stored in the r->ptr[] array.

	eptr() is called by re_recomp() during the POST_COMPILATION.
	The r->ptr[] array filled in by this call is used by the text
	matching routines.

	eptr() may also be called earlier during the OPTIMIZATION. 
	In these cases, the r->ptr[] array is used temporarily during
	the optimization, and is then freed.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
eptr(r, e)
register RE_RE *r;
register EXPR *e;
{
	/* Make sure that the ID of the expression is in range. */
	if ((e->id < 0) || (e->id >= r->maxid)) {
		exitmsg2("%s: eptr abort %d %d\n", e->id, r->maxid);
	}
	/* If "e" is a "lit", store a pointer to "e" in the array element
	 * corresponding to the ID.
	 */
	if (e->id) r->ptr[e->id] = e;

	/* Now, process the children recursively. */
	switch(e->type) {
	case Alternate: case Cat:
		eptr(r, e->l);
		eptr(r, e->r);
		break;
	case EOP: case Group: case Plus: case Star: case Quest:
		eptr(r, e->l);
		break;
	}
}

/******************************************************************************

Routine:
	newexpr()

Syntax:

	EXPR *newexpr(r, t, lit, left, right)

	RE_RE *r;	// Pointer to regular expression being compiled

	EXPR_TYPE t;	// Value to be stored in <newexpr>->type.
			// Type of expression to be created.

	CHAR lit;	// Value to be stored in <newexpr>->lit.
			// "lit" value of the expression.

	EXPR *left;	// Value to be stored in <newexpr>->left.
			// A pointer to the left (i.e. first) child of the
			//    expression, or null if the expression does
			//    not have children.
			// EXCEPTION: Backref type uses <left> for a pointer
			// to the associated Group.

	EXPR *right;	// Value to be stored in <newexpr>->right.
			// A pointer to the right (i.e. second) child of the
			//    expression, or null if the expression does not
			//    have two children.
			// EXCEPTION: Charclass type uses <right> for a pointer
			// to the character class table (a bit mask).

Description:

	newexpr() is called by various routines during COMPILATION PROPER
	to create a new expression for insertion into the expression tree.
	The tree is created from the leaves upward, so that children are
	created before the parent, and the root node is the last to be created.

	All children of the new expression, if any, MUST have previously
	been created.

	newexpr() may also be called during OPTIMIZATION if the optimization
	requires a new expression.  In this case, the parent may already exist,
	and it is up to the caller to set e->parent.

	If backreferencing is used, newexpr() is also called during the
	PRE-MATCHING phase, by re_brinit() or its subroutines.

Global Variables:

	Left_brackets	// Count of left angle brackets
	Right_brackets	// Count of right angle brackets

Files:
	None

Return Status:

	Returns a pointer to the newly allocated expression.

*******************************************************************************/

EXPR *
newexpr(r, t, lit, left, right)
RE_RE *r;
EXPR_TYPE t;
CHAR lit;
EXPR *left, *right;
{
	/* Allocate memory for the expression. */
	register EXPR *e = (EXPR *)cgmalloc(sizeof(EXPR), "newexpr");

	/* Fill in parameters from the arguments. */
	e->type = t;
	e->lit = lit;
	e->l = left;
	e->r = right;

	/* Expressions are normally allocated from the bottom up, so this
	 * expression probably has no parent yet.  If the parent already
	 * exists, it is up to the calling routine to set e->parent.
	 */
	e->parent = 0;

	/* The sets of possible preceding and following lit expressions
	 * will not be filled in until needed in matching.  Set the pointers
	 * to 0, indicating not allocated, and set the counts to 0 also.
	 */
	e->b.follow = e->f.follow = 0;
	e->b.flen = e->f.flen = 0;

	/* Set the angle bracket number (used for approximate matching)
	 * for the expression.  The angle bracket number is used in the
	 * matching algorithms only for atomic "lit" expressions and for
	 * Plus or Star.  In other cases, we set the bracket numbers
	 * deterministically to facilitate optimization.
	 * We first set the bracket number as if the expression were a "lit".
	 */
	if (r->cost_limit == 0 && r->cost_ins > 0) {
		/* We are exact matching.  Set all bracket numbers to 1. */
		e->n_bracket = 1;
	} else if (Left_brackets == Right_brackets) {
		/* We are approximate matching, and we have counted as many right
		 * brackets as left brackets. We are not inside angle brackets.
		 */
		e->n_bracket = 0;
	} else if (Left_brackets > Right_brackets) {
		/* We are approximate matching, and we have counted more left
		 * brackets than right brackets. We are inside angle brackets.
		 */
		if (r->cost_ins <= r->cost_limit) {
			/* Insertions are permitted.  Therefore, the bracket
			 * number is important.  Store the number of the
			 * left bracket.
			 */
			e->n_bracket = Left_brackets;
		} else {
			/* Insertions are not permitted.  Therefore, the bracket
			 * numbers are all effectively the same.
			 * Set all bracket numbers to 1 for those inside angle
			 * brackets.
			 */
			e->n_bracket = 1;
		}
	} else {
		/* We have counted more right brackets than left.
		 * This is an internal error, because we should have caught
		 * this earlier.
		 */
		exitmsg0("%s: INTERNAL ERROR: angle bracket imbalance\n");
	}

	/*
	 * For non-"lit" types, we may change the bracket number.
	 */
	switch (t) {
	case Alternate: case Cat:
		/* The bracket number of these composite objects has no
		 * significance in the matching algorithm.  For definitiveness,
		 * we set it to the bracket number of its children, if the
		 * same, or zero, otherwise.
		 */
		if (e->l->n_bracket == e->r->n_bracket) {
			e->n_bracket = e->l->n_bracket;
		} else {
			e->n_bracket = 0;
		}
		break;
	case Quest: case Group: case Backref: case EOP:
		/* The bracket number of these objects has no significance
		 * in the matching algorithm.  For definitiveness,
		 * we set it to the bracket number of its child.
		 *
		 * Quest Example:
		 * '<ab>?<cd>', '<ab><?><cd>', '<ab><?cd>', and '<ab?><cd>'
		 * are all equivalent, so we store it as '<ab?><cd>'
		 */
		e->n_bracket = e->l->n_bracket;
		break;
	case Plus: case Star:
		/* The bracket number of these objects has significance
		 * in the matching algorithm only if it is in the same bracket
		 * as its child.
		 * For definitiveness, we set the bracket number to zero
		 * if it is not in the same bracket as its child.
		 *
		 * Example:
		 *
		 * cgrep -E -A1 '<ab+><cd>' does not match abbxbbbbcd
		 *    (insertions not allowed in the chain of b's)
		 *    ('+' has same bracket number as 'b').
		 *
		 * while
		 * cgrep -E -A1 '<ab>+<cd>' matches abbxbbbbcd
		 *    (insertions allowed in the chain of b's, but still
		 *    not allowed between the 'a' and the first 'b'.)
		 *    ('+' has bracket number 0).
		 *
		 * '<ab><+><cd>' and '<ab><+cd>' are treated like '<ab>+<cd>'.
		 *    ('+' has bracket number 0).
		 */
		if (e->n_bracket != e->l->n_bracket) e->n_bracket = 0;
		break;
	}


	/* For "lit"s, set the expression ID to a unique
	 * non-zero identifier, and increment the count.
	 * For non-lit's set it to zero.
	 */
	switch (t) {
	case Dot: case Charclass: case Literal: case EOP:
		e->id = r->maxid++;
		break;
	default:
		e->id = 0;
		break;
	}

	/* If this expression has children, set the parent of each child to
	 * this expression.
	 */
	switch (t) {
	case Alternate: case Cat:
		left->parent = e;
		right->parent = e;
		break;
	case EOP: case Group: case Plus: case Star: case Quest:
		left->parent = e;
		break;
	}

	/* Calculate e->backref_no_star, e->backref, e->group,
	 *   e->rept, e->alter. 
	 *
	 * Calculate e->min_length, e->min_length_ex, and e->max_length_ex.
	 *
	 * Calculate e->dot, e->dot_star_l, and e->dot_star_r.
	 *
	 * Calculate e->nl_count, e->nl_first, e->nl_last, and e->nl_alt.
	 */

	fill_expr(e);

	/* If this expression is a Plus, Star, or Quest, and its child
	 * is also a Plus, Star, or Quest (not necessarily the same as the
	 * parent), the child expression can be collapsed into this one.
	 * This is an optimization.
	 */
	rept_compress(e, e->l);

	/* Return a pointer to the new expression. */
	return(e);
}

/******************************************************************************

Routine:
	lex()

Syntax:

	static void *lex(r)

	RE_RE *r;	// Pointer to regular expression being compiled

Description:

	lex() is called by eall(), e0(), e2(), and e3() during the
	COMPILATION PROPER to parse the next token in the regular expression.

	Usually, lex() processes one token and returns.  However, if the
	token is an angle bracket ('<' or '>'), lex() calls itself to
	also process the following token before returning.

	lex() uses <Ptr> to keep track of the position of the current token,
	and uses <Toktype> and <Toklit> to return information to the caller.
	See "Global Variables" in this prologue.

Global Variables:

	Ptr		// Pointer to current token
	End		// Pointer to end of uncompiled regular expression
	Left_brackets	// Count of left angle brackets
	Right_brackets	// Count of right angle brackets
	Toktype		// Returns the type of token
	Toklit		// Returns any value associated with the token

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
lex(r)
RE_RE *r;
{
	if (Ptr == End) {
		/* We have reached the end of the pattern. Set Toktype
		 * to EOP (End Of Pattern), and Toklit to 0.
		 */
		Toktype = EOP;
		Toklit = 0;
	} else switch(Toklit = Ptr++, *Toklit) {

	/* We haven't reached the end of the pattern.  Toklit points
	 * to the next character in the pattern.  We switched on the
	 * value of this character, and check for every possible
	 * metacharacter.
	 */

	/* '`' metacharacter.  Set Toktype to Backref, and
	 * Toklit to point to the following character, which should
	 * be a digit from '1' to '9', indicating the number of
	 * the corresponding group.
	 */
	case '`':	Toktype = Backref; Toklit = Ptr++; break;

	/* '#', '.', '*', '+', '?', '[', '(', ')' or new-line
	 * metacharacter.  Set Toktype to the corresponding value.
	 * Toklit will remain pointing to the metacharacter.
	 */
	case '#':	Toktype = Non_alpha; break;
	case '.':	Toktype = Dot; break;
	case '*':	Toktype = Star; break;
	case '+':	Toktype = Plus; break;
	case '?':	Toktype = Quest; break;
	case '[':	Toktype = Charclass; break;
	case '(':	Toktype = Lpar; break;
	case ')':	Toktype = Rpar; break;
	case '\n':	Toktype = Alternate; break;

	/* '<' metacharacter.  Increment the count of left angle brackets.
	 * Call lex() to parse the next token.
	 */
	case '<':	Left_brackets++; lex(r); break;

	/* '>' metacharacter.  Increment the count of right angle brackets.
	 * Call lex() to parse the next token.
	 */
	case '>':	Right_brackets++; lex(r); break;

	/* Backslash metacharacter.  It's an error if it is the last
	 * character of the pattern.  Otherwise, skip the backslash
	 * and treat the following character as a literal, regardless
	 * of what it is.
	 */
	case '\\':	if (Ptr == End) err("bad \\");
			Toktype = Literal; Toklit = &r->cmap[*Ptr++]; break;

	/* A literal character.  Set Toktype to Literal, and Toklit to
	 * point to the mapped value of the literal character.
	 */
	default:	Toktype = Literal; Toklit = &r->cmap[*Toklit]; break;
	}
}

/******************************************************************************

Routine:
	ccl()

Syntax:

	static int ccl(cmap, tab)

	CHAR *cmap;	// Pointer to character map for the regular expression
	CHAR *tab;	// Pointer to character class table to be filed in
			//    Memory must have been allocated by the caller.

Description:

	ccl() is called by e3() during the COMPILATION PROPER to parse a character
	class in the regular expression.  ccl() is also called by eall() during
	the COMPILATION PROPER to parse the character class used for matching
	non-alphanumeric characters for delimiting words.

	ccl() parses the character class until the ']' that terminates the
	character class specification is found, or the end of the regular expression
	is found.

	The character class table is a one dimensional CHAR array of size CCL_SIZ+1.

	If the character class contains two or more characters, ccl() fills in
	the first CCL_SIZ characters with a bit mask, with the appropriate
	bit set for each character in the class.  The last character, tab[CCL_SIZ],
	is set to the count of characters in the set (an 8-bit integer, not
	an ASCII digit).

	If the character class contains only one character, ccl() fills in tab[0]
	with the one ASCII character that comprises the class.  The rest of tab[]
	is undefined.

	In either case, the count of characters in the class is returned.

Global Variables:

	Ptr		// Pointer to current token
	End		// Pointer to end of uncompiled regular expression
	bittab		// Mask bits used internally by CCL_SET(), CCL_CLR(),
			//   and CCL_CHK() macros.

Files:
	None

Return Status:

	Returns the number of characters in the character class

*******************************************************************************/

static int
ccl(cmap, tab)
CHAR *cmap;
CHAR *tab;
{
	register int i;
	register CHAR c; 
	int range = 0;
	int lastc = -1;
	int count = 0;
	BOOL comp = NO;

	/* Zero the character class table. */
	(void)memset((char *)tab, 0, (int)((CCL_SIZ+1)*sizeof(CHAR)));

	/* If the first character inside the class is '^', we have a
	 * "complementary" character class.  Set "comp", and advance
	 * to the next character.
	 */
	if (*Ptr == '^') {
		Ptr++;
		comp = YES;
	}
	/* If the next character inside the class is a ']' or a '-',
	 * treat it as an ordinary character, without its special
	 * meaning.  Add the mapped character to the class, set
	 * "lastc" to point to this most recent character so far, and
	 * advance to the next character.
	 */
	if (*Ptr == ']' || *Ptr == '-') {
		c = cmap[*Ptr];
		CCL_SET(tab, c);
		lastc = *Ptr++;
	}
	/* We now loop through each character inside the class,
	 * until we come to the terminal ']' or the end of the pattern.
	 */
	for (; (Ptr < End) && (*Ptr != ']'); Ptr++) {

		if (!range && *Ptr == '-') {
			/* We have a '-', and we were not already in
			 * a range.  Assume that it indicates a range.
			 */
			range = 1;
			continue;
		}
		if (*Ptr == '\\') {
			/* We have a backslash.  In our internal format
			 * this is a metacharacter even inside a regular
			 * expression.  Skip the backslash, and the
			 * next character will be treated as an ordinary
			 * character, even if it is a '-' or ']'.
			 */
			if (++Ptr >= End) {
				/* It's an error if the backslash was the last
				 * character in the pattern.
				 */
				err(" [ ] imbalance");
			}
		}
		/* Add this character to the class. */
		c = cmap[*Ptr];
		CCL_SET(tab, c);

		if (range) {
			/* We have the final character of a range.
			 * Add every character in the interior of the range
			 * into the class, mapping it first.
			 * Then indicate we are no longer in a range.
			 * Note that *Ptr and lastc, the endpoints, have
			 * already been added to the class.
			 * Note also that if the first ASCII character of the range
			 * is greater than the last ASCII character, only the
			 * endpoint characters get added to the class.
			 * This is consistent with "grep" and "egrep".
			 */
			for (i = (*Ptr) - 1; i > lastc; i--) {
				CCL_SET(tab, cmap[i]);
			}
			range = 0;
		}
		/* Save the most recent character added. */
		lastc = *Ptr;
	}
	/* If we hit the end of the pattern without a closing ']', it is an error. */
	if (Ptr == End) err(" [ ] imbalance");

	/* If "range" is set, the last character before the closing ']' was a '-'.
	 * This is treated as a literal '-'.
	 */
	if (range) {
		c = cmap['-'];
		CCL_SET(tab, c);
		range = 0;
	}

	/* Advance beyond the closing ']', for subsequent parsing. */
	Ptr++;

	if (comp) {
		/* This is a complementary character class.  Turn the
		 * set of ASCII characters into its complement.
		 */
		for (i = 0; i < CCL_SIZ; i++) tab[i] ^= 0xff;

		/* The new-line character is a special case.  Since a
		 * complementary character class in "grep" or "egrep"
		 * matches every character not in the set, except new-line,
		 * we invert the status of the new-line character 
		 * to maintain compatibility.
		 */
		if (CCL_CHK(tab, cmap[NL])) {
			CCL_CLR(tab, cmap[NL]);
		} else {
			CCL_SET(tab, cmap[NL]);
		}
	}

	/* Loop through the character set. */
	for (i = 0; i < NCHARS; i++) {
		/* If the character does not map to itself, we can remove it
		 * from the character class if it is in it.
		 */
		if (cmap[i] != i) CCL_CLR(tab, i);

		/* If the character is still in the class, set "lastc" to
		 * it, and increment the character count.
		 */
		if (CCL_CHK(tab, i)) {
			lastc = i;
			count++;
		}
	}
	/* If there is only one character in the charclass, the caller
	 * will convert the Charclass to a Literal.  Copy the character
	 * into "tab".
	 * If there is more than one character, store the character
	 * count in the extra element at the end of the table.
	 */
	if (count == 1) *tab = (char)lastc;
	else tab[CCL_SIZ] = (CHAR)count;

	/* Return the character count. */
	return(count);
}

/******************************************************************************

Routine:
	e3()

Syntax:

	static EXPR *e3(r)

	RE_RE *r;	// Pointer to regular expression being compiled

Description:

	Grammar:

	EOP:		eall:   { e0 } EOP
	Alternation:	e0:	e1 { '|' e1 }*
	Concatenation:	e1:	e2 { e2 }*
	Repetition:	e2:	e3 { '+' | '*' | '?' }
	Literal:	e3:	Literal | '.' | '[]' | Backref | '(' e0 ')'

	e3() is called by e2() during the COMPILATION PROPER to process a
	single Literal, Charclass ('['), Non_alpha ('#'), Dot ('.'),
	Backref ('`'), or Lpar ('(') token.

	If Toktype is not one of the above types, it is a fatal error.

	For all of the legal token types, except Lpar, a single new expression
	of the corresponding type is created (with a call to newexpr()),
	except that a Charclass is used for a Non_alpha token, and a Charclass
	with only one member is converted to a Literal expression.

	For an Lpar token, lex() is called to parse the next token, and
	e0() is then called to process all the tokens up to but not including
	the corresponding Rpar (')') token.  e0() returns a pointer to the last
	expression that it allocated.  Finally, if and only if backreferencing
	is being used, an expression of type Group is created to delimit
	the expressions created by e0().

	e3() then calls lex() to parse the next token, and returns a pointer
	to the last new expression allocated.

Global Variables:

	Toktype		// Token type of next token
	Toklit		// Value associated with token.
	Groups		// Pointers to backreferencing groups
	Group_refs	// Reference counts to backreferencing groups

Files:
	None

Return Status:

	Returns a pointer to the last expression allocated.

*******************************************************************************/

static EXPR *
e3(r)
RE_RE *r;
{
	EXPR *e;
	CHAR *tab;
	int count;

	switch(Toktype) {
	case Backref:
		/* For a Backref, "lit" will be the digit from 1-9 that
		 * identifies the group.  "l" will point to the expression
		 * for the corresponding Group.
		 */
		e = newexpr(r, Backref, *Toklit, Groups[*Toklit - '0'], (EXPR *)0);
		/* Increment the reference count for the corresponding group. */
		Group_refs[*Toklit - '0']++;
		break;
	case Literal:
		/* For a Literal, "lit" will be the literal character itself.  */
		e = newexpr(r, Literal, *Toklit, (EXPR *)0, (EXPR *)0);
		break;
	case Non_alpha:
		/* For a Charclass, "lit" will be '[' and "r" will point
		 * to the table.  In the case of the word delimiter class,
		 * the class table is already compiled, and can be shared.
		 */
		e = newexpr(r, Charclass, '[', (EXPR *)0, (EXPR *)Non_alpha_ccl);
		break;
	case Dot:
		/* For a Dot, "lit" will be '.'.  */
		e = newexpr(r, Dot, '.', (EXPR *)0, (EXPR *)0);
		break;
	case Charclass:
		/* Allocate space for the class table, and compile the table.  */
		tab = (CHAR *)cgmalloc2(CCL_SIZ+1, sizeof(CHAR), "charclass defn");
		count = ccl(r->cmap, tab);
		if (count == 1) {
			/* If the Charclass has only one member, we convert it
			 * into a Literal.  In this case, ccl() has copied
			 * the literal character into the first byte of the
			 * Charclass table.
			 */
			e = newexpr(r, Literal, *tab, (EXPR *)0, (EXPR *)0);
			/* We no longer need the table.  Free it. */
			FREE(tab);
		} else {
			/* For a Charclass, "lit" will be '[' and the right
			 * child will point to the Charclass table.
			 */
			e = newexpr(r, Charclass, '[', (EXPR *)0, (EXPR *)tab);
		}
		break;
	case Lpar:
		/* Increment the count of left parentheses.
		 * We must save the current count, because if we have nested
		 * parentheses, Left_parno may be incremented inside the e0() call.
		 */
		count = ++Left_parno;

		/* Read the next token in the expression */
		lex(r);

		/* Process the portion of the expression inside the parentheses.
		 * We will stop at the corresponding right parenthesis.
		 */
		e = e0(r);
		if (Toktype != Rpar) err("expected a ')'");

		/* If we are backreferencing, we need a Group expression to
		 * reference.  "lit" will be a digit from 1-9, based on the
		 * sequential count of Left parentheses.  The left child will
		 * be <e>, the root of the subexpression defining the contents
		 * of the group.
		 */
		if (r->backref) {
			e = newexpr(r, Group, (CHAR)(count + '0'), e, (EXPR *)0);
			/* We store a pointer to the Group expression in
			 * the Groups array.
			 */
			Groups[count] = e;
			/* We store a count of the number of references to
			 * the group in the groups array.
			 */
			Group_refs[count] = 0;
		}
		break;
	default:
		err("expected a lit or '('");
		break;
	}
	/* Read the next token in the expression. */
	lex(r);

	/* Return a pointer to the last new expression allocated. */
	return(e);
}

/******************************************************************************

Routine:
	e2()

Syntax:

	static EXPR *e2(r)

	RE_RE *r;	// Pointer to regular expression being compiled

Description:

	Grammar:

	EOP:		eall:   { e0 } EOP
	Alternation:	e0:	e1 { '|' e1 }*
	Concatenation:	e1:	e2 { e2 }*
	Repetition:	e2:	e3 { '+' | '*' | '?' }
	Literal:	e3:	Literal | '.' | '[]' | Backref | '(' e0 ')'

	e2() is called by e1() during the COMPILATION PROPER to process
	a single Literal, Charclass ('['), Non_alpha ('#'), Dot ('.'),
	Backref ('`'), or Lpar...Rpar ("(...)") token, followed by 
	zero or more consecutive Plus ('+'), Star ('*'), or Quest ('?')
	tokens.

	e2() first calls e3() to process the next token.  If Toktype is not
	a Literal, Charclass ('['), Non_alpha ('#'), Dot ('.'), Backref ('`'),
	or Lpar ('(') token, it is a fatal error.

	e2() then checks if the next token is a Plus, Star, or Quest.
	If so, newexpr() is called to create a new Plus, Star, or Quest
	expression, and lex() is called to parse the next token.
	If the new token is also a Plus, Star, or Quest, this process
	repeats itself until a token other than Plus, Star, or Quest
	is encountered.

	NOTE: newexpr() will consolidate more than one consecutive Rept
	expression into a single Plus, Star, or Quest expression.

Global Variables:

	Toktype		// Token type of next token

Files:
	None

Return Status:

	Returns a pointer to the last new expression allocated.

*******************************************************************************/

static EXPR *
e2(r)
RE_RE *r;
{
	EXPR *e;

	/* Process one Literal, Charclass, Non_alpha, Dot, Backref, or
	 * an Lpar...Rpar.  Anything else will give an error.
	 */
	e = e3(r);

	/* Process all consecutive Plus, Star, or Quest.  */
	for ( ; PlusStarQuest(Toktype); lex (r)){

		/* Create the Plus, Star, or Quest expression. */
		e = newexpr(r, Toktype, '\0', e, (EXPR *)0);
	}
	/* We have a token other than a Plus, Star, or Quest. */

	/* Return a pointer to the last new expression allocated. */
	return(e);
}

/******************************************************************************

Routine:
	e1()

Syntax:

	static EXPR *e1(r)

	RE_RE *r;	// Pointer to regular expression being compiled

Description:

	Grammar:

	EOP:		eall:   { e0 } EOP
	Alternation:	e0:	e1 { '|' e1 }*
	Concatenation:	e1:	e2 { e2 }*
	Repetition:	e2:	e3 { '+' | '*' | '?' }
	Literal:	e3:	Literal | '.' | '[]' | Backref | '(' e0 ')'

	e1() is called by e0() during the COMPILATION PROPER to process
	a concatenation of one or more Literal, Charclass ('['), Non_alpha ('#'),
	Dot ('.'), Backref ('`'), or Lpar...Rpar ("(...)") tokens, each
	followed by zero or more consecutive Plus ('+'), Star ('*'),
	or Quest ('?') tokens.

	e1() first calls e2() to process the next token, plus any immediately
	following consecutive Plus ('+'), Star ('*'), or Quest ('?') tokens.
	If Toktype is not a Literal, Charclass ('['), Non_alpha ('#'), Dot ('.'),
	Backref ('`'), or Lpar ('(') token, it is a fatal error.

	e1() then checks if the next token is a Literal, Charclass ('['),
	Non_alpha ('#'), Dot ('.'), Backref ('`'), or Lpar ('(') token.
	If so, e2() is called again to process this token, plus any immediately
	following consecutive Plus ('+'), Star ('*'), or Quest ('?') tokens.
	Then newexpr() is called to create a new Cat expression, concatenating
	the expression returned by the previous call to e2() with the expression
	returned by this call to e2().

	The actions specified in the previous paragraph are performed repeatedly,
	until the next token is not one of the types checked for.  At this point,
	e1() will return.  This can only happen when e1() encounters the End of
	Pattern (EOP), an Alternate, or an Rpar(')'), (other than an Rpar
	encountered after a matching Lpar('(')).

Global Variables:

	Toktype		// Token type of next token

Files:
	None

Return Status:

	Returns a pointer to the last expression allocated.

*******************************************************************************/

static EXPR *
e1(r)
RE_RE *r;
{
	EXPR *e, *f;

	/* Process one Literal, Charclass, Non_alpha, Dot, Backref, or
	 * an Lpar...Rpar, and any following Plusses, Stars, or Quests.
	 * Anything else will give an error.
	 */
 	e = e2(r);
	/* Loop through subsequent Literal, Charclass, Non_alpha,
	 * Dot, Backref, or Lpar...Rpar.
	 */
	while (Toktype == Literal || Toktype == Charclass ||
	  Toktype == Non_alpha || Toktype == Dot || Toktype == Backref||
	  Toktype == Lpar) {
		/* Process another Literal, Charclass, Non_alpha, Dot, Backref, or
		 * an Lpar...Rpar, and any following Plusses, Stars, or Quests.
		 * Anything else will give an error.
		 */
		f = e2(r);
		/* Concatenate these processed expressions with the previous ones. */
		e = newexpr(r, Cat, '\0', e, f);
		/* Keep looping through subsequent Literal, Charclass, Non_alpha,
		 * Dot, Backref, or Lpar...Rpar.
		 */
	} /* end of while loop */

	/* We have an Alternate, Rpar, or EOP token. */

	/* Return a pointer to the last new expression allocated. */
	return(e);
}

/******************************************************************************

Routine:
	e0()

Syntax:

	static EXPR *e0(r)

	RE_RE *r;	// Pointer to regular expression being compiled

Description:

	Grammar:

	EOP:		eall:   { e0 } EOP
	Alternation:	e0:	e1 { '|' e1 }*
	Concatenation:	e1:	e2 { e2 }*
	Repetition:	e2:	e3 { '+' | '*' | '?' }
	Literal:	e3:	Literal | '.' | '[]' | Backref | '(' e0 ')'

	e0() is called by eall() or e3() during the COMPILATION PROPER to process
	an alternation of one or more concatenations of one or more Literal,
	Charclass ('['), Non_alpha ('#'), Dot ('.'), Backref ('`'),
	or Lpar...Rpar ("(...)") tokens, each followed by zero or more
	consecutive Plus ('+'), Star ('*'), or Quest ('?') tokens.

	e0() first calls e1() to process the next concatenation.
	If Toktype is not a Literal, Charclass ('['), Non_alpha ('#'), Dot ('.'),
	Backref ('`'), or Lpar ('(') token, it is a fatal error.

	e1() then checks if the next token is an Alternate.  If so, lex()
	is called to parse the next token, and e1() is called again to process
	the token and its corresponding concatenation.  Then newexpr() is called
	to create a new Alternate  expression, concatenating
	the expression returned by the previous call to e1() with the expression
	returned by this call to e1().

	The actions specified in the previous paragraph are performed repeatedly,
	until the next token is not an Alternate.  At this point, e0() will return.
	This can only happen when e1() encounters the End of Pattern (EOP) or an
	Rpar(')'), (other than an Rpar encountered after a matching Lpar('(')).

	e0() is called by eall() to process the entire pattern.  In this case,
	e0() should return when the End of Pattern (EOP) is encountered.
	e0() is called by e3() when an Lpar ('(') is encountered, to process the
	the portion of the pattern enclosed in the corresponding parentheses.
	In this case, e0() should return when the corresponding Rpar (')') is
	encountered.

Global Variables:

	Toktype		// Token type of next token

Files:
	None

Return Status:

	Returns a pointer to the last expression allocated.

*******************************************************************************/

static EXPR *
e0(r)
RE_RE *r;
{
	EXPR *e, *f;

	/* Process a concatenation of Literals, Charclasses, Non_alphas, Dots,
	 * Backrefs, or Lpar...Rpar's.  Each one may be followed by any number
	 * of Plusses, Stars, or Quests.  Anything else will give an error.
	 */
	e = e1(r);
	/* Is the next character an alternation? */
	while (Toktype == Alternate) {
		/* Process another concatenation of Literals, Charclasses,
		 * Non_alphas, Dots, Backrefs, or Lpar...Rpar's.  Each one may
		 * be followed by any number of Plusses, Stars, or Quests.
		 * Anything else will give an error.
		 */
		lex(r);
		f = e1(r);

		/* Alternate these processed expressions with the previous ones. */
		e = newexpr(r, Alternate, '\0', e, f);

		/* Loop if we have another alternation. */
	} /* end of while loop */

	/* We have an Rpar or EOP token. */

	/* Return a pointer to the last new expression allocated. */
	return(e);
}

/******************************************************************************

Routine:
	eall()

Syntax:

	static EXPR *eall(pat, r)

	PATTERN *pat;	// Pointer to pattern being compiled
	RE_RE *r;	// Pointer to regular expression being compiled

Description:

	eall() is called by re_recomp() to perform the COMPILATION PROPER.

	eall() does the following, in order:

	1) Initializes some global variables needed to perform the
	COMPILATION PROPER.  See "Global Variables" in this prologue.

	2) Calls lex() to parse the first token in the expression to be compiled.

	3) Calls e0() to compile the regular expression, generating the entire
	expression tree except for the root expression.

	4) Checks that e0() reached the end of the expression.

	5) Calls newexpr() to create the EOP expression at the root of the
	expression tree.

	6) Returns a pointer to the ROOT expression.

Global Variables:

	Ptr		// Pointer to current token
	End		// Pointer to end of uncompiled regular expression
	Groups		// Pointer to backreferencing groups
	Non_alpha_ccl	// Table for character class of non-alphanumerics
	Left_brackets	// Count of left angle brackets
	Right_brackets	// Count of right angle brackets
	Left_parno	// Count of left parentheses
	Toktype		// The type of the next token

Files:
	None

Return Status:

	None

*******************************************************************************/

static EXPR *
eall(pat, r)
PATTERN *pat;
RE_RE *r;
{
	EXPR *e;

	/* Calculate the non-alphanumeric character class only once for this
	 * pattern.  Due to possible character compression during optimization,
	 * each pattern needs its own copy.  However, this one copy can be shared
	 * within the same pattern.
	 */
	pat->non_alpha = (CHAR *)cgmalloc2(CCL_SIZ+1, sizeof(CHAR), "non-alpha charclass def");
	Ptr = non_alpha; 
	End = Ptr + sizeof(non_alpha) - 1;
	(void)ccl(pat->cmap, pat->non_alpha);
	Non_alpha_ccl = pat->non_alpha;

	/* Clear the pointers to backreferencing groups. */
	(void)memset(Groups, 0, 10*sizeof(EXPR *));

	/* Initialize the angle bracket counts to zero. */
	Left_brackets = Right_brackets = 0;

	/* Initialize the left parentheses count to zero. */
	Left_parno = 0;

	/* Now we begin the compilation.  First set up pointers to the
	 * beginning and end of the regular expression string.
	 */
	Ptr = pat->expression;
	End = pat->exp_end;

	/* Process the entire regular expression, up to the End of Pattern. */
	lex(r);
	e = e0(r);
	if (Toktype != EOP) err("expected end of pattern");

	/* Create the EOP expression.  */
	e = newexpr(r, EOP, '#', e, (EXPR *)0);

	/* Return a pointer to the EOP expression allocated. */
	return(e);
}

/******************************************************************************

Routine:
	err()

Syntax:

	void err(s)

	char *s;	// String to output in error message

Description:

	err() is called to output an error message and exit when an illegal
	token is encountered during the COMPILATION PROPER of an expression.

Global Variables:

	Toklit		// Illegal token that was encountered

Files:

	None

Return Status:

	Does not return

*******************************************************************************/

void
err(s)
char *s;
{
	char buf[4096];

	if (!Toklit)
		SPR(buf, "expression error: %s but got end of expression", s);
	else
		SPR(buf, "expression error: %s but got '%c'", s, *Toklit);
	re_error(buf);
}

/******************************************************************************

Routine:
	fill_expr()

Syntax:

	static void fill_expr(e)

	EXPR *e;	// Pointer to expression to be filled in.

Description:

	fill_expr() fills in information about an expression that depends
	on information in its children.

	fill_expr() is called by newexpr() when the new expression is created.
	This may occur during the COMPILATION, OPTIMIZATION, or PRE_MATCHING
	stages.  fill_expr() may also be called by percolate() during OPTIMIZATION
	to recalculate the information when an expression or a descendent
	is modified.

	A list of the members of the EXPR structure calculated by fill_expr is:

	group
	backref
	backref_no_star
	rept
	alter

	min_length
	min_length_ex
	max_length_ex

	dot
	dot_star_l
	dot_star_r

	nl_count
	nl_first
	nl_last
	nl_alt

	See file "re.h" or comments in this function for descriptions
	of these members.

Global Variables:

	Max_deletions

Files:

	None

Return Status:

	None

*******************************************************************************/

static void
fill_expr(e)
EXPR *e;
{
	EXPR *x;
	short lc, rc, l, r;
	int force_alt;
	BOOL lno, rno;

	/* Calculates whether the expression is or contains a Group
	 * or a Backref, assuming that this is already known for any children.
	 * Also calculates whether the expression is or contains a Backref
	 * other than a nulled (i.e. starred) backref.
	 */
	/* Also calculates whether the expression is or contains a repeat count,
	 * assuming that this is already known for any children.
	 */
	/* Also calculates whether the expression contains an INTERNAL alternation,
	 * assuming that this is already known for any children.  An internal
	 * alternation is an alternation below something other than an alternation
	 * or an EOP, i.e. it must be inside essential parentheses.
	 * Note that <alter> is not set for the internal alternation itself,
	 * only for its ancestors.  This is because, when creating expressions
	 * from the bottom up, we cannot yet tell whether an alternation
	 * is internal or external.
	 *
	 * Example of expression with internal alternation: "tom(boy|cat)"
	 * Example of expression with external alternation: "tom|boy|cat"
	 */
	switch(e->type) {
	case Group:
		e->backref_no_star = e->l->backref_no_star;
		e->group = YES;
		e->backref = e->l->backref;
		e->rept = e->l->rept;
		e->alter = (BOOL)(e->l->alter || e->l->type == Alternate);
		break;
	case Backref:
		e->group = NO;
		e->backref_no_star = e->backref = YES;
		e->rept = e->l->rept;
		e->alter = (BOOL)(e->l->alter || e->l->type == Alternate);
		break;
	case EOP:
		e->backref_no_star = e->l->backref_no_star;
		e->group = e->l->group;
		e->backref = e->l->backref;
		e->rept = e->l->rept;
		e->alter = e->l->alter;
		break;
	case Plus:
		e->backref_no_star = e->l->backref_no_star;
		e->group = e->l->group;
		e->backref = e->l->backref;
		e->rept = YES;
		e->alter = (BOOL)(e->l->alter || e->l->type == Alternate);
		break;
	case Star: case Quest:
		e->backref_no_star = NO;
		e->group = e->l->group;
		e->backref = e->l->backref;
		e->rept = YES;
		e->alter = (BOOL)(e->l->alter || e->l->type == Alternate);
		break;
	case Alternate:
		e->backref_no_star = (BOOL) (e->l->backref_no_star || e->r->backref_no_star);
		e->group = (BOOL) (e->l->group || e->r->group);
		e->backref = (BOOL) (e->l->backref || e->r->backref);
		e->rept = (BOOL)(e->l->rept || e->r->rept);
		e->alter = (BOOL)(e->l->alter || e->r->alter);
		break;
	case Cat:
		e->backref_no_star = (BOOL) (e->l->backref_no_star || e->r->backref_no_star);
		e->group = (BOOL) (e->l->group || e->r->group);
		e->backref = (BOOL) (e->l->backref || e->r->backref);
		e->rept = (BOOL)(e->l->rept || e->r->rept);
		e->alter = (BOOL)(e->l->alter || e->r->alter ||
		    e->l->type == Alternate || e->r->type == Alternate);
		break;
	default:
		e->backref_no_star = e->group = e->backref = e->rept = e->alter = NO;
		break;
	}

	/* Calculates "min_length_ex" and "max_length_ex", the minimum and maximum
	 * length of text that will exactly match this expression, assuming that
	 * this is already known for any children.
	 * Note that for the maximum length, -1 represents unlimited length.
	 *
	 * Also calculates "min_length", the minimum length of text that will
	 * match this expression, allowing for deletions due to approximate
	 * matching.
	 */
	switch(e->type) {
	case EOP: case Group: case Backref:

		/* These expressions don't change any lengths. */
		e->max_length_ex = e->l->max_length_ex;
		e->min_length_ex = e->l->min_length_ex;
		e->min_length = e->l->min_length;
		break;
	case Quest:
		/* Quest don't change the maximum length. */
		e->max_length_ex = e->l->max_length_ex;
		/* Quest makes the minimum lengths zero.  */
		e->min_length_ex = e->min_length = 0;
		break;
	case Star:
		/* Star makes the the maximum length unlimited. */
		e->max_length_ex = -1;
		/* Star makes the minimum lengths zero.  */
		e->min_length_ex = e->min_length = 0;
		break;
	case Plus:
		/* Star makes the the maximum length unlimited. */
		e->max_length_ex = -1;
		/* Plus don't change the minimum lengths. */
		e->min_length_ex = e->l->min_length_ex;
		e->min_length = e->l->min_length;
		break;
	case Alternate:
		/* The maximum length is the maximum of the values for the
		 * two children, treating "-1" as infinity.
		 */
		l = e->l->max_length_ex;
		r = e->r->max_length_ex;
		e->max_length_ex = ((l <= 0 || r <= 0) ? -1 : MAX(l, r));
		/* Each minimum length is the minimum of the values for
		 * the two children.
		 */
		e->min_length_ex = MIN(e->l->min_length_ex, e->r->min_length_ex);
		e->min_length = MIN(e->l->min_length, e->r->min_length);
		break;
	case Cat:
		/* The maximum length is the sum of the values for the
		 * two children, treating "-1" as infinity.
		 */
		l = e->l->max_length_ex;
		r = e->r->max_length_ex;
		e->max_length_ex = ((l <= 0 || r <= 0) ? -1 : l + r);
		/* The minimum length for exact matching is the sum of the values
		 * for the two children.
		 */
		e->min_length_ex = e->l->min_length_ex + e->r->min_length_ex;
		/* The minimum length, allowing for deletions, is the sum of the
		 * values for the two children, except that the total number
		 * of deletions allowed is capped by Max_deletions.
		 */
		e->min_length = MAX(e->min_length_ex - Max_deletions,
			e->l->min_length + e->r->min_length);

		break;
	default:
		/* The minimum and maximum lengths of text for matching the
		 * "lit" are each 1.
		 */
		e->min_length_ex = e->max_length_ex = 1;
		/* The minimum length of text that matches this "lit" is 1 if
		 * it is inside angle brackets or if deletions are not
		 * allowed, and is 0 otherwise.
		 */
		e->min_length = ((e->n_bracket || !Max_deletions) ? 1 : 0);

		break;
	}

	/*
	 * Sets Dot Star info needed for trim_nl()
	 *
	 * For a concatenation of one or more "lits", <dot> is YES if and only if at
	 * least one "lit" is a Dot, and every "lit" that is not a Dot is nullable
	 * (i.e., under a Star or Quest).
	 *
	 * Examples:
	 *
	 * 	".ab.c*":	<dot> is NO. ('a' and 'b' are not nullable)
	 * 	".a*b?.c*"	<dot> is YES ('a', 'b', 'c' all nullable)
	 * 	"a*b?c*"	<dot is NO> (contains no '.')
	 *
	 * For a concatenation of one or more "lits", <dot_star_l> (<dot_star_r>)
	 * is YES if and only if the concatenation contains a ".*" or a ".+", and
	 * every lit to the left (right) of the leftmost (rightmost) ".*" or ".+"
	 * is either nullable (i.e., under a * Star or Quest) or a Dot.
	 *
	 * Examples:
	 *
	 * 	".*abcd":	<dot_star_l> is YES (nothing to the left of ".*")
	 * 	".x*.y*.*abcd":	<dot_star_l> is YES (everything to the left of ".*"
	 * 				is either nullable or a Dot)
	 * 	".x.y*.*abcd":	<dot_star_l> is NO ('x' is not nullable)
	 *
	 * For an alternation, these quantities are true if they are true of
	 * any member of the alternation.
	 */
	switch(e->type) {
	case EOP: case Group: case Backref: case Quest:
		/* These expressions don't change anything. */
		e->dot = e->l->dot;
		e->dot_star_l = e->l->dot_star_l;
		e->dot_star_r = e->l->dot_star_r;
		break;
	case Plus: case Star:
		/* These expressions don't change <dot>. */
		e->dot = e->l->dot;

		/* If <dot> is on, turn on <dot_star_l> and <dot_star_r>. */
		e->dot_star_l = (BOOL)(e->l->dot || e->l->dot_star_l);
		e->dot_star_r = (BOOL)(e->l->dot || e->l->dot_star_r);
		break;
	case Alternate:
		/* OR the quantities */
		e->dot = (BOOL)(e->l->dot || e->r->dot);
		e->dot_star_l = (BOOL)(e->l->dot_star_l || e->r->dot_star_l);
		e->dot_star_r = (BOOL)(e->l->dot_star_r || e->r->dot_star_r);
		break;
	case Cat:
		/* Do the left and right members contain a non-nullable "lit"
		 * other than a <dot>?  Each does so if and only if it is
		 * non-nullable, and has <dot> == NO.
		 */
		lno = (BOOL)(e->l->min_length_ex && !e->l->dot);
		rno = (BOOL)(e->r->min_length_ex && !e->r->dot);
		/* If either member contains a non-nullable non-Dot, set <dot>
		 * to no.  Otherwise, set <dot> to the OR of <dot> for the two
		 * members.
		 */
		if (lno || rno) e->dot = NO;
		else e->dot = (BOOL)(e->l->dot || e->r->dot);

		/* If the left member contains a non-nullable non-Dot,
		 * then the right member cannot effect <dot_star_l>,
		 * so <dot_star_l> is just inherited from the left member.
		 * Otherwise, set <dot_star_l> to the OR of <dot_star_l> for
		 * the two members.  <dot_star_r> is handled similarly.
		 */
		if (lno) e->dot_star_l = e->l->dot_star_l;
		else e->dot_star_l = (BOOL)(e->l->dot_star_l || e->r->dot_star_l);
		if (rno) e->dot_star_r = e->r->dot_star_r;
		else e->dot_star_r = (BOOL)(e->l->dot_star_r || e->r->dot_star_r);
		break;
	default:
		/* For a "lit", <dot> will be YES if and only if the expression
		 * is a Dot, and <dot_star_l> and <dot_star_r> will be NO.
		 */
		e->dot = (BOOL)(e->type == Dot);
		e->dot_star_l = e->dot_star_r = NO;
		break;
	}

	/* Calculates some quantities relating to the maximum number of
	 * new-line characters in text that can match an expression,
	 * assuming they are already known for the children.
	 *
	 * A subexpression is any of the possible alternate expressions.
	 *
	 * The "dominant" subexpression is the one that can match the most
	 * new-line characters.  See the comments in the "Alternate" case
	 * below on resolving ties!
	 *
	 * int nl_count:  The maximum number of new-lines in text that can match
	 * 			the dominant subexpression, but not more than MANY.
	 *
	 * int  nl_first:  Does the the leftmost "lit" in the dominant
	 * 			subexpression match a new-line? (0 or 1)
	 *
	 * int nl_last:  Does the rightmost "lit" in the dominant
	 *			subexpression match a new-line? (0 or 1)
	 *
	 * BOOL nl_alt:  Set to YES if and only if one of the following holds:
	 *
	 * 		  1) The dominant subexpression has one, but not both,
	 * 		  of nl_first and nl_last set.  Another subexpression
	 * 		  that has the opposite settings of nl_first and nl_last
	 * 		  can match the same number of new-lines as the dominant.
	 * 		  OR
	 * 		  2) The dominant subexpression has both nl_first and
	 * 		  nl_last set.  Another subexpression that has neither
	 * 		  nl_first nor nl_last set can match one less new-line
	 * 		  as the dominant.
	 *
	 * 		  Basically, nl_alt is set if the characteristics of
	 * 		  the dominant subexpression are uncertain (i.e., the
	 * 		  other subexpression cannot be ignored).
	 */
	switch(e->type) {
	case Alternate:
		/* Find the dominant subexpression.  The dominant is the one
		 * that can match the most new-line characters, except that
		 * new-line matches in the first and last positions count
		 * only half as much (they may not contribute to line spanning).
		 * A subexpression with a new-line in the interior will
		 * tie one with a new-line at the beginning and a new-line at
		 * the end.
		 */
		lc = 2*e->l->nl_count - e->l->nl_first - e->l->nl_last;
		rc = 2*e->r->nl_count - e->r->nl_first - e->r->nl_last;
		x = 0;
		force_alt = 0;
		if (lc > rc) {
			/* Left guaranteed at least as many NL's as e->r. */
			x = e->l;
		} else if (rc > lc) {
			/* Right guaranteed at least as many NL's as e->l. */
			x = e->r;
		} else if (e->l->nl_alt) {
			/* Left and Right have the same "count", but Left
			 * already has nl_alt set, so Left covers both cases.
			 */
			x = e->l;
		} else if (e->r->nl_alt) {
			/* Left and Right have the same "count", but Right
			 * already has nl_alt set, so Right covers both cases.
			 */
			x = e->r; 
		} else if (e->l->nl_first == e->r->nl_first) {
			/* Left and Right must be equivalent, so we
			 * arbitrarily pick Left.  It does not matter which.
			 */
			x = e->l;
		} else if (e->l->nl_first) {
			/* Left and Right have the same "count", but are not
			 * equivalent.  Since Left has nl_first set, Left must
			 * be guaranteed at least as many NL's as Right.
			 * So pick Left, but force nl_alt to be set.
			 */
			x = e->l;
			force_alt = 1;
		} else {
			/* Left and Right have the same "count", but are not
			 * equivalent.  Since Left does not have nl_first set,
			 * Right must be guaranteed at least as many NL's as
			 * Left.  So pick Right, but force nl_alt to be set.
			 */
			x = e->r;
			force_alt = 1;
		}
		/* nl_first, nl_last, nl_count, and nl_alt are taken
		 * from the dominant subexpression, except that force_alt
		 * forces nl_alt to be on.
		 */
		e->nl_count = x->nl_count;
		e->nl_first = x->nl_first; 
		e->nl_last = x->nl_last;
		e->nl_alt = (force_alt ? YES : x->nl_alt);

		break;
	case Cat:
		/* If either child has "nl_alt" set and only one of
		 * "nl_first" or "nl_last" set (Case 1 in the prologue),
		 * we make sure that the one in the interior of the
		 * concatenation is set.  The child remains ambiguous,
		 * but the concatenation will not be!
		 */
		if (e->l->nl_alt && !e->l->nl_last) {
			e->l->nl_first = 0;
			e->l->nl_last = 1;
		}
		if (e->r->nl_alt && !e->r->nl_first) {
			e->r->nl_first = 1;
			e->r->nl_last = 0;
		}
		e->nl_alt = NO;

		/* The count is the sum of the pieces being concatenated.
		 * nl_first comes from the left piece,
		 * while nl_last comes from the right piece.
		 */
		e->nl_count = MIN(e->l->nl_count + e->r->nl_count, MANY);
		e->nl_first = e->l->nl_first;
		e->nl_last = e->r->nl_last;
		break;
	case EOP: case Group: case Backref: case Quest:
		/* These do not affect anything. */
		e->nl_count = e->l->nl_count;
		e->nl_first = e->l->nl_first;
		e->nl_last = e->l->nl_last;
		e->nl_alt = e->l->nl_alt;
		break;
	case Plus: case Star:
		/* Plus and Star do not affect anything, except that
		 * the new-line count becomes the maximum if it was non-zero.
		 */
		e->nl_count = (e->l->nl_count ? MANY : 0);
		e->nl_first = e->l->nl_first;
		e->nl_last = e->l->nl_last;
		e->nl_alt = e->l->nl_alt;
		break;
	case Dot:
		/* Dot does not match new-line */
		e->nl_first = e->nl_last = e->nl_count = 0;
		e->nl_alt = NO;
		break;
	case Charclass:
		/* Charclass must be checked for match to new-line. */
		e->nl_first = e->nl_last =
		    e->nl_count = CCL_CHK((CHAR *)e->r, NL);
		e->nl_alt = NO;
		break;
	case Literal:
		/* Literal matches new-line only if it is one. */
		e->nl_first = e->nl_last = 
		    e->nl_count = (e->lit == NL ? 1 : 0);
		e->nl_alt = NO;
		break;
	}
}

/******************************************************************************

Routine:
	optimize_exp()

Syntax:

	static void optimize_exp(pat, r)

	PATTERN *pat;	// Pointer to pattern to be matched. 
	RE_RE *r;	// Pointer to regular expression being compiled

Description:

	optimize_exp() is the main routine for the OPTIMIZATION stage
	of compilation.  Wherever possible, the compiled regular
	expression is converted to an equivalent regular expression that
	will match the same lines of code and provide better performance.

	The various optimizations are explained in the in-line comments in
	optimize_exp(), and in the prologues and comments in the individual
	optimization functions.

Global Variables:

	Groups		// Array of pointers to backreferencing groups
	Group_refs	// Array of reference counts to backreferencing Groups

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
optimize_exp(pat, r)
PATTERN *pat;
RE_RE *r;
{
	register int i;

	/* Check for a nullable pattern or a pattern that must match
	 * every line.  In either case, change it to a pattern consisting
	 * of a single new-line.
	 */
	if (matches_every_line(pat, r)) return;

	/* Allocate some arrays we need during optimization. */
	r->fpos.base = (PSFID *)cgmalloc(r->maxid * sizeof(PSFID), "fpos for optimization()");
	r->ptr = (EXPR **)cgcalloc(r->maxid, sizeof(EXPR *), "ptr");

	/* Fill in the "r->ptr[]" array with pointers to each "lit" expression. */
	eptr(r, r->root);

	/* Put nullable alternations into "standard" form.  See the prologue
	 * of null_alter() for details.
	 */
	null_alter(r, r->root, NO);

	/* Delete any backreference Group expressions that are not referenced
	 * by a backref.  The descendents of the Group are preserved.
	 */
	for (i = 1; i < 10; i++) {
		if (Groups[i] && Group_refs[i] == 0) {
			delete_rept(Groups[i]);
		}
	}

	/* Often, expressions can be trimmed off the ends of patterns without
	 * affecting the lines that are matched.  For example, "55" matches
	 * a line if and only if it contains two consecutive '5' characters.
	 * However, "a*55b*c*" matches the very same lines.  Hence, "a*55b*c*"
	 * can be trimmed to "55".
	 */
	if (trim(pat, r, RIGHT)) return;
	if (trim(pat, r, LEFT)) return;

	/* Compress adjacent items when possible.  For example, in "egrep"
	 * mode, "e+e*" is equivalent to "e+", and "[a-z]*e*" is equivalent
	 * to "[a-z]*" Also moves repeat counts from LEFT to RIGHT where possible, 
	 * e,g, "e+e" will be changed to "ee+".  This helps make sure that
	 * all possible compressions are done.
	 */
	adj_compress(r->root, LEFT);

	/* Compress adjacent items again when possible.  Also moves repeat counts
	 * from Right to Left where possible, * e,g, "ee+" will be changed to "e+e".
	 * This helps make sure that all possible compressions are done.
	 */
	adj_compress(r->root, RIGHT);

	/* Often, expressions can be trimmed off the ends of patterns without
	 * affecting the lines that are matched.  For example, "55" matches
	 * a line if and only if it contains two consecutive '5' characters.
	 * However, "a*55b*c*" matches the very same lines.  Hence, "a*55b*c*"
	 * can be trimmed to "55".
	 */
	/* Trim irrelevant stuff off left end */
	if (trim(pat, r, LEFT)) return;

	/* Compress adjacent items again when possible, moving repeat counts
	 * to the right. */
	adj_compress(r->root, LEFT);

	/* Trim irrelevant stuff off right end */
	if (trim(pat, r, RIGHT)) return;

	/* Create new character map that maps all equivalent text characters
	 * into a single character.  For example, in matching "[LBH]arry",
	 * 'L', 'B', and 'H' text characters are equivalent.  Mapping them
	 * into a single character allows the character class "[LBH]" to be
	 * replaced by a Literal, which can greatly help performance.
	 */
	char_compress(pat, r);

	/* We don't need these arrays any more. */
	FREE(r->fpos.base);
	FREE(r->ptr);
}

/******************************************************************************

Routine:
	char_compress()

Syntax:

	static void char_compress(pat, r)

	PATTERN *pat;	// Pointer to pattern to be matched. 
	RE_RE *r;	// Pointer to regular expression being optimized

Description:

	char_compress() creates a character map for a regular expression
	that maps all "equivalent" ASCII characters into a single ASCII
	character.  All "equivalent" ASCII characters will be mapped into
	the character in the set of equivalent characters that has the
	lowest ASCII code, except that if one of those characters is
	new-line, the characters will be mapped into new-line.

	char_compress() starts out with a map that maps each character to
	itself, unless the "-i" option applies to the pattern, in which case
	each lower case character is equivalent to and mapped into the
	corresponding upper case character.

	Other than equivalence due to the case insensitive option, two
	ASCII characters are equivalent (with respect to a pattern to
	be matched) if:

	1) Neither is used in the pattern as a literal.
	AND
	2) For each Charclass in the pattern, either both characters
	are in the Charclass, or neither is.
	AND
	3) If one of the two characters is a new-line, there is no Dot in
	the pattern.
	AND
	4) Neither character can match any "lit" in a backreferencing Group.

	The advantages of character mapping for performance are two-fold:

	1) If every character in a character class is equivalent, the
	character class can be replaced by a Literal.  This may convert a
	regular expression into fixed strings and allow a much faster
	algorithm to be used, or increase the length of the fixed strings
	embedded in the regular expression, which may speed up a multistage
	match.

	2) When an entry in the state transition table is filled in
	during a match to a regular expression, it applies to all
	equivalent characters.  This speeds up matching by eliminating many
	calls to followstate() and stateof() during text matching.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
char_compress(pat, r)
PATTERN *pat;
RE_RE *r;
{
	CHAR mappit[NCHARS];
	register EXPR *p;
	register int i, j, k, l;
	int ncharclass = 0;
	CHAR *tab;

#ifdef DEBUG
	print_expr(R->root);
#endif
	/* No character compression if backreferencing, for now. */
	if (pat->backref) {
		return;
	}

	/* Allocate a new character map, and copy the existing map into it. */
	r->cmap = cgmalloc(NCHARS*sizeof(char), "compressed cmap");
	(void)memcpy(r->cmap, pat->cmap, NCHARS*sizeof(char));
	pat->cmap = r->cmap;

	/* Fill in the "r->ptr" array with all the "lit" expressions
	 * in the regular expression.  This array will be used by
	 * char_compress() and freed afterwards.
	 */
	FREE(r->ptr);
	r->ptr = (EXPR **)cgcalloc(r->maxid, sizeof(EXPR *), "ptr");
	eptr(r, r->root);

	/* Set every entry in the mappit array to 1, indicating that
	 * every character is potentially mappable.
	 */
	(void)memset(mappit, '\001', NCHARS*sizeof(CHAR));

	/* If the cost of insertions is zero, the cost of the insertion
	 * of a new-line is still 1.  In that case, new-line is not equivalent
	 * to any other character, and is not mappable.
	 */
	if (r->cost_ins == 0) mappit[NL] = 0;

	/* Loop through the "lits" in the expression. */
	for (i = 1; i < r->maxid; i++) {
		/* If there is no entry for this ID, that "lit" must have been
		 * deleted in an earlier optimization.  Continue with the next one.
		 */
		if (!r->ptr[i]) continue;

		/* Get the entry. */
		p = r->ptr[i];

		switch (p->type) {
		/* If a Literal, the character it maps to is uniquely used,
		 * and cannot be mapped.
		 */
		case Literal: mappit[r->cmap[p->lit]] = 0; break;

		/* If a Dot, it matches everything except new-line.
		 * New-line is unique and cannot be mapped.
		 */
		case Dot: mappit[NL] = 0; break;

		/* If a Charclass, we don't do anything now, except
		 * to change the ID numbers of charclass entries
		 * so they are densely packed, starting at index 1.
		 * This is not necessary, but can greatly speeds performance
		 * of the compression.
		 */
		case Charclass:
			/* Increment the count of charclasses. */
			ncharclass++;

			/* The charclass has ID "i", and we would like to
			 * give it ID "ncharclass".  So, to be non-destructive,
			 * we exchange the ID's of the two "lit" expressions,
			 * the one with ID "i" and ID "ncharclass".
			 * If this charclass ID number already has
			 * the desired ID number, don't change anything.
			 */
			if (i == ncharclass) break;

			/* Note that "i" is greater than "ncharclass", so
			 * we will not hit either of these again in this for loop.
			 */
			p->id = ncharclass;
			if (r->ptr[ncharclass]) r->ptr[ncharclass]->id = i;

			/* Now, exchange the pointers in the ptr[] array to
			 * match the new ID's.
			 */
			r->ptr[i] = r->ptr[ncharclass];
			r->ptr[ncharclass] = p;
			break;
		}
	}

	/* Don't map any character that is already mapped into a different char. */
	for (i = 0; i < NCHARS; i++) if (r->cmap[i] != i) mappit[i] = 0;

	/* Now check character classes.  Loop over the character set.
	 * We always map a character into the equivalent with the smallest
	 * ASCII value, except that we map characters to new-line, rather
	 * than mapping new-lines to other characters.  We do this by
	 * starting at -1, and treating new-line on this first pass,
	 */
	if (ncharclass > 0) {
	    for (k = -1; k < NCHARS; k++) {
		/* To put new-line first, we handle it when k == -1,
		 * not when k == NL.
		 */
		if (k == NL) continue;
		i = (k < 0 ? NL : k);

		/* If the character whose ASCII code is <i> is mappable,
		 * loop over the remainder of the character set, to check if
		 * the character whose ASCII code is <j> should be mapped into
		 * the character whose ASCII code is <i>.  If at some point
		 * it is determined that the character whose ASCII code is <i>
		 * is not mappable, don't keep looping.
		 */
		for (j = k+1; mappit[i] && j < NCHARS; j++) {
			/* If this character <j> can't be mapped, skip it. */
			if (!mappit[j]) continue;

			/* Don't map new-line into anything else. */
			if (j == NL) continue;

			/* Loop over the character classes, and try to find
			 * one in which <i> is included and <j> not included,
			 * or vice versa.
			 */
			for (l = 1; l <= ncharclass; l++) {
				p = r->ptr[l];
				if (p->type == Literal) continue;
				tab = (CHAR *)(p->r);
				/* Compare the presence of <i> and <j> in the
				 * character class.
				 */
				if (CCL_CHK(tab, (CHAR)i) != CCL_CHK(tab, (CHAR)j)) { 
					/* This character class includes one
					 * of <i> and <j>, but not the other.
					 * <j> can't be mapped to <i>.
					 */
					break;
				}
			}
			/* If <j> can't be mapped to <i>, continue. */
			if (l <= ncharclass) continue;

			/* Map <j> to <i>.  */
			r->cmap[j] = (CHAR)i;

			/* Indicate <j> has already been mapped. */
			mappit[j] = 0;

			/* Loop through all character classes, and remove
			 * <j> from all classes that it is in.
			 */
			for (l = 1; l <= ncharclass; l++) {
				p = r->ptr[l];
				if (p->type == Literal) continue;
				tab = (CHAR *)(p->r);
				if (CCL_CHK(tab, (CHAR)j)) {
					/* Remove <j>, and decrement the count.
					 * Note that if this table is shared,
					 * this effectively removes <j>
					 * from all instances of the class.
					 */
					CCL_CLR(tab, (CHAR)j);
					tab[CCL_SIZ]--;
				}
				if (tab[CCL_SIZ] == 1) {
					/* The character class has been reduced to
					 * a single member.  Convert it to a Literal.
					 * <i> must be the literal, because if <j>
					 * was in the class, <i> must be also.
					 * Don't free the table because it may
					 * still be shared.
					 */
					p->type = Literal;
					p->r = 0;
					p->lit = (CHAR)i;
					/* <i> is a Literal, so it can't be mapped.*/
					mappit[i] = 0;
				}
			}
		}
	    }
	}

	/* We may have a case where 'b' maps to 'B' and 'B' maps to 'A'.
	 * This can occur because of prior mappings due to the -i option.
	 * We resolve these transitive mappings.  We need go only one level
	 * deep.
	 */
	for (i = 0; i < NCHARS; i++) r->cmap[i] = r->cmap[r->cmap[i]];

#ifdef DEBUG
	print_expr(R->root);
#endif
}

/******************************************************************************

Routine:
	trim()

Syntax:

	static BOOL trim(pat, r, direction)

	PATTERN *pat;	// Pointer to pattern to be matched. 
	RE_RE *r;	// Pointer to regular expression being optimized
	DIRECTION direction;	// Which end of the pattern is being
				//   trimmed, RIGHT or LEFT.

Description:

	trim() is the main routine for calling the routines that
	remove irrelevant expressions from the ends of the pattern
	as part of the OPTIMIZATION stage of compilation.   trim()
	calls trim_nl() and trim_null() to actually do the trimming.
	trim_nl() trims irrelevant new-line characters from the ends
	of the pattern to be matched, while trim_null() trims irrelevant
	nullable expressions from the ends of the patterns to be matched.

	After calling trim_nl(), trim() calls matches_every_line() to determine
	if the remaining pattern must match every line of text.  If so,
	matches_every_line() optimizes the expression and returns YES, and trim()
	then returns YES, indicating to the caller that no further
	optimization need be attempted.

	If pat->pbounds is set, no trimming is done, as the exact extent
	of the match to the pattern needs to be known.

Global Variables:

	None

Files:
	None

Return Status:

	Returns YES if matches_every_line() has determined that the pattern
	now will match every line of text.

	Returns NO otherwise.

*******************************************************************************/

static BOOL
trim(pat, r, direction)
PATTERN *pat;
RE_RE *r;
DIRECTION direction;
{
	/* If <pbounds> is set for this pattern, we need to find the exact
	 * length of the match.  Therefore, do not do any trimming.
	 */
	if (pat->pbounds) {
		return(NO);
	}
	/* Trim new-lines from the specified end of the pattern, where
	 * possible.
	 */
	trim_nl(r->root, direction, YES);

	/* If the pattern now must match every line, optimize the pattern,
	 * and inform the caller that no further optimization is needed.
	 */

	if (matches_every_line(pat, r)) return(YES);

	/* Trim nullable expressions from the ends of the pattern, where
	 * possible.
	 */
	trim_null(r, r->root, direction);
	return(NO);
}

/******************************************************************************

Routine:
	trim_nl()

Syntax:

	static void trim_nl(e, direction, top)

	EXPR *e;	// Pointer to expression to be trimmed, where possible.
	DIRECTION direction;	// Which end of the pattern is being
				//   trimmed, RIGHT or LEFT.
	BOOL top;	// Are we traversing the highest level Alternation?

Description:

	During the OPTIMIZATION stage of compilation, trim_nl() trims
	new-line characters (or character classes containing new-line) off
	the ends of patterns when it is possible to do so without affecting
	matches.

	A new-line character (or a character class containing a new-line)
	can be trimmed if it is at one end of the pattern, and is
	adjacent to a ".*" or ".+", or is separated from a ".*" or ".+"
	only by Dots or nullable characters (characters below a Star or Quest),
	AND if the new-line character is not itself in a Plus or Star, AND
	if none of the new-line character, the ".*" or ".+", and intervening
	expressions are in a backreference Group or Backref (other than
	a Backref Star).

	Examples:

	In the following cases, the new-line character, indicated by '\n',
	can be deleted from the pattern:

	"\n.*abc"
	"\na?b*..c*.*xyz"
	"\n?a?b*.c*.*xyz"
	"\na*.*\(..\)\1"
	"(\na*|rst)a?b*.c*.*xyz"
	"(\n.*|rst)abc"
	"\(..\).*\1*\n"

	In the following cases, the new-line character cannot be deleted:

	"a*\n.*abc"	new-line is not at end of pattern
	"\na*b*"	no ".+" or ".*"
	"\nx.*abc"	'x' (not nullable or a Dot) is between '\n' and ".*"
	"\n*.*abc"	'\n' is in a '*'
	"\(\n.*\).*\1"	'\n' is in a backreference Group
	"\n\(..\).*\1"	backreference Group is between '\n' and ".*"
	"\n\(.*\)\1"	".*" is in backreference Group
	"\(..\).*\1\n"	Backref (not Backref Star) is between '\n' and ".*"

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
trim_nl(e, direction, top)
register EXPR *e;
DIRECTION direction;
BOOL top;
{
	register EXPR *p;

	/* If the expression does not contain a new line, just return. */
	if (!e->nl_count) return;

	switch(e->type) {
	case EOP:
		/* A new-line can be trimmed only if the pattern as a whole
		 * contains a Plus or a Star.
		 */
		if (e->l->max_length_ex == -1) trim_nl(e->l, direction, YES);
		return;
	case Plus: case Star: case Group: case Backref:
		/* A new-line inside a Plus, Star, or backreference Group
		 * cannot be trimmed.
		 */
		return;
	case Quest:
		/* A new-line inside a Quest can be trimmed. */
		trim_nl(e->l, direction, NO);
		return;
	case Cat:
		/* In a concatenation, always take the member in the
		 * direction we are matching, so we get the characters
		 * at the desired end of the pattern.
		 */
		trim_nl(((direction == RIGHT) ? e->r : e->l), direction, NO);
		return;
	case Alternate:
		/* In an alternation, we attempt to trim both members.
		 * However, if this is a top level alternation, we
		 * can't trim a new-line unless there is a Plus or Star
		 * in the same member of the alternation.
		 */
		if (!top || e->r->max_length_ex == -1) trim_nl(e->r, direction, top);
		if (!top || e->l->max_length_ex == -1) trim_nl(e->l, direction, top);
		return;
	}
	/* We have reached a new-line, or a character class containing
	 * a new-line, at the desired end of the pattern.  Get the neighbor
	 * of the new-line.  Note that we pass in the same direction,
	 * because get_next() searches FROM <direction>, not TO <direction>.
	 */
	p = e;
	while ((p = get_next(p, direction, (BOOL *)0)) != 0) {
		/* If the neighbor is or contains a Group or Backref,
		 * other than a Backref Star, we cannot trim the new-line.
		 */
		if (p->group || p->backref_no_star) break;
		/* Otherwise if the neighbor is a ".*", or ".+", or a composite
		 * with a ".*" or ".+" at the proper end, we can delete
		 * the new-line.
		 */
		if ((direction == RIGHT) ? p->dot_star_r : p->dot_star_l) {
			delete_expr(e);
			break;

		/* Otherwise, if the neighbor contains no non-nullable
		 * characters other than Dot, we can skip over it and
		 * check ITS neighbor.
		 */
		} else if (!p->min_length_ex || p->dot) {
			continue;
		}
		/* Otherwise, we cannot trim the new-line. */
		break;
	}
}

/******************************************************************************

Routine:
	trim_null()

Syntax:

	static void trim_null(e, direction, top)

	RE_RE *r;	// Pointer to the regular expression we are optimizing.
	EXPR *e;	// Pointer to expression to be trimmed, if possible.
	DIRECTION direction;	// Which end of the pattern is being
				//   trimmed, RIGHT or LEFT.

Description:

	During the OPTIMIZATION stage of compilation, trim_null() trims
	nullable expressions (composite expressions allowed) off the ends
	of patterns when it is possible to do so without affecting the lines
	that will be matched.

	Nullable expressions inside backreferencing Groups or other
	backreferencing constructs cannot be trimmed.  Other than
	in backreferencing, a nullable expression (an expression that
	can match the null string) can always be removed from one end
	of a regular expression without affecting when a match is found.
	However, deleting the null expression may reduce the extent of
	the text that is matched.  If this reduction in extent might cause
	the match to extend to fewer lines, the null expression may not
	be deleted.  This can occur in two cases:

	1) The expression to be potentially deleted itself can match text that
	contains a new-line, other than a new line at the end to be trimmed.

	OR

	2) The expression to be potentially deleted can match text adjacent
	to a new-line, thereby extending the match onto another line.

	Examples ('\n' indicates a new-line):

	In the following cases, the patterns can be trimmed as follows:

	"x*abc" 	=> "abc"
	"\n?abc"	=> "abc"	// The new-line is at the end to be trimmed.
	"(\nx)?abc"	=> "abc"	// The new-line is at the end to be trimmed.
	"x+abc"		=> "xabc"	// "x+" is equivalent to "x*x", and "x*"
					//	can be trimmed.
	"x*\nabc"	=> "x?abc"	// Matches to more than one 'x' cannot
					//	extend the match to a new-line.
	"\(x*\)abc\1\1*" => "\(x*\)abc\1"  // The "x*" is inside a backreferencing
					   //    Group, and therefore cannot
					   //    be trimmed.  The "\1*" can be
					   //    trimmed, but the "\1" cannot be.  
	"\(x*\)abc\1*" => "\(x*\)abc" => "x*abc" => "abc"
					// The "\1*" can be trimmed first.
					// Since the backreference group is no
					// longer referenced, the Group
					// expression, although not its
					// contents, will be deleted when the
					// "\1*" is deleted.  After that, the
					// "x*" can be deleted.

	In the following cases, nothing can be trimmed:

	"x?\nabc"	// The "x?" is adjacent to a new-line, and therefore
			//    can extend the match to an additional line.
	"\n*abc"	// The "\n*" can match multiple new-lines, and therefore
			//    can extend the match to an additional line.
	"(x\n)?abc"	// The "(x\ny)? can match a new-line not at the end of the
			//    matched text, and can therefore extend the match
			//    to an additional line.
	"\(x*\)abc\1"	// The "x*" is inside a backreferencing Group.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
trim_null(r, e, direction)
RE_RE *r;
register EXPR *e;
DIRECTION direction;
{
	register EXPR *oldchild, *child;

	/* If the expression or its descendents do not contain a repeat count,
	 * no portion of the expression is nullable, so neither it nor any of
	 * its parts can be trimmed.
	 */
	if (!e->rept) {
		return;
	}

	/* If <e> does not contain any new-lines, we can trim repeats. */
	if (!e->nl_count) {
		/* Neither <e> nor any descendents contain a new-line,
		 * (and <e> is at one end of the pattern).  Therefore,
		 * "<e>+" ("<e>*") can be changed to <e> (<e>?),
		 * since the matches to the second and subsequent occurrences
		 * of <e> are not required, and cannot extend the match onto
		 * another line.
		 */
		if (e->type == Plus) { 
			/* Change "<e>+" to "<e>". */
			e = e->l;
			delete_rept(e->parent);
		} else if (e->type == Star) {
			/* Change "<e>*" to "<e>?". */
			change_rept(e, Quest);
		}
	}
	/* If <e> is nullable, (i.e. exactly matches the null text string),
	 * and does not contain a backreferencing group or a backreference,
	 * (excluding a Backref Star, which is trimmable)
	 * we may be able to delete it entirely.
	 */
	if (e->min_length_ex == 0 && !e->group && !e->backref_no_star) {
		/* <e> is nullable, and does not contain a backreference Group
		 * or a Backref.  We MAY be able to delete it if it does not
		 * contain a new line, except perhaps at the end we are trimming.
		 */
		if (!interior_nl(e, direction)) {
			POSITION_SET *fpos; 
			ID id;
			BOOL backward;

			/* "e" is nullable and contains no interior new-line.
			 * We can delete it entirely, unless the characters
			 * that it matches can be adjacent to a new-line in
			 * the text.  If it can be adjacent to a new-line,
			 * text matching "e" could extend a match onto an
			 * additional line, which needs to be output.
			 *
			 * Examples:   (\n represents NL)
			 *
			 *	     "a?bc":  "a?" can be deleted. 
			 *
			 * 	     "a?x\nbc:"  "a?" can be deleted.
			 *
			 * 	     "a?\nbc:"  "a?" cannot be deleted,
			 * 	     because if the text contains "a\nbc",
			 * 	     the line ending in 'a' would not get printed.
			 *
			 * 	     "a?x*\nbc:"  "a?" cannot be deleted, for
			 * 	     the same reason as in the previous example.
			 *
			 * We calculate the set of "lits" that can follow
			 * "e" in matching text, and see if any can match a
			 * new-line.
			 */

			/* Save the match direction. */
			backward = r->backward;

			/* Set the match direction to match the direction
			 * we are trimming, so that the adjacency will
			 * be calculated in the correct direction.
			 */
			r->backward = (direction == RIGHT ? YES : NO);

			/* This will fill up the POSITION_SET r->fpos
			 * with the "lits" that can be adjacent to "e"
			 * in the correct direction.
			 */
			efollow (r, e, (FOLLOW_SET *)0);

			/* Restore the match direction. */
			r->backward = backward;

			/* Step through the ID's in the position set.
			 * Break out if and when one is encountered that
			 * can match a new-line text character.
			 */
			fpos = &r->fpos;
			for (id = fpos->last; id > 0; id = fpos->base[id].previd) {
				EXPR *f = r->ptr[id];
				if (f->type == Literal && f->lit == NL) {
					break;
				}
				if (f->type == Charclass && CCL_CHK((CHAR *)f->r, NL)) {
					break;
				}
			}
			if (id < 0) {
				/* "e" is at one end of the pattern.
				 * Neither "e" nor any descendents contain a
				 * new-line.  "e" can match the null string,
				 * and any text matching "e" cannot be
				 * adjacent to a new-line.  "e" and all its
				 * descendents can be deleted.
				 */
				delete_expr(e);
				return;
			}
		}
	}

	/* <e> was not deleted, but perhaps a portion of it can be deleted
	 * or trimmed.  Step down the tree.
	 */
	switch(e->type) {
	/* We can't trim inside backreferencing entities. */
	case Group: case Backref:
		break;
	case Plus: case Star:
		/* We can't trim inside a Plus or a Star, because the
		 * entities can be matched repeatedly.  For example,
		 * we can't trim (a?bc)* to (bc)*, because "bcabc"
		 * will no longer be matched.
		 */
		break;
	case EOP:
	case Quest:
		/* Call trim_null() for the child.  Keep calling it repeatedly
		 * if, after the call, the child is a different expression.
		 */
		oldchild = 0;
		while (e->l && e->l != oldchild) {
			oldchild = e->l;
			trim_null(r, e->l, direction);
		}
		break;
	case Cat:
		/* We only process one child, because we are trimming the
		 * ends of the pattern.  If <direction> is RIGHT, we are
		 * trimming the right end of the pattern, and vice-versa.
		 *
		 * Call trim_null() for the child.  Keep calling it repeatedly
		 * if, after the call, the child is a different expression.
		 */
		oldchild = 0;
		child = ((direction == RIGHT) ? e->r : e->l);
		while (child && child != oldchild) {
			oldchild = child;
			trim_null(r, child, direction);
			child = ((direction == RIGHT) ? e->r : e->l);
		}
		break;
	case Alternate:
		/* Call trim_null() for the right child.  Keep calling it repeatedly
		 * if, after the call, the child is a different expression.
		 */
		oldchild = 0;
		while (e->r && e->r != oldchild) {
			oldchild = e->r;
			trim_null(r, e->r, direction);
		}
		/* Call trim_null() for the left child.  Keep calling it repeatedly
		 * if, after the call, the child is a different expression.
		 */
		oldchild = 0;
		while (e->l && e->l != oldchild) {
			oldchild = e->l;
			trim_null(r, e->l, direction);
		}
		break;
	default:
		/* We have an atomic "lit".  It has no children to process.  */
		break;
	}
	return;
}

/******************************************************************************

Routine:
	interior_nl()

Syntax:

	BOOL interior_nl(e, direction)

	EXPR *e;		// Pointer to expression to be checked
				//    for new-lines.
	DIRECTION direction;	// LEFT: Ignore new-line that can only match
				//   first character of text.
				// RIGHT: Ignore new-line that can only match
				//   last character of text.

Description:

	interior_nl() determines if <e> can match text containing a new-line,
	excluding a possible new-line as the first (last) character of
	the text if <direction> is LEFT (RIGHT).

Global Variables:

	None

Files:
	None

Return Status:

	YES	// The expression may match text that contains a new-line,
		//   excluding the first (last) character if direction is
		//   LEFT (RIGHT).
	NO	// The expression cannot match text that contains a new-line,
		//   except perhaps for the first (last) character if
		//   direction is LEFT (RIGHT). 

*******************************************************************************/

BOOL
interior_nl(e, direction)
EXPR *e;
DIRECTION direction;
{
	BOOL intnl;
	/* Switch on the maximum number of new-lines in the expression. */
	switch (e->nl_count) {
	case 0:
		/* <e> contains no new-lines. */
		intnl = NO;
		break;
	case 1:
		/* <e> can match an expression containing at most
		 * one new-line.
		 */
		if (e->nl_alt) {
			/* The new-line can be at either end (due to
			 * an alternation).  We may have an interior
			 * new-line.
			 */
			intnl = YES;
		} else if (e->nl_first && direction == LEFT) {
			/* The new-line is at the end we are trimming. */
			intnl = NO;
		} else if (e->nl_last && direction == RIGHT) {
			/* The new-line is at the end we are trimming. */
			intnl = NO;
		} else {
			/* The new-line is not at the end we are trimming. */
			intnl = YES;
		}
		break;
	default:
		/* <e> can match an expression containing two or
		 * more new-lines. Since at most one of these can
		 * be at the ned we are trimming, we must have an
		 * interior new-line.
		 */
		intnl = YES;
		break;
	}
	return (intnl);
}

/******************************************************************************

Routine:
	adj_compress()

Syntax:

	static void adj_compress(e, direction)

	EXPR *e;	// Pointer to expression to be compressed, if possible.
	DIRECTION direction;	// From which direction are we compressing,
				//   RIGHT or LEFT.

Description:

	During the OPTIMIZATION stage of compilation, adj_compress() is called
	to compress adjacent "lit" expressions, where possible.  adj_compress
	traverses the expression tree, and calls adj_lit_compress() when an
	atomic "lit" expression is encountered.  The actual compression is
	performed by adj_lit_compress().  See the prologue of adj_lit_compress()
	for details.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
adj_compress(e, direction)
register EXPR *e;
DIRECTION direction;
{
	/* For non-"lits" or EOP, call ourselves recursively for any children,
	 * but don't bother unless there is a Plus, Star, or Quest, here or below.
	 */
	switch(e->type) {
	case EOP:
	case Group:
		/* Attempt compression only if there is a Plus, Star, or Quest below. */
		if (e->l->rept) adj_compress(e->l, direction);
		return;
	case Plus: case Star: case Quest:
		/* Attempt compression. */
		adj_compress(e->l, direction);
		return;
	case Alternate: case Cat:
		/* Attempt compression in each branch only if there is a Plus,
		 * Star, or Quest in that branch.  For Cat, the order matters.
		 * WARNING:  We cannot assume that e->l or e->r is non-zero
		 * on the second call, because the Cat may have been removed
		 * from the tree on the first call.  When this happens,
		 * e->l and e->r will be zeroed.
		 */
		if (direction == RIGHT) {
			if (e->r->rept) adj_compress(e->r, direction);
			if (e->l && e->l->rept) adj_compress(e->l, direction);
		} else {
			if (e->l->rept) adj_compress(e->l, direction);
			if (e->r && e->r->rept) adj_compress(e->r, direction);
		}
		return;
	}
	/* We have a Literal, Charclass, Dot, or Backref.  Don't try compression
	 * unless the parent is a Plus, Star, or Quest, because nothing will
	 * be compressed otherwise.
	 */
	if (!e->parent->rept) return;

	/* We attempt to compress adjacent "lits". */
	adj_lit_compress(e, direction);
}

/******************************************************************************

Routine:
	adj_lit_compress()

Syntax:

	static void adj_lit_compress(e, direction)

	EXPR *e;	// Pointer to expression to be compressed, if possible.
	DIRECTION direction;	// From which direction are we compressing,
				//   RIGHT or LEFT.

Description:

	During the OPTIMIZATION stage of compilation, adj_lit_compress() is called
	to compress adjacent "lit" or Backref expressions, where possible.

	<e> must be a Literal, Charclass, Dot, or Backref expression,
	and <e->parent> should be a Plus, Star, or Quest expression.

	adj_lit_compress() will first find the adjacent "lit" or Backref
	expression <p> in the direction indicated, and then attempt to compress
	<e> and/or <p>.

	Case 1:	Identical expressions in same angle brackets.

		If <e> and <p> are identical "lit" expressions, and if both <e>
		and <p> are in the same angle brackets, or both are in no angle
		brackets, we can perform the following compressions
		(with compress_like()):

		"e+p+"	=> "e+p"
		"e+p*"	=> "e+"
		"e+p?"	=> "e+"
		"e*p*"	=> "e*"
		"e*p"	=> "e+"
		"e*p?"	=> "e*"
		"e*p+"	=> "p+"
		"e?p+"	=> "p+"
		"ep*"	=> "p+"
		"e?p*"	=> "p*"

		In all these cases except the first, one of the two expressions
		has been deleted, and we call adj_lit_compress() recursively
		to attempt to compress the surviving expression with the next
		following "lit" expression. 
		NOTE: Since <e> and <p> are identical, we could have deleted
		either expression.  Which expression is deleted is a matter
		of convenience only.

		After doing any possible compressions, if neither <e> nor <p>
		was deleted, we make the following order changes:

		"e+p"	=> "ep+"
		"e?p"	=> "ep?"

		Then, we call adj_lit_compress() recursively to attempt to compress
		the <p> with the following "lit" expression.

	Case 2: One-way compression.

		There are cases where one of the expressions <e> and <p> can be
		compressed, whereas the other expression cannot be.

		1) If <p> is a proper subset of <e>, <e> cannot be compressed,
		although <p> may be able to be compressed (e.g. "[a-z]+x*" is
		equivalent to "[a-z]+").

		2) If <p> is inside angle brackets, and <e> is outside angle
		brackets, <e> cannot be compressed, although <p> may be able to
		be compressed (e.g. "[a-z]+<x*>" is equivalent to "[a-z]+",
		while "<[a-z]+>x*" cannot be compressed).

		3) If <p> is inside an alternation or a nested '?', and <e>
		is not, <e> cannot be compressed, although <p> may be able to
		be compressed (e.g. "[a-z]+(x*5)?" is equivalent to "[a-z]+5?",
		while "(5[a-z]+)?x*" cannot be compressed).

		In any of these cases, we can perform the following compressions
		(with compress_unlike()) (assuming that <p> is the compressible
		expression):

		"e+p+"	=> "e+p"
		"e*p+"	=> "e*p"
		"e+p*"	=> "e+"
		"e+p?"	=> "e+"
		"e*p*"	=> "e*"
		"e*p?"	=> "e*"

		In all these cases except the first two, <p> has been deleted,
		and we call adj_lit_compress() recursively to attempt to compress
		<e> with the next following "lit" expression. 

		In the corresponding cases where <e> has been deleted,
		we call adj_lit_compress() recursively to attempt to compress
		<p> with the next PREVIOUS "lit" expression. 

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
adj_lit_compress(e, direction)
register EXPR *e;
DIRECTION direction;
{
	register EXPR *p;

	BOOL e_compressible = YES;
	BOOL p_compressible = YES;

	/* <e->parent> should always be a Plus, Star, or Quest.
	 * This check is just to be defensive.
	 */
	if (!PlusStarQuest(e->parent->type)) return;

	/* Get the adjacent "lit" to <e> in the current direction.
	 * If the adjacent lit is under a Plus, Star, or Quest,
	 * the search will stop at the Plus, Star, or Quest.
	 *
	 * If the search indicates that <e> is in a complex Alternate or Quest,
	 * (e.g. "(xyz|cde*)p", "(cde*)?p*", or "(xyz|cde*)?p*")
	 * <p> cannot be compressed, although <e> still can be.
	 * p_compressible will be set to no in these cases.
	 *
	 * If <e> is in a complex Plus or Star, (e.g. "(cde*)+p*", or "(cde*)*p*")
	 * <p> will not be found and get_next() will return null.
	 */
	p = get_next(e, direction, &p_compressible);

	if (!p) return;

	/* If we have a Plus, Star, or Quest, get the child. */
	if (PlusStarQuest(p->type)) p = p->l;

	/* If <p> is not a "lit" or a Backref, the situation is complicated,
	 * and we don't attempt to compress anything.  This will happen
	 * when the adjacent lit <k> is in a complex expression,
	 * e.g. "e*(k*qr)+", "e*(k*qr)*", "e*(k*qr)?" or "e*(k*qr|xyz)".
	 */
	if (!p->id && p->type != Backref) return; 

	/* We can't do anything with "e?p?". */
	if (e->parent->type == Quest && p->parent->type == Quest) return;

	/* If <e> and <p> are not in the same angle brackets
	 * (for approximate matching), we may not be able to compress <p>.
	 * and/or <e>.
	 */
	if (p->n_bracket != e->n_bracket) {
		if (!p->n_bracket) {
			/* <p> is not in angle brackets (and <e> is).
			 * We cannot compress <p>.
			 */
			p_compressible = NO;
		}
		if (!e->n_bracket) {
			/* <e> is not in angle brackets (and <p> is).
			 * We can't compress <e>.
			 */
			e_compressible = NO;
		} else {
			/* <e> and <p> are both in angle brackets, but
			 * in different ones.  We can't compress <e> or <p>.
			 */
			return;
		}
	}
			
	/* <p> must only match a subset of characters that <e> matches
	 * in order to compress <p>.
	 */
	if (p_compressible && !lit_contains(e, p)) {
		p_compressible = NO;
	}

	/* <e> must only match a subset of characters that <p> matches
	 * in order to compress <p>.
	 */
	if (e_compressible && !lit_contains(p, e)) {
		e_compressible = NO;
	}

	/* Exit if we can't compress either <e> or <p>. */
	if (!e_compressible && !p_compressible) return;

	/* If <e> is not compressible, compress <p> if possible, and continue
	 * compressing adjacent elements if <p> was deleted in compression.
	 *
	 * e+p+	=> e+p
	 * e+p* => e+
	 * e+p?	=> e+
	 * e*p+	=> e*p
	 * e*p* => e*
	 * e*p?	=> e*
	 */ 
	if (!e_compressible && compress_unlike(e, p)) {
		adj_lit_compress(e, direction);
		return;
	}

	/* If <p> is not compressible, compress <e> if possible, and continue
	 * compressing adjacent elements if <e> was deleted in compression.
	 *
	 * e+p+	=> ep+
	 * e*p+ => p+
	 * e?p+	=> p+
	 * e+p*	=> ep*
	 * e*p* => p*
	 * e?p*	=> p*
	 */ 
	if (!p_compressible && compress_unlike(p, e)) {
		/* We reverse direction, because <e> was BEFORE <p>
		 * in the direction we are now going.
		 */
		adj_lit_compress(p, (direction == RIGHT ? LEFT : RIGHT));
		return;
	}

	/* We are done if we can't compress both. */
	if (!p_compressible || !e_compressible) return;

	/* <p> and <e> are both compressible. */ 

	/* Compress <p> if possible, and continue compressing adjacent elements
	 * if <p> was deleted in compression.
	 *
	 * e+p* => e+
	 * e+p?	=> e+
	 * e*p* => e*
	 * e*p?	=> e*
	 * e*p  => e+
	 * e+p+ => e+p
	 */ 
	if (compress_like(e, p)) {
		adj_lit_compress(e, direction);
		return;
	}

	/* Compress <e> if possible, and continue compressing adjacent elements
	 * if <e> was deleted in compression.  In this case, we keep going
	 * in the same direction, because if <e> is deleted, we could have
	 * deleted <p> instead.  We delete <e> in these cases
	 * just because it is easier than deleting <p>.
	 *
	 * e*p+ => p+
	 * e?p+	=> p+
	 * e?p* => p*
	 */ 
	if (compress_like(p, e)) {
		adj_lit_compress(p, direction);
		return;
	}

	/* We have "e+p" or "e?p", where <e> == <p>, and <e> and <p>
	 * are in equivalent angle brackets.  Since under these
	 * conditions, "e+p" is equivalent to "ep+", we can convert
	 * "e+p" to "ep+", and similarly we can convert "e?p" to the
	 * equivalent "ep?".
	 *
	 * We "move" the '+' or '?' for two reasons:
	 *
	 * 1) It may migrate to the end of the pattern, where it may
	 * be trimmed by trim_null().
	 *
	 * 2) It may be compressible in the new location.
	 */
	{
		EXPR *pq = e->parent;
		EXPR *ep = pq->parent;
		EXPR *pp = p->parent;

		if (pp == ep) {	
			if (ep->l == pq) {
				ep->l = e;
				ep->r = pq;
			} else {
				ep->l = pq;
				ep->r =e;
			}
		} else {
			(ep->l == pq) ? (ep->l = e) : (ep->r = e);
			(pp->l == p) ?  (pp->l = pq) : (pp->r = pq); 
			pq->parent = pp; 
		}

		pq->l = p;
		e->parent = ep; 
		p->parent = pq; 

		/* Percolate the changes up the tree. */
		percolate(pq);
		percolate(ep);
	}
#ifdef DEBUG
	print_expr(R->root);
#endif
	/* Continue with possible compression of <p> with the
	 * following expression.
	 */
	adj_lit_compress(p, direction);
}

/******************************************************************************

Routine:
	compress_unlike()

Syntax:

	static BOOL compress_unlike(e, p)

	EXPR *e;	// Pointer to "lit" expression adjacent to <p>.
	EXPR *p;	// Pointer to "lit" expression to be compressed, if possible.

Description:

	During the OPTIMIZATION stage of compilation, compress_unlike() is called
	to compress adjacent "lit" expressions for those cases where only one
	expression, <p>, is compressed or deleted, while the adjacent expression
	<e> is left undisturbed.  A necessary (but not sufficient) condition for
	compress_unlike() to be called is that <e> and <p> are adjacent "lit"
	expressions where the set of characters matched by <p> is a subset
	(not necessarily a proper subset) of the set of characters matched by <e>.

	compress_unlike() performs one of the following compressions, where possible:

	"e+p+"	=> "e+p"
	"e*p+"	=> "e*p"
	"e+p*"	=> "e+"
	"e+p?"	=> "e+"
	"e*p*"	=> "e*"
	"e*p?"	=> "e*"

	If <p> is deleted in the compression (i.e., if any of the last four
	compressions are done), compress_unlike() returns YES.  Otherwise,
	compression_unlike() returns NO.

Global Variables:

	None

Files:
	None

Return Status:

	YES	// Expression <p> has been deleted.
	NO	// Expression <p> has not been deleted.

*******************************************************************************/

static BOOL
compress_unlike(e, p)
EXPR *e;
EXPR *p;
{
	BOOL p_deleted;

	/* We may be able to compress <p>, but we must leave <e> alone. */

	/* We can't compress unless <e> is under a Plus or a Star. */

	if (!PlusStar(e->parent->type)) return(NO);

	/* Switch on the type of p->parent. */
	switch (p->parent->type) {
	case  Star: case Quest:
		/*
		 * <e> is a superset of <p>.
		 *
		 * "e+p*" or "e+p?" can be simplified to "e+", or
		 * "e*p*" or "e*p?" can be simplified to "e*".
		 * 
		 * Delete "p*" or "p?".
		 */
		delete_expr(p->parent);

		p_deleted = YES;
		break;
	case Plus:
		/*
		 * <e> is a superset of <p>.
		 *
		 * "e+p+" can be simplified to "e+p", or
		 * "e*p+" can be simplified to "e*p".
		 * 
		 * Change "p+" to "p".
		 */
		delete_rept(p->parent);

		/* <p> itself has not been deleted. */
		p_deleted = NO;
		break;
	default:
		/* We have "e+p" or "e*p".  We can't compress. */
		p_deleted = NO;
		break;
	}
	return(p_deleted);
}

/******************************************************************************

Routine:
	compress_like()

Syntax:

	static BOOL compress_like(e, p)

	EXPR *e;	// Pointer to "lit" expression adjacent to <p>.
	EXPR *p;	// Pointer to "lit" expression to be compressed, if possible.

Description:

	During the OPTIMIZATION stage of compilation, compress_like() is called
	to compress adjacent "lit" expressions for those cases where both
	expressions may be compressed.   A necessary (but not sufficient)
	condition for compress_like() to be called is that <e> and <p>
	are adjacent "lit" expressions that match the identical set of characters.
	
	NOTE: When compressing adjacent expressions, unless the first call to
	compress_like() returns YES, indicating that <p> has been deleted,
	compress_like() must be called a second time with the input arguments
	reversed to perform all possible compressions.

	compress_like() performs one of the following compressions, where possible:

	"e+p+"	=> "e+p"
	"e*p+"	=> "e*p"
	"e+p*"	=> "e+"
	"e+p?"	=> "e+"
	"e*p*"	=> "e*"
	"e*p?"	=> "e*"
	"e*p?"	=> "e+"

	All except the last compression are performed by calling compress_unlike().

	If <p> is deleted in the compression (i.e., if any of the last five
	compressions are done), compress_unlike() returns YES.  Otherwise,
	compression_unlike() returns NO.

Global Variables:

	None

Files:
	None

Return Status:

	YES	// Expression <p> has been deleted.
	NO	// Expression <p> has not been deleted.

*******************************************************************************/

static BOOL
compress_like(e, p)
EXPR *e;
EXPR *p;
{
	if (e->parent->type == Star && !PlusStarQuest(p->parent->type)) {
		/* "e*p" can be simplified to "e+", since e == p. 
		 *
		 * Change the "e*" to "e+" and delete the "p".
		 */
		change_rept(e->parent, Plus);
		delete_expr(p);
		/* <p> has been deleted. */
		return(YES);
	}

	/* In all other cases, we just call compress_unlike() to do the
	 * possible compression.
	 */
	return(compress_unlike(e, p));
}

/******************************************************************************

Routine:
	lit_contains()

Syntax:

	static BOOL lit_contains(a, b)

	EXPR *a;	// Pointer to an expression.
	EXPR *b;	// Pointer to another lit" expression.

Description:

	During the OPTIMIZATION stage of compilation, lit_contains() is called
	to determine if the set of text characters that can match expression
	<b> is a subset (not necessarily a proper subset) of the set of text
	characters that can match expression <a>.  <a> and <b> must be
	Literal, Charclass, Dot, or Backref expressions.

	Examples:

	'x' is a subset of 'x'
	'x' is a subset of "[a-z]"
	'x' is a subset of '.'
	"[a-h]" is a subset of "[a-z]"
	"[a-z]" is a subset of "[a-z]"
	"[a-z]" is a subset of "."
	'.' is a subset of '.'
	"\1" is a subset of "\1"	(Backrefs)
	
	'x' is not a subset of 'y'
	'x' is not a subset of "[A-Z]"
	"[a-z]" is not a subset of "[a-h]"
	'.' is not a subset of "[a-z]"
	'.' is not a subset of 'x'
	"\1" is not a subset of "\2"	(Backrefs)
	"\1" is not a subset of 'x'
	"\1" is not a subset of "[a-z]"
	"\1" is not a subset of '.'
	'x' is not a subset of "\1"
	"[a-z]" is not a subset of "\1"
	'.' is not a subset of "\1"


Global Variables:

	None

Files:
	None

Return Status:

	YES	// Expression <b> is a subset of <a> (i.e. <a> contains <b>).
	NO	// Expression <b> is not a subset of expression <a>.

*******************************************************************************/

static BOOL
lit_contains(a, b)
register EXPR *a, *b;
{
	BOOL ret;
	switch (a->type) {
	case Charclass:
		switch (b->type) {
		case Charclass:
			/* ccl_compare will return YES if and only if
			 * Charclass <b> is a subset of Charclass <a>.
			 */
			ret = (BOOL)(ccl_compare((CHAR *)a->r, (CHAR *)b->r));
			break;
		case Literal:
			/* Literal <b> is a subset of Charclass <a>
			 * if and only if <b> is in the Charclass.
			 */
			ret = (BOOL)(CCL_CHK((CHAR *)a->r, b->lit));
			break;
		case Backref: case Dot:
		default:
			/* We can assume that the Charclass does not
			 * contain every character (except perhaps
			 * new-line).  All that happens if we're wrong
			 * is some optimization does not get done.
			 */
			ret = NO;
			break;
		}
		break;
	case Literal:
		/* Literal <b> is a subset of <a> if and only if
		 * <b> is a Literal identical to <a>.
		 */
		ret = (BOOL)(b->type == Literal && a->lit == b->lit);
		break;
	case Dot:
		switch (b->type) {
		case Charclass:
			/* Charclass <b> is a subset of Dot <a>
			 * if and only if new-line is not in the Charclass.
			 */
			ret = (BOOL)!(CCL_CHK((CHAR *)b->r, NL));
			break;
		case Literal:
			/* Literal <b> is a subset of Dot <a>
			 * if and only if the Literal is not new-line.
			 */
			ret = (BOOL)(b->lit != NL);
			break;
		case Dot:	
			/* Dot <b> is a subset of Dot <a>.
			 * All Dots are identical.
			 */
			ret = YES;
			break;
		case Backref:
		default:
			ret = NO;
			break;
		}
		break;
	case Backref:
		/* Backref <b> is a subset of <a> if and only if
		 * <b> is a Literal identical to <a>.
		 */
		ret = (BOOL)(b->type == Backref && a->lit == b->lit);
		break;
	}
	return(ret);
}

/******************************************************************************

Routine:
	ccl_compare()

Syntax:

	static BOOL ccl_compare(a, b)

	CHAR *a;	// Pointer to Charclass table.
	CHAR *b;	// Pointer to another Charclass table.

Description:

	During the OPTIMIZATION stage of compilation, ccl_compare() is called
	to determine if the set of text characters that can match Charclass
	expression <b> is a subset (not necessarily a proper subset) of the set
	of text characters that can match Charclass expression <a>.

	Examples:

	'[a-h]' is a subset of "[a-z]"
	'[a-z]' is a subset of "[a-z]"
	
	'[a-z]' is not a subset of "[a-h]"

Global Variables:

	None

Files:
	None

Return Status:

	YES	// Charclass <b> is a subset of <a> (i.e. <a> contains <b>).
	NO	// Charclass <b> is not a subset of expression <a>.

*******************************************************************************/

static BOOL
ccl_compare(a, b)
register CHAR *a, *b;
{
	/* Set a pointer to a CHAR one past the end of the CCL masks.  */
	CHAR *af = a + CCL_SIZ;

	/* If <a> and <b> point to the same table, we are done. */
	if (a == b) return(YES);

	/* Loop through the masks, one CHAR at a time. */	
	for (; a < af; a++, b++) {
		/* Do a bitwise and of <b> and the ones-complement of <a>.
		 * If the result is non-zero, there must be a bit set in <b>
		 * that is not set in <a>.  Therefore <b> is not a subset of <a>.
		 */
		if ((*b) & (~(*a))) return(NO);
	}
	/* <b> is a subset of <a> (although not necessarily a proper subset). */
	return(YES);
}

/******************************************************************************

Routine:
	matches_every_line()

Syntax:

	static BOOL matches_every_line(pat, r)

	PATTERN *pat;	// Pointer to pattern to be matched. 
	RE_RE *r;	// Pointer to regular expression being optimized

Description:

	Determine if the pattern will match every line of text.  This
	can happen in either of two ways (also taking into account
	approximate matching).

	1) The pattern matches the null string, e.g., a pattern "a*b*".
	2) The pattern matches a single new-line character, e.g, a
		pattern "a*\nb*"

Global Variables:

	None

Files:
	None

Return Status:

	YES	// Pattern will match every line of text, and therefore
		//	has been converted to a single new-line, with
		//	exact matching.

	NO	// Pattern will not match every line of text.

*******************************************************************************/

static BOOL
matches_every_line(pat, r)
PATTERN *pat;
RE_RE *r;
{
	/* If <pbounds> is set for this pattern, we need to find the exact
	 * length of the match.  Therefore, do not do any replacement.
	 */
	if (pat->pbounds) {
		return(NO);
	}
	/* Check if the pattern matches the null text string, taking 
	 * into account possible deletions or substitutions due to approximate
	 * matching).
	 */
	if (r->root->min_length == 0) {

		/* The pattern matches the null string.  It will match EVERY line
		 * of text.  As an optimization, we delete the current pattern and
		 * replace it with a single new-line character with exact matching.
		 */
		free_expr(r->root);
		pat->approx = NO;
		pat->cost_limit = r->cost_limit = 0;
		pat->cost_ins = r->cost_ins = 1;
		pat->cost_del = r->cost_del = 1;
		pat->cost_sub = r->cost_sub = 1;
		r->maxid = 1;
		r->root = newexpr(r, Literal, '\n', (EXPR *)0, (EXPR *)0);
		r->root = newexpr(r, EOP, '#', r->root, (EXPR *)0);
		return(YES);
	}
	return(NO);
}

/******************************************************************************

Routine:
	free_expr()

Syntax:

	static void free_expr(e)

	EXPR *e;	// Pointer to expression to be freed.

Description:

	free_expr() is invoked during the OPTIMIZATION stage of the compilation
	after an expression and its descendents are removed from the expression
	tree.  free_expr() frees the memory for expression <e> and its descendents
	from the bottom up, by calling itself recursively for each child of the
	specified expression.

	When a Backref expression is freed, the count of references for
	the associated backreferencing Group is decremented.  If the count
	should reach zero, delete_rept() is called to delete the Group
	expression only.  Descendents of the Group expression are unaffected.

Global Variables:

	Group_refs

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
free_expr(e)
EXPR *e;
{
	/* Backref and Charclass need special attention, because e->l
	 * or e->r is set, even though there are no descendents.
	 */
	switch (e->type) {
	case Backref:
		/* Decrement the reference count for the associated group.
		 * If the reference count becomes zero, delete the Group
		 * expression for the corresponding group, but not its
		 * descendents.
		 */
		if (--Group_refs[e->lit - '0'] == 0) {
			delete_rept(e->l);
		};
		/* Zero the pointer to the corresponding Group
		 * so it and its descendents are not deleted.
		 */
		e->l = 0;
		break;
	case Charclass:
		/* Don't free the table, because it may be shared,
		 * but zero the pointer to it.
		 */
		e->r = 0;
		break;
	}

	/* Free the descendent expressions from the bottom up. */
	if (e->r) free_expr(e->r);
	if (e->l) free_expr(e->l);

	/* Free this expression. */
	FREE(e);
}

/******************************************************************************

Routine:
	delete_expr()

Syntax:

	static void delete_expr(e)

	EXPR *e;	// Pointer to expression to be deleted.

Description:

	delete_expr() is invoked during the OPTIMIZATION stage of the compilation
	to remove expression <e> and its descendents from the expression tree.
	If <e->parent> is a Plus, Star, or Quest expression, the parent
	is removed also, by calling delete_expr(e->parent).  If <e->parent>
	is an Alternate or a Cat, the Alternate or Cat expression is
	pruned from the tree, as there is now one less member of the
	alternation or concatenation.  If <e->parent> is anything other
	than a Plus, Star, Quest, Alternate, or Cat, it is a fatal error.

	After the Alternate or Cat is pruned from the tree, delete_expr()
	calls percolate() to recalculate the expression attributes
	for those remaining expressions that had descendents pruned from the tree.
	Then, delete_expr() calls free_expr() to free the memory for the pruned
	expressions, except that the memory for the Alternate or Cat expression
	itself is not freed.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
delete_expr(e)
register EXPR *e;
{
	register EXPR *parent = e->parent;
	register EXPR *grandparent, *sibling;

	/* If the expression has no parent, it is a fatal error. */
	if (!parent) {
		exitmsg("%s: INTERNAL ERROR: delete_expr abort %d\n", EOP);
	}

	/* Check the type of the parent. */

	switch(parent->type) {
	case Plus: case Star: case Quest: case Alternate:
		/* The parent is a repeat expression or an alternation.
		 * Delete the parent and it's descendents.
		 */
		delete_expr(parent);
		return;
	case Cat:
		/* The parent is a concatenation.
		 * We can delete "e" from the concatenation.
		 */
		break;
	default:
		/* The parent is not a repeat expression, nor an alternation
		 * or a concatenation.  This is a fatal internal error.
		 */
		exitmsg("%s: INTERNAL ERROR: delete_expr abort %d\n", parent->type);
	}
#ifdef DEBUG
	print_expr(R->root);
#endif
	/* The parent of "e" is a Cat.  Remove this parent
	 * from the tree.  We first get the parent of the parent, i.e., the
	 * grandparent of "e".
	 */
	grandparent = parent->parent;

	/* Now, get the sibling of "e".  */
	if (e == parent->r) {
		sibling = parent->l;
	} else {
		sibling = parent->r;
	}

	/* The grandparent is the new parent of the sibling. */
	sibling->parent = grandparent;

	/* The sibling is a new child of the grandparent (replacing the parent). */
	if (parent == grandparent->r) grandparent->r = sibling;
	else grandparent->l = sibling;

	/* Recalculate all the quantities in the grandparent and its ancestors
	 * that may have changed due to this deletion.
	 */
	percolate(grandparent);

	/* Free the memory for the expression and its descendents. */
	/* Careful!!  Freeing nodes of a tree while traversing the tree
	 * recursively is dangerous.  However, we can delete expressions
	 * with zero or one child safely, because we only traverse down
	 * one branch.  When a node has two children, like a Cat,
	 * adjacency compression may delete one child while we were
	 * traversing the other child.  When we return to the Cat,
	 * we may attempt to traverse the deleted child.
	 * We solve this problem by NOT freeing the parent Cat (or Alternate),
	 * but rather setting both its children to 0.  adj_compress()
	 * must check for this.
	 */
	parent->r = parent->l = 0;
	free_expr(e);

	/* In the (unlikely) event that the grandparent and the sibling (now
	 * parent and child) are both repeat counts, they can be compressed.
	 */
	rept_compress(grandparent, sibling);
#ifdef DEBUG
	print_expr(R->root);
#endif
}

/******************************************************************************

Routine:
	change_rept()

Syntax:

	static void change_rept(e, type)

	EXPR *e;	// Pointer to expression to be modified.
	EXPR_TYPE type	// New type of expression (Plus, Star, or Quest)1

Description:

	change_rept() is invoked during the OPTIMIZATION stage of the compilation
	when the type of a Rept expression (a Plus, Star, or Quest expression)
	must be changed to a different type of Rept expression.  <e> must
	be a Rept expression.  There is no check for this.  The new type
	must also be a Rept type.  There is also no check for this.

	After the type of <e> is modified, change_rept() calls percolate() to
	recalculate expression attributes for <e> and its ancestors.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
change_rept(e, type)
register EXPR *e;
register EXPR_TYPE type;
{
#ifdef DEBUG
	print_expr(R->root);
#endif
	e->type = type;

	/* Recalculate all the quantities in the expression and its ancestors
	 * that may have changed due to this type change.
	 */
	percolate(e);
#ifdef DEBUG
	print_expr(R->root);
#endif
}

/******************************************************************************

Routine:
	delete_rept()

Syntax:

	static void delete_rept(e)

	EXPR *e;	// Pointer to expression to be deleted.

Description:

	change_rept() is invoked during the OPTIMIZATION stage of the compilation
	when a Rept expression (a Plus, Star, or Quest expression) or a Group
	expression must be deleted, while keeping the descendents of the deleted
	expression in the tree.  The parent of the deleted expression and the child
	of the deleted expression become parent and child.

	After <e> is deleted, delete_rept() calls percolate() to recalculate
	expression attributes for <e> and its ancestors.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
delete_rept(e)
register EXPR *e;
{
#ifdef DEBUG
	print_expr(R->root);
#endif
	/* Our parent becomes the parent of our child. */
	e->l->parent = e->parent;

	/* Our child replaces us as a child of our parent. */
	if (e->parent->l == e) e->parent->l = e->l;
	else e->parent->r = e->l;

	/* Recalculate all the quantities in the grandparent and its ancestors
	 * that may have changed due to this deletion.
	 */
	percolate(e->parent);

	/* If the new parent and child are both Rept expressions (very
	 * unlikely), compress them.
	 */
	rept_compress (e->parent, e->l);

	/* Free the memory for the expression. */
	FREE(e);
#ifdef DEBUG
	print_expr(R->root);
#endif
}

/******************************************************************************

Routine:
	compress_rept()

Syntax:

	static void compress_rept(parent, child)

	EXPR *parent;	// Pointer to parent expression.
	EXPR *child;	// Pointer to an expression that is a child of <parent>.

Description:

	compress_rept() is invoked during the OPTIMIZATION stage of the compilation
	to compress two Rept expressions that are parent and child into
	a single equivalent Rept expression.

	compress_rept() first checks that the specified expressions are both
	Rept expressions.  If not, compress_rept() exits without doing anything.
	compress_rept() does not check that the expressions are parent and child.

	The compressions are:

	e++ => e+
	e+* => e*
	e+* => e*
	e*+ => e*
	e** => e*
	e*? => e*
	e?+ => e*
	e?* => e*
	e?? => e?

	The <child> is deleted by a call to delete_rept().  However,
	if the types of the parent and child are different, the type
	of the parent is first changed to Star. 

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
rept_compress(parent, child)
register EXPR *parent, *child;
{
	/* Check that parent and child are both Rept expressions.
	 * If not, do nothing.
	 */
	switch(parent->type) {
	case Plus: case Star: case Quest: break;
	default: return;
	}
	switch(child->type) {
	case Plus: case Star: case Quest: break;
	default: return;
	}

	/* We consolidate the expressions into the parent, and delete the child.
	 * If the types are different, the consolidated type will be a Star.
	 * Otherwise, the type of the parent remains unchanged.
	 */
	if (parent->type != child->type && parent->type != Star) {
		change_rept(parent, Star);
	}

	/* Remove the child from the tree. */

	delete_rept(child);
}

/******************************************************************************

Routine:
	percolate()

Syntax:

	static void percolate(e)

	EXPR *e;	// Pointer to expression that has been modified,
			//   or had a descendent modified or deleted.

Description:

	percolate() is invoked during the OPTIMIZATION stage of the compilation
	when an expression is modified or deleted.  When an expression is modified,
	percolate() is called for that expression.  When an expression is deleted
	from the tree, percolate is called for the nearest ancestor that is still
	in the tree.

	percolate() first calls fill_expr(e) to recalculate the expression
	attributes of <e> that are filled in by fill_expr(), in case any of
	them have changed.  percolate() then calls itself recursively for
	<e->parent> to propagate the changes upwards in the tree.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
percolate(e)
EXPR *e;
{
	/* Recalculate these quantities in <e> */
	fill_expr(e);

	/* Recursively percolate any changes upward to ancestors of "e". */
	if (e->parent) percolate(e->parent);
}

/******************************************************************************

Routine:
	get_next()

Syntax:

	static EXPR *get_next(e, direction, no_quest_or_alt)

	EXPR *e;		// Pointer to current expression
	DIRECTION direction;	// Which end of the pattern are we traversing
				//   from, RIGHT or LEFT.
	BOOL *no_alt_or_quest;  // If non-null, will be set to NO if we have
				//    traversed out of an alternation or
				//    traversed out of a concatenation.

Description:

	get_next() is called during the OPTIMIZATION stage to return the
	expression "adjacent" to <e>, traversing the pattern FROM the
	direction indicated.

	The adjacent expression returned (if any) will never be a Cat,
	and will always have a parent that is a Cat.

	The search algorithm is as follows:

	0) If no_alt_or_quest is not null, initialize *no_alt_or_quest to YES;
	Go to step 1.

	1) Search up the tree, beginning at <e>, until an expression
	whose parent is other than a "Rept" expression is found.
	Reset <e> to this expression.  Go to step 2.

	2) If the parent of <e> is not a Cat, go to step 3.
	If the parent of <e> is a Cat, traverse the concatenation
	to find the sibling of <e> in the direction we are traversing.
	We traverse the entire concatenation if necessary, not just
	a single Cat expression.  If <direction> is RIGHT (LEFT),
	a sibling will always be found unless <e> is the leftmost (rightmost)
	member of the concatenation.  If a sibling is found, return a pointer
	to the sibling.  If a sibling is not found, set <e> to the highest Cat
	expression in the concatenation, and go to Step 3.

	3) Set <e> to its (non-Cat) parent.  If <e> is an Alternate or a Quest,
	set *no_alt_or_quest to NO (unless no_alt_or_quest is NULL), and go to
	Step 2.  If <e> is any other type of expression, give up and return 0.

	In the following examples, we assume <direction> is LEFT, i.e., we are
	traversing from left to right.  We also assume that <e> was a pointer to
	'a' or to the '+', '*', or '?' just above 'a'.

	Examples:

	Expression:	<e>		Returns pointer to:	null_alt_or_quest

	"yza*x*cde"	'a' or "a*"	"x*"				YES
	"yzax*cde"	'a'		"x*"				YES
	"yza*xcde"	'a' or "a*"	"x"				YES
	"yzaxcde"	'a'		"x"				YES
	"yza"		'a'		returns 0
	"yza*(xcde)*"	'a' or "a*"	"(xcde)*"			YES
	"(za+)?x+c"	'a' or "a+"	"x+"				NO
	"(za+|rst)x+c"	'a' or "a+"	"x+"				NO
	"(a+|rst)x+c"	'a' or "a+"	"x+"				NO
	"(za+)*x+c"	'a' or "a+"	returns 0
	"(za+)+x+c"	'a' or "a+"	returns 0

Global Variables:

	None

Files:
	None

Return Status:

	Returns a pointer to the adjacent expression in the concatenation,
	traversing from the direction indicated.

	Returns 0 if there is no adjacent expression, based on the
	restrictions of the search.

*******************************************************************************/

static EXPR *
get_next(e, direction, no_alt_or_quest)
register EXPR *e;
DIRECTION direction;
BOOL *no_alt_or_quest;
{
	if (no_alt_or_quest) *no_alt_or_quest = YES;

	/* While the parent of "e" is a Plus, Star, or Quest, we move up to the
	 * parent.  We don't stop at anything when moving up the tree.
	 */
	while (PlusStarQuest(e->parent->type)) {
		EXPR *child = e;
		e = e->parent;
		/* Just in case we have two adjacent Rept expressions,
		 * compress them.  This is unlikely here, but we
		 * will be defensive.  If <child> is also a Rept
		 * expression, it will be removed by this compression.
		 */
		rept_compress(e, child);
	}

	/* Loop until we find the expression we want, or determine that
	 * it does not exist.
	 */
	while (1) {
		/* We loop upwards in the tree, as long as the parent of "e"
		 * is a Cat, and we haven't found our desired expression.
		 */
		for (; e->parent && e->parent->type == Cat; e = e->parent) {
			/* If we are traveling FROM the left, and "e" is
			 * the left child of the parent, our desired
			 * expression is the leftmost non-Cat expression
			 * in the right child of the parent.
			 */
			if (direction == LEFT && e->parent->l == e) {
				e = e->parent->r;
				while (e->type == Cat) e = e->l;
				return(e);
			}
			/* If we are traveling FROM the right, and "e" is
			 * the right child of the parent, our desired
			 * expression is the rightmost non-Cat expression
			 * in the left child of the parent.
			 */
			if (direction == RIGHT && e->parent->r == e) {
				e = e->parent->l;
				while (e->type == Cat) e = e->r;
				return(e);
			}
			/* If we are traveling FROM the left, and "e" is
			 * the right child of the parent, or if we are
			 * traveling from the right and "e" is the left
			 * child of the parent, we go further up the tree
			 * if we can.  Keep looping.
			 */
		}
		/* We haven't found our expression yet.  "e" is now an
		 * expression whose parent is not a Cat.  Set "e" to
		 * its (non-Cat) parent.
		 */
		e = e->parent;

		/* We give up, and will return 0, in the following cases:
		 *
		 * 1) We have reached the root of the expression (type EOP).
		 *
		 * 2) We have reached a Plus, a Star, or a backreference
		 * Group.  This means that the original <e> was inside a
		 * Plus, Star, or Group, which our search is now leaving.
		 * Since we cannot compress a character inside a Plus,
		 * Star, or Group with a character outside the Plus, Star,
		 * or Group, give up and return 0.
		 */
		if (!e || e->type == EOP || PlusStar(e->type) || e->type == Group) {
			break;
		}
		/* If we are inside a Quest or an Alternate, we can
		 * be compressed, but adjacent items cannot be.  Inform
		 * the caller, if the caller supplied a pointer to store
		 * the info in.
		 */
		if (e->type == Alternate || e->type == Quest) {
			if (no_alt_or_quest) *no_alt_or_quest = NO;
		}
	}
	return((EXPR *)0);
}

/******************************************************************************

Routine:
	null_alter()

Syntax:

	static void null_alter(r, e, null_alt)

	RE_RE *r;	// Pointer to regular expression being optimized

	EXPR *e;	// Pointer to an expression in the expression tree
			//    When called externally, <e> should be the
			//    root expression (i.e. r->root).

	BOOL null_alt;	// Are we currently in a null alternation?
			//    When called externally, <null_alt> should be NO.

Description:

	null_alter() is called by optimize_exp() during the OPTIMIZATION stage
	of compilation, before any trimming or compression is done.
	null_alter() changes any "nullable alternations" to "standard form",
	to facilitate the trimming and compression of expressions.

	A nullable alternation is an alternation that can match a null text string.
	For purposes of this definition, if the alternation is directly under
	a Rept expression, the Rept expression is considered part of the alternation.

	There is an ambiguity in how nullable alternations are expressed in that
	the same nullable alternation can be specified in different ways.

	In the following examples, 'a' and 'b' can represent any simple or
	composite non-nullable expression.

	For example, the following nullable alternations are all equivalent:

	"(a?|b)" "(a|b?)" "(a?|b?)" "(a|b)?" "(a?|b)?" "(a|b?)?" "(a?|b?)?"

	Also, for example, the following nullable alternations are equivalent:

	"(a*|b)" "(a+|b?)" "(a*|b?)" "(a+|b)?" "(a*|b)?" "(a+|b?)?" "(a*|b?)?"

	Also, for example, the following nullable alternations are equivalent:

	"(a?|b)+" "(a|b?)+" "(a?|b?)+" "(a|b)*" "(a?|b)*" "(a|b?)*" "(a?|b?)*"

	Also, for example, the following nullable alternations are equivalent:

	"(a*|b)+" "(a+|b?)+" "(a*|b?)+" "(a+|b)*" "(a*|b)*" "(a+|b?)*" "(a*|b?)*"

	We choose as the "standard form" of these nullable alternations that
	form that has each member of the alternation nullable, with no
	nullable Rept expression for the entire alternation.  The standard
	forms in the above examples are "(a?|b?)", "(a*|b?)", "(a?|b?)+",
	and "(a*|b?)+" respectively.

	The "standard form" facilitates deleting members of an alternation,
	and avoids deletion anomalies that could arise if the alternations
	were stored in a form where one member is nullable and the other member
	is not (e.g. "(a|b?)").  In this case, if "b?" is deleted from the
	alternation, we would be left with "a", which is incorrect, rather than
	"a?", which is correct.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
null_alter(r, e, null_alt)
RE_RE *r;
register EXPR *e;
BOOL null_alt;
{
	register EXPR *parent;
	register EXPR *quest;

	if (null_alt) {
		switch(e->type) {
		case Plus:
			change_rept(e, Star);
			break;
		case Alternate: case EOP: case Quest: case Star:
			break;
		default:
			parent = e->parent;
			quest = newexpr(r, Quest, '\0', e, (EXPR *)0);
			quest->parent = parent;
			(e == parent->l) ?
			 (parent->l = quest) : (parent->r = quest);
			break;
		}
	}
	switch(e->type) {
	case EOP: case Group:
		/* Don't bother unless there is a repeat count and an internal
		 * alternation below.
		 */
		if (e->rept && e->alter) null_alter(r, e->l, NO);
		return;
	case Plus: case Star: case Quest:
		/* Continue down the tree */
		null_alter(r, e->l, NO);
		return;
	case Cat:
		/* Don't bother unless there is a repeat count and an internal
		 * alternation below.
		 */
		if (e->rept && e->alter) {
			null_alter(r, e->r, NO);
			null_alter(r, e->l, NO);
		}
		return;
	case Alternate:
		/* If the alternation is nullable, we may need special processing.
		 * The special processing is done when we enter the n-fold
		 * alternation.  So, if our parent is an Alternate, we've
		 * already done the special processing.
		 */
		if (e->parent->type != Alternate) {
			/* We have just entered an alternation.  Check whether
			 * it is nullable.  If our parent is a Quest, it is
			 * considered part of the alternation.  (Our parent
			 * cannot be a Plus or Star, because we don't step
			 * inside Plus or Star.)
			 */
			null_alt = ((StarQuest(e->parent->type) || e->min_length_ex == 0) ? YES : NO); 
			if (e->parent->type == Quest) {
				delete_rept(e->parent);
			} else if (e->parent->type == Star) {
				change_rept(e->parent, Plus);
			}
		}
		null_alter(r, e->r, null_alt);
		null_alter(r, e->l, null_alt);
		return;
	}
}

#ifdef DEBUG

static void
print_expr(e)
EXPR *e;
{
	switch(e->type) {
	case EOP:
		print_expr(e->l); printf("#\n"); return;
	case Plus:
		print_expr(e->l); printf("+"); return;
	case Star:
		print_expr(e->l); printf("*"); return;
	case Quest:
		print_expr(e->l); printf("?"); return;
	case Cat:
		if (PlusStarQuest(e->parent->type)) printf("(");
		print_expr(e->l); print_expr(e->r);
		if (PlusStarQuest(e->parent->type)) printf(")");
		return;
	case Alternate:
		if (PlusStarQuest(e->parent->type)) printf("(");
		if (e->parent->type == Cat) printf("(");
		print_expr(e->l); printf("|"); print_expr(e->r);
		if (PlusStarQuest(e->parent->type)) printf(")");
		if (e->parent->type == Cat) printf(")");
		return;
	case Group:
		printf("(");
		print_expr(e->l);
		printf(")");
		return;
	case Backref:
		printf("%c%c", '\\', e->lit); return;
	case Dot:
		printf("."); return;
	case Charclass:
		printf("[]"); return;
	case Literal:
		switch (e->lit) {
		case '\n':
			printf("\\n", e->lit); return;
			break;
		case '\v':
			printf("\\v", e->lit); return;
			break;
		case '\f':
			printf("\\f", e->lit); return;
			break;
		case '\t':
			printf("\\t", e->lit); return;
			break;
		case '\r':
			printf("\\r", e->lit); return;
			break;
		default:
			printf("%c", e->lit); return;
			break;
		}
	}
}

#endif
