	static char sccsid[]="%Z%	recw.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	recw.c [%P%]

Description:

	Takes a compiled regular expression, and determines whether or not
	it can be fully or partially matched as a set of fixed strings.

	If so, it sets up the fixed strings for a match.

Functions:
	re_recw()
	altlist()
	traverse()
	patstore()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

static int altlist();
static void traverse();
static void patstore();

static PATTERN *Pat;
static CHAR **Pats;
static int Npat;

#define	MAXLIT	256	/* is plenty big enough */
static CHAR Temp[MAXLIT], Best[MAXLIT];
static CHAR *Tempc;
static int Bestlen;
static unsigned short Bracket_no;

/* Starts the scan of a potential fixed string. */
#define	START(n)	{ Tempc = Temp; Bracket_no = n; }

/* Adds a character to a fixed string. */
#define	ADD(c)	{ if (Tempc < &Temp[MAXLIT]) *Tempc++ = c; }

/* If the fixed string is the longest yet, store it. */
#define	FINISH	{ if ((Tempc-Temp) > Bestlen) (void)memcpy(Best, Temp, Bestlen = Tempc-Temp); }

/* FINISH the current string, and START a new one.  Also sets Pat->succfn
 * to indicate that the longest fixed string is not equivalent to the
 * entire regular expression.
 */
#define	RESTART(n)	{FINISH; Pat->succfn = re_reexec; START(n); }

/******************************************************************************

Routine:
	re_recw()

Syntax:

	int re_recw(pat, pats)

	PATTERN	pat;	// Pointer to pattern to be matched
	CHAR **pats;	// If 0, store patterns in Commentz-Walter structure
			// If non-zero, store patterns in pointer array
			//    beginning at <pats> 

Description:

	re_recw() is called by egrep_compile() after a regular expression
	has been compiled, and pat->re_ptr points to the compiled regular
	expression.

	re_recw() determines the best set of fixed strings can be found
	that form a subset of the regular expression, and that will match
	a superset of lines of text that match the regular expression.
	For example, if the regular expression is "abcd*e|fghij",  the fixed
	strings "abc" and "fghij" will find a superset of matches to the regular
	expression.

	If no such set of strings is possible, re_recw() returns 0.
	If such a set exists, re_recw() returns the length of the shortest
	string in the set (e.g. 3, in the above example).  Furthermore,
	the set of strings will have been stored in a Commentz-Walter structure,
	or in a pointer array beginning at <pats>.  Furthermore, if the strings
	are entirely equivalent to the regular expression, pat->succfn will
	be set to 0.  If the strings are not entirely equivalent to the
	regular expression, pat->succfn will be set to re_reexec, to indicate
	that a subsequent stage of matching tot he full regular expression
	is needed.

Global Variables:

	Pat		// Set to "pat" argument
	Pats		// Set to "pats" argument
	Npat		// Initialized to 0.

Files:
	None

Return Status:

	0	// No match to a set of fixed strings possible.
	>0	// Match to a set of fixed strings is possible.
		//   The value returned is the length of the shortest string
		//   in the set.

*******************************************************************************/

int
re_recw(pat, pats)
register PATTERN *pat;
CHAR **pats;
{
	/* Store the arguments as file scoped statics. */
	Pat = pat;
	Pats = pats;
	Npat = 0;

	/* Initialize the Commentz-Walter structures. */
	if (!pats) pat->cw_ptr = re_cwinit(pat);

	/* Initialize pat->succfn to 0. */
	pat->succfn = 0;

	/* Call the recursive function altlist() to do the work.
	 * Pass in the child of the root expression.
	 */
	return(altlist(pat->re_ptr->root->l));
}

/******************************************************************************

Routine:
	altlist()

Syntax:

	static int altlist(e)

	EXPR *e;	// Pointer to expression to be processed.

Description:

	altlist() is called by re_recw() to do the actual work of determining
	the set of fixed strings that can duplicate or approximate the regular
	expression.

	If "e" is an alternation, altlist() calls itself recursively for each
	of the two children of the alternation, and returns the minimum of the
	two return values.  However, if the return value from the first is zero,
	altlist() knows that no set of fixed strings is possible, so the 
	call for the second child is not needed.

	If "e" is not an alternation, altlist() calls traverse() to recursively
	traverse the expression, and returns the length of the longest fixed
	string that guarantees a match to the expression.

Global Variables:

	Pat		// Pointer to pattern to be matched.
	Pats		// Pointer array in which to store strings.
			//    If 0, store in Commentz-Walter structure.
	Temp[]		// Temporary storage for string
	Best[]		// Storage for current longest string
	Tempc		// Pointer to character being processed
	Bestlen		// Length of longest string.
	Bracket_no	// Identifier of angle bracket surrounding character
			//    (used for approximate matching).

Files:
	None

Return Status:

	0	// No match to a set of fixed strings possible.
	>0	// Match to a set of fixed strings is possible.

*******************************************************************************/

static int
altlist(e)
register EXPR *e;
{
	int l_best, r_best;
	CHAR *lb, *le;

	/* If we have an alternation, we return the minimum string length
	 * from each of the two alternatives.  If the first is zero, we
	 * need not check the other.
	 */
	if (e->type == Alternate) {
		l_best = altlist(e->l);
		if (l_best == 0) return(0);
		r_best = altlist(e->r);
		return(MIN(l_best, r_best));
	}
	/* We have an expression other than an alternation. */

	/* Initialize the best length to zero. */
	Bestlen = 0;

	/* Initialize the a traversal of a potential string. */
	START(0);

	/* Do the traversal */
	traverse(e);

	/* Finish off the traversal of the last string. */
	FINISH;

	/* Bestlen is the length of the fixed string.
	 * If there is no non-zero length string, return 0.
	 */
	if (Bestlen == 0) return(0);

	/* Allocate memory for the longest string found, plus an extra
	 * character for the null that FINISH put after it.
	 */
	lb = cgmalloc2((unsigned int)Bestlen, sizeof(CHAR), "Commentz-Walter string");

	/* Copy the longest string to the allocated memory. */
	(void)memcpy(lb, Best, Bestlen);

	/* Set le to be one character past the end. */
	le = lb + Bestlen;

	/* Store the string in a pointer array or in a Commentz-Walter structure. */
	if (Pats) {
		patstore(lb, le);
	} else {
		re_cwadd(Pat->cw_ptr, lb, le);
	}

	/* Return the string length. */
	return(Bestlen);
}

/******************************************************************************

Routine:
	traverse()

Syntax:

	static void traverse(e)

	EXPR *e;	// Pointer to expression to be processed.

Description:

	traverse() is called by altlist() to determine the longest fixed string
	that can duplicate or approximate one regular expression.

Global Variables:

	Pat		// Pointer to pattern to be matched.
	Temp[]		// Temporary storage for string
	Best[]		// Storage for current longest string
	Tempc		// Pointer to character being processed
	Bestlen		// Length of longest string, plus 1.
	Bracket_no	// Identifier of angle bracket surrounding character

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
traverse(e)
register EXPR *e;
{
	/* We are accumulating a fixed string of consecutive literal
	 * characters.
	 */
	switch(e->type) {
	case Literal:
		/* We have a literal character. */ 
		if (!Pat->approx) {
			/* We are not approximate matching.
			 * We can add the character to our current string.
			 */
			ADD(e->lit);
		} else if (e->n_bracket == 0) {
			/* We are approximate matching, and this character
			 * is outside an angle bracket.  This character
			 * therefore need not be matched.  Finish the current
			 * string, and start a new one after this character.
			 */
			RESTART(0);
		} else if (Bracket_no == 0) {
			/* We are approximate matching, this character is
			 * inside an angle bracket, and either the previous
			 * character was outside an angle bracket or there
			 * was no previous character.  We can add the character
			 * to our current (empty) string.
			 */
			Bracket_no = e->n_bracket;
			ADD(e->lit);
		} else if (e->n_bracket == Bracket_no) {
			/* We are approximate matching, and this character
			 * is within the same angle bracket as the last one.
			 * We can add it to our current string.
			 */
			ADD(e->lit);
		} else {
			/* We are approximate matching and this character is
			 * within the a different angle bracket than the last
			 * character.  We can't add this  character to our
			 * current string, but it can begin a new string.
			 */
			RESTART(e->n_bracket);
			ADD(e->lit);
		}
		break;
	case Cat:
		/* We have a concatenation.  Traverse the children in the
		 * proper order.
		 */
		if (Pat->backward) {
			traverse(e->r);
			traverse(e->l);
		} else {
			traverse(e->l);
			traverse(e->r);
		}
		break;
	case Plus:
		/* We have a '+'.  The expression(s) underneath must be
		 * matched one or more times.  So we can possibly end the
		 * current string with the expressions underneath, and
		 * possibly begin the new string with the same expressions.
		 */
		traverse(e->l);
		RESTART(0);
		traverse(e->l);
		break;
	case Group:
		/* We have a group for backreferencing.  Traverse the group. */
		traverse(e->l);
		break;
	case Star:
	case Quest:
	case Alternate:
	case Dot:
	case Charclass:
	case Backref:
	default:
		/* We either have something that need not be matched (Star or
		 * Quest), or something that can match more than one literal
		 * character (Alternate, Dot, Charclass, or Backref).
		 * Finish the current string, and start another potential
		 * one after this.
		 */
		RESTART(0);
		break;
	}
}

/******************************************************************************

Routine:
	patstore()

Syntax:
	static void patstore(s, e)

	CHAR *s;	// Pointer to first character of string.
	CHAR *e;	// Pointer to one past last character of string.

Description:

	Adds one string to pointer array Pats.

	NOTE:  If this structure is to be used for a match in the
	backward direction, each string must be reversed before being
	passed into this routine.  "s" must point to the first character
	of the reversed string, and "le" must point one character past
	the end of the reversed string.

Global Variables:

	Pats	// Pointer array for storage.
	Npat	// Array element count

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
patstore(s, e)
register CHAR *s;
register CHAR *e;
{

	
	/* Store the string itself. */
	Pats[Npat] = (CHAR *)cgmalloc((unsigned int)(e + 1 - s), "pats");
	(void)memcpy((char *)Pats[Npat], (char *)s, e - s);
	Pats[Npat][e-s] = '\0';

	Npat++;
}
