	static char sccsid[]="%Z%	restate.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	re_state.c [%P%]

Description:

	Contains low level routines used to access and manage state information
	for the finite stat machine used for matching regular expressions.

Functions:
	eg_stateinit()
	eg_savestate()
	eg_clrstates()
	eg_getstates()
	eg_posalloc()

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include	"re.h"

#ifdef __MSDOS__
#define		STATESTART	63
#define		STATELIM	STATESTART
#define		POSSTEP		8191
#define		POSLIMIT	POSSTEP
#else
#define		STATESTART	128
#define		STATELIM	512
#define		POSSTEP		8192
#define		POSLIMIT	(64*8192)
#endif

static void eg_clrstates();

/******************************************************************************

Routine:
	eg_stateinit()

Syntax:

	void eg_stateinit();

	RE_RE *r;	// Pointer to regular expression to be matched.

Description:

	Called by re_reinit(). 
	Allocates memory for states used in matching a regular expression,
	and initializes related quaintities.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
eg_stateinit(r)
RE_RE *r;
{
	/* Number of states that we will initially allocate. */
	r->statelim = STATESTART;

	/* Number of times we flush state buffers before we enlarge the buffers. */
	r->threshold = 1;

	/* Allocate memory for the STATE structures */
	r->states = (STATE *)cgmalloc2(r->statelim, sizeof(STATE), "states");

	/* Number of FID's for which we will allocate memory.  A set of
	 * FID's are associated with each state.
	 */
	if (r->nposalloc == 0) r->nposalloc = POSSTEP;

	/* Allocate memory for the FID structures. */
	r->posbase = (FID *)cgmalloc2(r->nposalloc, sizeof(FID), "posbase");

	/* Initialize the flush count, the number of states in memory,
	 * and the number of FID's in memory to zero.
	 */
	r->flushed = r->nstates = r->posnext = 0;
}

/******************************************************************************

Routine:
	eg_savestate()

Syntax:

	void eg_savestate();

	RE_RE *r;	// Pointer to regular expression to be matched.

Description:

	Called by re_reinit() after the initial state and its associated
	FID's have been stored.  Ensures that the FID's associated with the
	initial state will not be flushed when the states are flushed.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

void
eg_savestate(r)
RE_RE *r;
{
	/* r->posreset is set to point to the next FID after those for
	 * the initial state.  When states are flushed, only FID's
	 * after this point will be flushed.
	 */
	r->posreset = r->posnext;
}

/******************************************************************************

Routine:
	eg_clrstates()

Syntax:

	static void eg_clrstat();

	RE_RE *r;	// Pointer to regular expression to be matched.

Description:

	Called by eg_getstate() when there is no room for the next state.
	eg_clrstates() effectively flushes from memory all stored states
	except for the initial state.

Global Variables:

	None

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
eg_clrstates(r)
RE_RE *r;
{
	register int i;

	/* Reset the number of states so that all the states, except for the
	 * initial state, will be written over.
	 */
	r->nstates = 1;

	/* Reset the number of FID's so that all the FID's, except for those
	 * for the initial state, will be written over.
	 */
	r->posnext = r->posreset;

	/* Since all of the states, except for the initial state, will be
	 * overwritten, all references to states other than the initial state
	 * in the transition table of the initial state must be reinitialized.
	 */
	for (i = 0; i < NCHARS ; i++) {
		if (r->states[0].tab[i] != &r->states[0]) r->states[0].tab[i] = 0;
	}
}

/******************************************************************************

Routine:
	eg_getstate()

Syntax:

	int eg_getstate();

	RE_RE *r;	// Pointer to regular expression to be matched.

Description:

	Called by addstate() before storing a new state in memory.
	eg_getstate() makes sure that there is room to store
	the new state in memory, returns the next entry in the
	array of states in which to store the new state, and increments
	the state count.

Global Variables:

	None

Files:
	None

Return Status:

	Returns the array element in which to store the new state.

*******************************************************************************/

int
eg_getstate(r)
register RE_RE *r;
{
	/* Will the next state overflow the end of the array? */
	if (r->nstates >= r->statelim) {
		/* We have no room for the next state.  We will flush the
		 * state buffer.  Increment the flush count.
		 */
		r->flushed++;

		/* If this is more flushes than our threshold, we increase
		 * the size of the array first, and reset the flush count
		 * before flushing.  However, we do not go above STATELIM
		 * in the number of elements.
		 */
		if (r->statelim < STATELIM && r->flushed > r->threshold) {
			r->statelim *= 2;
			if (r->statelim > STATELIM) r->statelim = STATELIM;
			r->states = (STATE *)cgrealloc2((char *)r->states,
					r->statelim, sizeof(STATE), "states");
			/* We raise the threshold for the number of flushes
			 * before the next memory expansion.
			 */
			r->threshold++;
			r->flushed = 0;
#ifdef	DEBUG
			printf("%d STATES INCREASED\n", r->statelim);
#endif
		}
		/* Now, flush the state buffer. */
		eg_clrstates(r);
#ifdef	DEBUG
		printf("%d STATES FLUSHED\n", r->statelim);
#endif
	}
	/* Return the next slot in the array of states,
	 * and increment the state count.
	 */
	return(r->nstates++);
}

/******************************************************************************

Routine:
	eg_posalloc()

Syntax:

	int eg_posalloc();

	RE_RE *r;	// Pointer to regular expression to be matched.
	int n;		// Number of FID's associated with the new state.

Description:

	Called by addstate() before storing a new state to memory.
	eg_posalloc() makes sure that there is room to store the FID's
	associated with new state in memory, returns the next entry in the
	array of FID's in which to store the new FID's, and increments
	the FID count.

Global Variables:

	None

Files:
	None

Return Status:

	Returns the array element in which to begin to store the FID for the state.

*******************************************************************************/

int
eg_posalloc(r, n)
RE_RE *r;
register unsigned int n;
{
	register int j;

	/* Get the next slot in which to store the FID's. */
	j = r->posnext;

	/* Increment the FID count. */
	r->posnext += n;

	/* Check if there is enough memory to store all the FID's for the state */
	if (r->posnext >= r->nposalloc) {
#ifdef	DEBUG
		printf("OLD NPOSALLOC %d\n", r->nposalloc);
#endif
		/* We need more memory.  Keep doubling the memory needed until
		 * we have enough, but don't go over POSLIMIT FID's.
		 */
		while ((r->nposalloc < r->posnext) && (r->nposalloc < POSLIMIT))
			r->nposalloc *= 2;
		/* If this is still not enough, increment the memory to
		 * be the smallest multiple of POSSTEP that is enough.
		 */
		if (r->nposalloc < r->posnext) {
			r->nposalloc = (r->posnext+POSSTEP-1)/POSSTEP;
			r->nposalloc *= POSSTEP;
		}
#ifdef	DEBUG
		printf("NEW NPOSALLOC %d\n", r->nposalloc);
#endif
		/* Now, do a realloc to get the memory, without disturbing
		 * the current contents of the FID array.
		 */
		r->posbase = (FID *)cgrealloc2((char *)r->posbase,
			r->nposalloc, sizeof(FID), "posbase");
	}
	/* Return array element in which to begin storing FID's for the state. */
	return(j);
}
