	static char sccsid[]="%Z%	trails.c	%E% %U%";
/******************************************************************************

                ###############################################
                # Copyright 1987-2000 Lucent Technologies Inc #
		# Author:  Bill Tanenbaum		      #
                ###############################################

Module:	trails.c [%P%]

Description:

	Implements trail pattern options (-t, +t, -T).

Functions:
	trail_patterns
	store_trail
	check_trail

*******************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "re.h"

static void store_trail();
static int check_trail();
static BOOL Need_compile = NO;

/******************************************************************************

Routine:
	trail_patterns()

Syntax:
	void
	trail_patterns(begin, end, pat, beg_delim, end_delim, tpat, cpat, savematches)
	CHAR	*begin;		// Pointer to beginning of text to search
	CHAR	*end;		// Pointer to end of text to search
	PATTERN *pat;		// Pointer to match pattern
	PATTERN *beg_delim;	// Pointer to delimiter pattern for beg. of window
	PATTERN *end_delim;	// Pointer to delimiter pattern for end of window
	PATTERN *tpat;		// Pointer to trail pattern
	PATTERN *cpat;		// Pointer to "combination pattern", used
				//  for two separate purposes:
				//  1) to merge matches to the match pattern with
				//  matches to the trail pattern.
				//  2) to store fixed strings in a Commentz-Walter
				//  structure, so that matches to these fixed
				//  strings can be determined.
	int	savematches;	// Number of stored matches left over from the
				//  previous text buffer that must be saved.
				//  Note that these matches must all be BEFORE
				//  "begin".  None can be within the "begin"
				//  to "end" range.

Description:

	trail_patterns() is the main routine for implementing the -t (or +t or -T)
	options.  It is called by exec_buffer() after all the matches to the
	match pattern in the text buffer have been found and windowed.

	The purpose of trail patterns() is:

	1) To find all matches to the trail pattern that occur within windows
	around matches to the match pattern.  The corresponding fixed strings
	of text that are so matched are stored, so that subsequent matches to the
	fixed strings will be processed similarly to matches to the match pattern.

	2) To find all matches to fixed strings previously stored, as indicated
	above in 1).  These matches are then merged with the matches to the
	match pattern for possible output.

Global Variables:

	CHAR *Boffset		// Pointer to beginning of file in buffer
	CHAR *Buffend		// Pointer to last character of text in text buffer
	CHAR *Buffer		// Pointer to beginning of text buffer
	BOOL File_more 		// Does more of file still need to be read in?
	int  Flags[]		// Command option flags
	CHAR *Last_NL		// Pointer to last new-line character in text buffer

Files:
	None

Return Status:

	None

*******************************************************************************/

void
trail_patterns(begin, end, pat, beg_delim, end_delim, tpat, cpat, savematches)
CHAR *begin, *end;
PATTERN *pat, *beg_delim, *end_delim, *tpat, *cpat;
int savematches;
{
	register CGWINDOW *p, *t, *c, *pkept;
	CGWINDOW *pf, *tf, *cf;
	int keep;
	unsigned int utmp;

	/* If the -t option (or +t or -T) was not used, just return. */
	if (!Flags['t']) return;

	/* If this is the first call of trail_patterns for this file,
	 * we must do some initialization.
	 */
	if (Boffset == Buffer) {
		/* If there is a structure already allocated
		 * to hold stored fixed strings, it must be left over
		 * from the last file.
		 * In order not to gobble memory, we free the tab
		 * structure,  the NODE structures, and finally
		 * the RE_CW structure itself.
		 */
		if (cpat->cw_ptr) {
			FREE(cpat->cw_ptr->bmpat);
			FREE(cpat->cw_ptr->shift1);
			FREE(cpat->cw_ptr->tab);
			re_cwfreenodes(cpat->cw_ptr->root);
			FREE(cpat->cw_ptr);
		}
		/* Initialize a Commentz-Walter structure for holding the
		 * fixed strings, and set the matching algorithm
		 * to Commentz-Walter.
		 */
		cpat->cw_ptr = re_cwinit(cpat);
		cpat->procfn = re_cwexec;
		cpat->succfn = 0;
	}

	/* Get matches to trail pattern in text buffer. */
	tpat->nmatches = 0;

	if (cpat->cw_ptr->nodeid > 1) {
		/* We have previously saved fixed strings.
		 * Therefore, we have to look everywhere for matches to
		 * the trail pattern.  Find all such matches.
		 */
		tpat->nmatches = pattern_match(tpat, begin, end, 0);
	} else {
		/* We have no previously saved fixed strings.
		 * Therefore, until we find a string that will be saved,
		 * we only need to look for matches to trail patterns
		 * inside windows.  So, for better performance, we
		 * search only in windows.
		 */
		/* If we have no windows, just return. */
		if (!pat->nmatches) return;

		/* Skip past any windows that end before "begin".
		 * We can skip these, since they are outside our search range.
		 */
		p = pat->matches;
		pf = p + pat->nmatches;
		while (p < pf && p->wend <= begin) p++;

		/* If we have a window that straddles "begin", search the
		 * portion of that window after "begin" for matches to the
		 * trail pattern.
		 */
		if (p < pf && begin > p->wbegin) {
			tpat->nmatches = pattern_match(tpat, begin, p->wend, 0);
			++p;
		}
		/* If we didn't just match a trail pattern, search each
		 * subsequent window until we find a match or run out of
		 * windows.
		 */
		--p;
		while (!tpat->nmatches && ++p < pf) {
			tpat->nmatches = pattern_match(tpat, p->wbegin, p->wend, 0);
		}
		/* We have either stopped at a match, OR we have searched
		 * all windows.  If we found a match, we now have to search
		 * the entire remainder of the buffer following the last
		 * window searched, including the area outside windows.
		 * This is because the string that we just found will be
		 * saved, and any subsequent matches to trail patterns
		 * may match it.
		 */
		if (tpat->nmatches) {
			/* Find remaining matches to trail pattern. */
			tpat->nmatches = pattern_match(tpat, p->wend, end, tpat->nmatches);
		}
	}

	/* If we found no matches to trail patterns in this invocation
	 * of trail_patterns(), just return.
	 */
	if (!tpat->nmatches) return;

	/* 
	 * Check matches to trail patterns to see if they are in windows,
	 * and if not, if they match stored text.
	 */
	/* Set pointers to first and last pattern match windows, but skip those
	 * saved from last pass, since they are before "begin".
	 */
	p = pat->matches;
	pf = p + pat->nmatches;
	p += savematches;
	/* Set pointers to first and last trail pattern matches. */
	t = tpat->matches;
	tf = t + tpat->nmatches;
	/* Some trail pattern matches will be kept and some discarded.
	 * Save a pointer to the slot in which to save the next kept
	 * match.
	 */
	pkept = t;

	/* Loop over trail pattern matches */
	for(; t < tf; t++) {
		/* Skip over pattern matches that are before trail pattern. */
		while (p < pf && p->wend <= t->begin) p++;
		/* Check if the trail pattern is (at least partly) inside
		 * the next pattern match window.
		 */
		if (p < pf && p->wbegin < t->end) {
			CGWINDOW *pp;
			/* The match is in (or overlaps) a window.
			 * Save a fixed string for future matching.
			 */
			store_trail(t, cpat);
			/* The fixed string that matched the trail pattern has
			 * been saved.  Now, we must decide whether or not
			 * this match needs to be kept for possible output.
			 * If the match is not inside an actual pattern match,
			 * we must keep it, because it may extend the window
			 * or need to be highlighted, or both.
			 */
			/* keep the match, unless determined otherwise. */
			keep = 1;
			/* Check for overlap of this trail pattern with
			 * pattern matches.
			 */
			for (pp = p; keep && pp < pf && pp->begin < t->end; pp++) {
				/* Skip over pattern matches that precede
				 * this trail pattern match.
				 */
				if (pp->end <= t->begin) continue;

				if (pp->begin <= t->begin && pp->end >= t->end) {
					/* The trail pattern match is entirely
					 * within a pattern match.  It is
					 * therefore redundant, and we can
					 * discard it.
					 */
					keep = 0;
				} else if (pp->begin >= t->begin && pp->end <= t->end) {
					/* The pattern match is entirely
					 * within the trail pattern match.
					 * Discard the trail pattern match,
					 * because it is the simplest course
					 * of action for this very low runner.
					 */
					keep = 0;
				} else if (pp->end > t->begin) {
					/* The trail pattern match overlaps
					 * a pattern match.  Keep it, but
					 * truncate the overlapping portion
					 * of the trail pattern match.
					 */
					t->begin = pp->end;
				} else if (pp->begin < t->end) {
					/* The trail pattern match overlaps
					 * a pattern match.  Keep it, but
					 * truncate the overlapping portion
					 * of the trail pattern match.
					 */
					t->end = pp->begin;
				}
			}
		} else {
			/* The match is not in a window.  Keep it only if
			 * it matches a previously saved fixed string.
			 */
			keep = check_trail(t, cpat);
		}
		if (keep) {
			/* The trail pattern match is a keeper.
			 * Expand it to complete lines.
			 * Join it with the previous kept trail pattern
			 * match if they touch or overlap.
			 */
			CHAR *b = memchr_back(t->begin, NL);
			CHAR *f = memchr_fwd(t->end, NL);
			if (pkept > tpat->matches && b <= (pkept-1)->end) {
				/* Join the match with the previous kept
				 * match, since they overlap.
				 */
				pkept--;
				if (b < pkept->begin) pkept->begin = b;
				if (f > pkept->end) pkept->end = f;
			} else {
				/* Keep this match. */
				pkept->begin = b;
				pkept->end = f;
				pkept->wbegin = pkept->wend = 0;
			}
			pkept++;
		}
	}
	/* If no trail pattern were kept, return. */
	tpat->nmatches = pkept - tpat->matches;
	if (!tpat->nmatches) return;

	/* 
	 * Check kept matches to trail patterns to see if they overlap with
	 * the windows around pattern matches.  If so, truncate the portions
	 * of the windows that overlap the trail pattern matches.
	 * Don't do this truncation if -o was specified on the command line.
	 */
	p = pat->matches;
	pf = p + pat->nmatches;
	t = tpat->matches;
	tf = t + tpat->nmatches;

	if (!Flags['o']) {
	/* Loop over trail pattern matches. */
		for(; t < tf; t++) {
			CGWINDOW *pp;
			/* Skip windows that are completely before trail pattern. */
			while (p < pf && p->wend <= t->begin) p++;

			/* truncate windows that extend into trail pattern. */
			for (pp = p; pp < pf && pp->wbegin < t->end; pp++) {
				if (pp->end <= t->begin) {
					if (pp->wend > t->begin) pp->wend = t->begin;
				} else if (pp->begin >= t->end) {
					if (pp->wbegin < t->end) pp->wbegin = t->end;
				}
			}
		}
	}
	/* We now merge the pattern ("pat") and trail pattern  ("tpat") matches.
	 * However, for convenience, we first "move" the "pat" matches to "cpat".
	 * and make sure there are enough array slots in pat->matches.
	 * We then merge "tpat" and "cpat" matches back into "pat".
	 */
	/* Exchange the "pat" and "cpat" match buffers.  This effectively
	 * moves the "pat" matches to "cpat".
	 */
	cpat->nmatches = pat->nmatches;

	t = cpat->matches;
	cpat->matches = pat->matches;
	pat->matches = t;

	utmp = cpat->match_buffers;
	cpat->match_buffers = pat->match_buffers;
	pat->match_buffers = utmp;

	/* Set pat->nmatches to the sum of the number of "cpat" and "tpat"
	 * matches, and allocate enough slots in "pat" for this number.
	 * (match_alloc() uses pat->nmatches to determine how many slots to
	 * allocate).
	 */
	pat->nmatches = cpat->nmatches + pat->nmatches;
	match_alloc(pat);

	/* Now that we have allocated slots, initialize the number of merged matches
	 * to zero and do the merge.  We are merging "cpat" and "tpat" matches
	 * into "pat".
	 */
	pat->nmatches = 0;

	p = pat->matches;
	t = tpat->matches;
	tf = t + tpat->nmatches;
	c = cpat->matches;
	cf = c + cpat->nmatches;

	/* OK. Now do the merge.  */
	/* Loop over the trail pattern matches, but go one past the last one!  */
	for(; t <= tf; t++, p++) {
		/* Store all "cpat" matches before the current "tpat" match
		 * or, if t >= tf, after there are no more "tpat" matches.
		 */
		while (c < cf && (t >= tf || c->wend <= t->begin)) {
			pat->nmatches++;
			*p++ = *c++;
		}
		/* If no more "tpat" matches, we are done with the merge. */
		if (t >= tf) break;
		/* Now store the trail pattern match. */
		pat->nmatches++;
		*p = *t;
		/* The trail pattern match needs to be windowed.  Indicate
		 * this by setting the window limits to 0.
		 */
		p->wbegin = p->wend = 0;
	}
	/* The merge is complete.  Now, do the windowing of the merged
	 * matches.   Matches already windowed will be passed over.
	 */
	windows(pat, beg_delim, end_delim, Buffer, File_more ? Buffend+1 : Last_NL);
}

/******************************************************************************

Routine:

	store_trail()

Syntax:

	static void store_trail(t, cpat)

	CGWINDOW *t;	// Pointer to a match to the trail pattern
	PATTERN *cpat;	// Pointer to a pattern structure that will
			//   be used for storing the text
			//   that matched the trail pattern

Description:

	store_trail() is called when a match is found to the trail pattern,
	and the match is inside a window around a match to a pattern.

	store trail() stores the text that matched the trail pattern,
	so that future matches to this exact text will be output.

Global Variables:

	Need_compile	// Does the pattern consisting of fixed strings
			//   need recompilation.

Files:
	None

Return Status:

	None

*******************************************************************************/

static void
store_trail(t, cpat)
register CGWINDOW *t;
register PATTERN *cpat;
{
	CHAR cb, ce;

	/* We store the text with additional new-line characters delimiting
	 * the text.  The reason we do this is explained in check_trail(),
	 * the routine that checks for matches to these stored strings.
	 */
	/* Save the values of the preceding and following characters, and
	 * then set them to new-lines.
	 */
	cb = *(t->begin-1);
	ce = *(t->end+1);
	*(t->begin-1) = NL;
	*(t->end+1) = NL;

	/* Store the matched text, including the surrounding new-lines.
	 * The last argument to re_cwadd() must be one past the last character
	 * of the string to be stored.
	 */
	re_cwadd(cpat->cw_ptr, t->begin-1, t->end+2);

	/* Restore the text. */
	*(t->begin-1) = cb;
	*(t->end+1) = ce;

	/* Since we have stored a new string, we will have to recompile before
	 * we can check for matches to the stored strings.
	 */
	Need_compile = YES;
}

/******************************************************************************

Routine:

	check_trail()

Syntax:

	static void check_trail(t, cpat)

	CGWINDOW *t;	// Pointer to a match to the trail pattern
	PATTERN *cpat;	// Pointer to the pattern structure that
			//   contains previously stored strings.

Description:

	check_trail() checks if the text that matched the trail pattern
	exactly matches one of the previously stored fixed strings.

Global Variables:

	Need_compile	// Does the pattern consisting of fixed strings
			//   need recompilation.

Files:
	None

Return Status:

	1	//  The text that matched the trail pattern
		//    matches a previously saved fixed string.
	0	//  The text that matched the trail pattern
		//    does not match a fixed string.

*******************************************************************************/

static int
check_trail(t, cpat)
register CGWINDOW *t;
register PATTERN *cpat;
{
	CHAR cb, ce;
	int ret;

	/* If we haven't stored any strings, we can have no matches. */
	if (cpat->cw_ptr->nodeid <= 1) {
		return(0);
	}

	/* If we stored an additional string since the last call to
	 * check_trail(), we need to recompile the pattern.
	 * In order not to gobble memory, we first free the structures
	 * from the previous compilation.  The RE_CW and NODE structures
	 * are still needed.
	 */
	if (Need_compile) {
		FREE(cpat->cw_ptr->bmpat);
		cpat->cw_ptr->bmpat = 0;
		FREE(cpat->cw_ptr->shift1);
		cpat->cw_ptr->shift1 = 0;
		FREE(cpat->cw_ptr->tab);
		cpat->cw_ptr->tab = 0;
		re_cwcomp(cpat->cw_ptr);
		Need_compile = NO;
	}
	/* Set the counts to find the first match. */
	cpat->N_start = 0;
	cpat->N_stop = 1;

	/* We call pattern_match() to actually check for a match.
	 * However, we first delimit the text with new-line characters.
	 * The fixed strings were also stored with surrounding new-line
	 * delimiters.  We add the new-line delimiters to guarantee that
	 * pattern_match() will not find a match unless a stored fixed string
	 * matches the entire text, rather than a substring of the text.
	 * For example, if we had stored "EVENT=1234" as a fixed string,
	 * a subsequent occurrence of "EVENT=12345" would match it.
	 * However, if we store "<NL>EVENT=1234<NL>", <NL>EVENT=12345<NL>
	 * will not match it.
	 */
	/* Save the values of the preceding and following characters, and
	 * then set them to new-lines.
	 */
	cb = *(t->begin-1);
	ce = *(t->end+1);
	*(t->begin-1) = NL;
	*(t->end+1) = NL;

	/* Check for a match. */
	ret = pattern_match(cpat, t->begin-1, t->end+1, 0);

	/* Restore the text. */
	*(t->begin-1) = cb;
	*(t->end+1) = ce;
	return(ret);
}
